package struclust.eval.comparison;

import graph.Graph;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.stream.Collectors;

import struclust.Cluster;

import com.google.common.base.Preconditions;
import com.google.common.collect.Sets;

/**
 * Contingency matrix is defined as
 * 
 * <pre>
 * n_ij = |clusters1_i \cap clusters2_j|
 * </pre>
 * 
 * where clusters1_i is the i-th cluster from the fist clustering and
 * clusters2_j the j-th cluster form the second clustering.
 * 
 * @author Till Schäfer
 * 
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class ContingencyMatrix<NL, EL, G extends Graph<NL, EL>> {

    private final int[][] matrix;
    private final int[] sizes1;
    private final int[] sizes2;
    private final int numElem;

    /**
     * Constructor
     * 
     * Please note that clusters1 and clusters2 must contain the same set of
     * graphs.
     * 
     * @param clusters1
     *            the first Clustering
     * @param clusters2
     *            the second Clustering
     */
    public ContingencyMatrix(Collection<Cluster<NL, EL, G>> clusters1, Collection<Cluster<NL, EL, G>> clusters2) {
        numElem = clusters1.stream().mapToInt(c -> c.size()).sum();
        Preconditions.checkArgument(numElem == clusters2.stream().mapToInt(c -> c.size()).sum(),
                "clusterings does not have the same size (in terms of cluster elements)");

        matrix = new int[clusters1.size()][clusters2.size()];
        sizes1 = new int[clusters1.size()];
        sizes2 = new int[clusters2.size()];
        ArrayList<Set<Graph<NL, EL>>> sets1 = new ArrayList<>(clusters1.size());
        ArrayList<Set<Graph<NL, EL>>> sets2 = new ArrayList<>(clusters2.size());

        Iterator<Cluster<NL, EL, G>> it1 = clusters1.iterator();
        for (int i = 0; i < clusters1.size(); i++) {
            Cluster<NL, EL, G> cluster = it1.next();
            sets1.add(cluster.stream().map(gc -> gc.getGraph()).collect(Collectors.toSet()));
            sizes1[i] = cluster.size();
        }
        Iterator<Cluster<NL, EL, G>> it2 = clusters2.iterator();
        for (int i = 0; i < clusters2.size(); i++) {
            Cluster<NL, EL, G> cluster = it2.next();
            sets2.add(cluster.stream().map(gc -> gc.getGraph()).collect(Collectors.toSet()));
            sizes2[i] = cluster.size();
        }

        for (int i = 0; i < clusters1.size(); i++) {
            for (int j = 0; j < clusters2.size(); j++) {
                matrix[i][j] = Sets.intersection(sets1.get(i), sets2.get(j)).size();
            }
        }
    }

    /**
     * Returns the contingency value n_ij
     * 
     * @param i
     *            index i
     * @param j
     *            index j
     * @return n_ij
     */
    public int getContingencyValue(int i, int j) {
        return matrix[i][j];
    }

    /**
     * Return the size of the i-th cluster in clustering 1
     * 
     * @param i
     *            index
     * @return the cluster size
     */
    public int getClusterSize1(int i) {
        return sizes1[i];
    }

    /**
     * Return the size of the j-th cluster in clustering 2
     * 
     * @param j
     *            index
     * @return the cluster size
     */
    public int getClusterSize2(int j) {
        return sizes2[j];
    }

    /**
     * Return the number of clusters in clustering 1
     * 
     * @return the number of clusters
     */
    public int numClusters1() {
        return matrix.length;
    }

    /**
     * Return the number of clusters in clustering 2
     * 
     * @return the number of clusters
     */
    public int numClusters2() {
        if (numClusters1() > 0) {
            return matrix[0].length;
        } else {
            return 0;
        }
    }

    /**
     * Return the sum of all entries
     * 
     * @return the sum of all entries
     */
    public int sumAll() {
        int retVal = 0;
        int numClusters1 = numClusters1();
        int numClusters2 = numClusters2();
        for (int i = 0; i < numClusters1; i++) {
            for (int j = 0; j < numClusters2; j++) {
                retVal += matrix[i][j];
            }
        }

        return retVal;
    }

    /**
     * Return the sum of all squared entries
     * 
     * @return the sum of all squared entries
     */
    public BigInteger sumAllSquared() {
        BigInteger retVal = new BigInteger("0");
        int numClusters1 = numClusters1();
        int numClusters2 = numClusters2();
        for (int i = 0; i < numClusters1; i++) {
            long innerSum = 0;
            for (int j = 0; j < numClusters2; j++) {
                innerSum += Math.pow(matrix[i][j], 2);
            }
            retVal = retVal.add(BigInteger.valueOf(innerSum));
        }

        return retVal;
    }

    /**
     * Return the sum of all entries in row i (from cluster1)
     * 
     * @param i
     *            the row index
     * @return the sum of all entries in row i (from cluster1)
     */
    public int sumCluster1(int i) {
        int retVal = 0;
        for (int j = 0; j < matrix[0].length; j++) {
            retVal += matrix[i][j];
        }

        return retVal;
    }

    /**
     * Return the sum of all entries in column j (from clusters2)
     * 
     * @param j
     *            the column index
     * @return the sum of all entries in column j (from cluster2)
     */
    public int sumCluster2(int j) {
        int retVal = 0;
        for (int i = 0; i < matrix.length; i++) {
            retVal += matrix[i][j];
        }

        return retVal;
    }

    /**
     * Return the number of elements contained in each clustering.
     * 
     * @return the number of elements
     */
    public int numElements() {
        return numElem;
    }
}
