package struclust.eval.comparison;

import graph.Graph;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.RoundingMode;
import java.util.Collection;

import struclust.Cluster;
import struclust.util.Numerical;

/**
 * Implements the Fowlkes and Mallows Comparison Method.
 * 
 * A Method for Comparing Two Hierarchical Clusterings, E. B. Fowlkes and C. L.
 * Mallows, Page 553 of 553-569 (Journal of the American Statistical
 * Association)
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class FowlkesMallowsCC<NL, EL, G extends Graph<NL, EL>> implements ClusteringComparison<NL, EL, G> {
    @Override
    public String getDescription() {
        return "FowlkesMallowsCC";
    }

    @Override
    public double calc(Collection<Cluster<NL, EL, G>> clusters1, Collection<Cluster<NL, EL, G>> clusters2) {
        ContingencyMatrix<NL, EL, G> m = new ContingencyMatrix<>(clusters1, clusters2);

        BigInteger t = m.sumAllSquared().subtract(BigInteger.valueOf(m.numElements()));
        BigInteger p = new BigInteger("0");
        for (int i = 0; i < m.numClusters1(); i++) {
            p = p.add(BigInteger.valueOf((long) Math.pow(m.sumCluster1(i), 2)));
        }
        BigInteger q = new BigInteger("0");
        for (int i = 0; i < m.numClusters2(); i++) {
            q = q.add(BigInteger.valueOf((long) Math.pow(m.sumCluster2(i), 2)));
        }


        BigDecimal mult = new BigDecimal(p.multiply(q));
        BigDecimal sqrtMult = Numerical.bigSqrt(mult);
        BigDecimal val = new BigDecimal(t);
        val = val.divide(sqrtMult, 5, RoundingMode.HALF_UP);

        double retVal = val.doubleValue();
        return retVal;
    }

    @Override
    public double maxValue() {
        return 1;
    }

}
