package struclust.eval.comparison;

import graph.Graph;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.OptionalInt;
import java.util.Set;

import struclust.Cluster;
import struclust.datastructures.CountingInt;
import struclust.graph.GraphContainer;

import com.google.common.base.Preconditions;

/**
 * Purity measure as defined by:
 * 
 * Zhao, Ying, George Karypis, and Ding-Zhu Du. Criterion functions for document
 * clustering. Technical Report, 2005.
 * 
 * Other than PurityCC this measure does not take the reference clustering into
 * account, but uses attached properties of the GraphContainers to calculate the
 * value. THis allows to assign multiple labels to a single graph and therefore
 * comparison with overlapping ground truths.
 * 
 * Note: the second clustering argument is ignored!
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class OverlappingPurityCC<NL, EL, G extends Graph<NL, EL>> implements ClusteringComparison<NL, EL, G> {

    private Set<String> properties;
    private String multiLabelSplit;
    private boolean useSplitting;

    /**
     * Constructor
     * 
     * @param properties
     *            the properties used to measure the
     * @param multiLabelSplit
     *            if not null, the properties are split by this string and each
     *            resulting substring is considered a separate label
     */
    public OverlappingPurityCC(Set<String> properties, String multiLabelSplit) {
        Preconditions.checkArgument(!properties.isEmpty(), "properties must have at least one property defined");

        this.multiLabelSplit = multiLabelSplit;
        this.useSplitting = multiLabelSplit != null;
        this.properties = properties;
    }

    @Override
    public String getDescription() {
        return "OverlappingPurityCC";
    }

    @Override
    public double calc(Collection<Cluster<NL, EL, G>> clusters1, Collection<Cluster<NL, EL, G>> clusters2) {
        double retVal = 0;
        long numGraphs = clusters1.stream().flatMap(c -> c.stream()).count();

        for (Cluster<NL, EL, G> c : clusters1) {
            HashMap<String, CountingInt> labelCount = new HashMap<>();
            for (GraphContainer<NL, EL, G> gc : c) {
                List<String> graphlabels = getGraphlabels(gc);
                for (String label : graphlabels) {
                    CountingInt.incrementCount(labelCount, label);
                }
            }
            OptionalInt maxClusterLabelCount = labelCount.entrySet().stream().mapToInt(e -> e.getValue().get()).max();
            retVal += maxClusterLabelCount.orElse(0);
        }

        return retVal / numGraphs;
    }

    /**
     * returns the labels associated with a graph
     * 
     * @param gc
     *            the graph
     * @return the graph labels
     */
    private List<String> getGraphlabels(GraphContainer<NL, EL, G> gc) {
        ArrayList<String> retVal = new ArrayList<>();
        for (String prop : properties) {
            String propVal = gc.getProperty(prop);
            if (propVal != null && !propVal.isEmpty()) {
                if (useSplitting) {
                    for (String subProp : propVal.split(multiLabelSplit)) {
                        String trimmed = subProp.trim();
                        if (!trimmed.isEmpty()) {
                            retVal.add(trimmed);
                        }
                    }
                } else {
                    retVal.add(propVal);
                }
            }
        }
        return retVal;
    }

    @Override
    public double maxValue() {
        return 1;
    }

}
