package struclust.eval.comparison;

import java.util.Collection;
import java.util.HashMap;
import java.util.Set;
import java.util.stream.Collectors;

import graph.Graph;
import struclust.Cluster;
import struclust.graph.GraphContainer;

/**
 * Purity measure as defined by:
 * 
 * Zhao, Ying, George Karypis, and Ding-Zhu Du. Criterion functions for document
 * clustering. Technical Report, 2005.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class PurityCC<NL, EL, G extends Graph<NL, EL>> implements ClusteringComparison<NL, EL, G> {

    @Override
    public String getDescription() {
        return "PurityCC";
    }

    @Override
    public double calc(Collection<Cluster<NL, EL, G>> clusters1, Collection<Cluster<NL, EL, G>> clusters2) {
        assert assertSameElements(clusters1, clusters2) : "the comparing clusterings contain different elements";
        
        HashMap<G, Integer> labels = new HashMap<>();
        for (Cluster<NL, EL, G> c : clusters2) {
            for (GraphContainer<NL, EL, G> gc : c) {
                labels.put(gc.getGraph(), c.getId());
            }
        }

        long numGraphs = clusters1.stream().flatMap(c -> c.stream()).count();
        assert numGraphs == clusters2.stream().flatMap(c -> c.stream()).count();

        return (double) clusters1
                .stream()
                .mapToInt(
                        c -> c.stream()
                                .collect(Collectors.groupingBy((GraphContainer<NL, EL, G> gc) -> labels.get(gc.getGraph())))
                                .entrySet().stream().mapToInt(e -> e.getValue().size()).max()
                                .orElseThrow(() -> new IllegalStateException("each cluster must contain an element")))
                .sum()
                / numGraphs;
    }

    private boolean assertSameElements(Collection<Cluster<NL, EL, G>> clusters1, Collection<Cluster<NL, EL, G>> clusters2) {
        Set<GraphContainer<NL, EL, G>> graphs1 = clusters1.stream().flatMap(c -> c.stream()).collect(Collectors.toSet());
        Set<GraphContainer<NL, EL, G>> graphs2 = clusters2.stream().flatMap(c -> c.stream()).collect(Collectors.toSet());
        return graphs1.equals(graphs2);
    }

    @Override
    public double maxValue() {
        return 1;
    }

}
