/*
 * Scaffold Hunter
 * Copyright (C) 2006-2008 PG504
 * Copyright (C) 2010-2011 PG552
 * Copyright (C) 2012 LS11
 * See the file README.txt in the root directory of the Scaffold Hunter
 * source tree for details.
 *
 * Scaffold Hunter is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Scaffold Hunter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package struclust.eval.metamodules;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.Map.Entry;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.comparison.AccFunction;
import struclust.eval.EvalResult;
import struclust.util.Numerical;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;

/**
 * Computes the accumulated values for all {@link EvalResult}s with the same
 * settings.
 * 
 * Works only with numerical Values!
 * 
 * @author Till Schäfer
 */
public class AccEMM extends AbstractEMM {
    private static Logger logger = LoggerFactory.getLogger(AccEMM.class);

    private final AccFunction accumulationFuncition;

    private Collection<String> filter;

    /**
     * Constructor.
     * 
     * Filters the metadata key "series uuid" for settings comparison.
     * 
     * @param accumulationFuncition
     *            the {@link AccFunction}
     */
    public AccEMM(AccFunction accumulationFuncition) {
        this.accumulationFuncition = accumulationFuncition;
        filter = Collections.singleton("series uuid");
    }

    /**
     * Constructor
     * 
     * @param accumulationFuncition
     *            the {@link AccFunction}
     * @param filter
     *            specify, which metadata keys should be ignored for settings
     *            comparison. No filter is applied if the value is null.
     */
    public AccEMM(AccFunction accumulationFuncition, Collection<String> filter) {
        this.accumulationFuncition = accumulationFuncition;
        this.filter = filter;
    }


    /**
     * Calculates the accumulation of the results
     * 
     * @param results
     *            the {@link EvalResult}s to accumulate
     * @return the accumulated {@link EvalResult}s
     */
    @Override
    public Collection<EvalResult> process(Collection<EvalResult> results) {
        LinkedList<EvalResult> retVal = Lists.newLinkedList();
        Collection<Collection<EvalResult>> groupedResults = groupResults(results);

        for (Collection<EvalResult> sameMeasurement : groupedResults) {
            retVal.add(accumulateSingleGroup(sameMeasurement));
        }

        return retVal;
    }

    /**
     * Return one new {@link EvalResult} with the accumulated values of the
     * {@link EvalResult}s with the same settings.
     * 
     * @param results
     * @return
     */
    protected EvalResult accumulateSingleGroup(Collection<EvalResult> results) {
        Preconditions.checkArgument(results.size() > 0);

        int collectionSize = results.size();
        LinkedHashMap<String, Double> accumulatedValues = Maps.newLinkedHashMap();

        /*
         *  create accumulated EvalResult
         */
        EvalResult arbitraryResult = results.iterator().next();
        String measurement = arbitraryResult.getMeasurement() + "-" + accumulationFuncition.name() + " of "
                + collectionSize + " values";
        EvalResult retVal = new EvalResult(measurement, arbitraryResult.getMetadata());
        // filter series uuid as it does not make sense for acc values
        retVal.removeMetadata("series uuid");
        // add references to acc EvalResults 
        retVal.addRefs(results);

        // initialize retVal results
        for (String key : arbitraryResult.getResults().keySet()) {
            switch (accumulationFuncition) {
            case MIN:
                accumulatedValues.put(key, Double.POSITIVE_INFINITY);
                break;
            case MAX:
                accumulatedValues.put(key, Double.NEGATIVE_INFINITY);
                break;
            case AVG:
                // same as Sum
            case SUM:
                accumulatedValues.put(key, 0.0);
                break;
            default:
                break;
            }
        }

        // accumulate retVal results
        for (String key : arbitraryResult.getResults().keySet()) {
            int accumulationCount = 0;
            for (EvalResult result : results) {
                try {
                    Double currentVal = Double.parseDouble(result.getResults().get(key));
                    Double oldVal = accumulatedValues.get(key);

                    accumulationCount++;
                    switch (accumulationFuncition) {
                    case MIN:
                        accumulatedValues.put(key, Math.min(currentVal, oldVal));
                        break;
                    case MAX:
                        accumulatedValues.put(key, Math.max(currentVal, oldVal));
                        break;
                    case AVG:
                        accumulatedValues.put(key, Numerical.incAvg(accumulationCount, oldVal, currentVal));
                        break;
                    case SUM:
                        accumulatedValues.put(key, oldVal + currentVal);
                        break;
                    default:
                        break;
                    }
                } catch (NumberFormatException ex) {
                    logger.warn("Number format {} could not be parsed. Skipping value for key {} in accumulation.",
                            result.getResults().get(key), key);
                }
            }
        }

        // add new accumulated results to retVal
        for (Entry<String, Double> singleResult : accumulatedValues.entrySet()) {
            retVal.addResult(singleResult.getKey(), singleResult.getValue().toString());
        }

        return retVal;
    }

    /**
     * Group {@link EvalResult}s based on their settings
     * 
     * @param results
     *            all {@link EvalResult}s
     * @return the groups of {@link EvalResult}s
     */
    private Collection<Collection<EvalResult>> groupResults(Collection<EvalResult> results) {
        LinkedList<Collection<EvalResult>> retVal = Lists.newLinkedList();
        while (!results.isEmpty()) {
            LinkedList<EvalResult> currentGroup = Lists.newLinkedList();
            Iterator<EvalResult> it = results.iterator();
            /*
             * the leader is compared to all other objects and defines the
             * settings for this group
             */
            EvalResult currentLeader = it.next();

            // remove leader from results
            it.remove();
            // find results with equal settings
            for (EvalResult result : results) {
                if (currentLeader.equalSettings(result, filter)) {
                    currentGroup.add(result);
                }
            }
            // remove results with equal settings from results
            results.removeAll(currentGroup);
            currentGroup.add(currentLeader);

            retVal.add(currentGroup);
        }

        return retVal;
    }

}
