package struclust.eval.metamodules;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;

import struclust.eval.EvalException;
import struclust.eval.EvalResult;
import struclust.eval.modules.EvalModule;

/**
 * Repeats the argument n times
 * 
 * @author Till Schäfer
 * 
 */
public class RepeaterEMM extends AbstractEMM {
    private static Logger logger = LoggerFactory.getLogger(RepeaterEMM.class);

    private final int count;
    private final boolean leaveFirstOut;

    private boolean resetModule;

    /**
     * Constructor <br>
     * 
     * calls {@link RepeaterEMM#RepeaterEMM(int, boolean, boolean)} with
     * resetModule=true
     * 
     * @param count
     *            the count of repeats
     * @param leaveFirstOut
     *            when true the first run is not added to the evaluation results
     */
    public RepeaterEMM(int count, boolean leaveFirstOut) {
        this(count, leaveFirstOut, true);
    }

    /**
     * Constructor
     * 
     * @param count
     *            the count of repeats
     * @param leaveFirstOut
     *            when true the first run is not added to the evaluation results
     * @param resetModule
     *            reset the {@link EvalModule} before each run
     */
    public RepeaterEMM(int count, boolean leaveFirstOut, boolean resetModule) {
        Preconditions.checkArgument((!leaveFirstOut && count > 0) || (leaveFirstOut && count > 1));
        this.count = count;
        this.leaveFirstOut = leaveFirstOut;
        this.resetModule = resetModule;
    }

    @Override
    public Collection<EvalResult> run(EvalModule<?> module) {
        Preconditions.checkNotNull(module);

        LinkedList<EvalResult> retVals = new LinkedList<>();

        for (int i = 0; i < count; i++) {
            LinkedList<EvalResult> retVal = new LinkedList<>();
            logger.info("Running repeat {} of {}", i + 1, count);

            retVal.addAll(module.run());
            if (!leaveFirstOut || i > 0) {
                retVals.addAll(retVal);
            }
        }

        return retVals;
    }

    @Override
    public Collection<EvalResult> run(List<EvalMetaModule> metaModules, EvalModule<?> module) {
        Preconditions.checkNotNull(module);
        Preconditions.checkNotNull(metaModules);

        LinkedList<EvalResult> retVals = Lists.newLinkedList();

        if (metaModules.isEmpty()) {
            return run(module);
        } else {
            EvalMetaModule metaModule = metaModules.iterator().next();
            List<EvalMetaModule> metaModulesRest = metaModules.subList(1, metaModules.size());

            for (int i = 0; i < count; i++) {
                LinkedList<EvalResult> retVal = Lists.newLinkedList();
                logger.info("Running repeat {} of {}", i + 1, count);

                if (resetModule) {
                    module.reset();
                }
                retVal.addAll(metaModule.run(metaModulesRest, module));

                if (!leaveFirstOut || i > 0) {
                    retVals.addAll(retVal);
                }
            }

            return retVals;
        }
    }

    @Override
    public Collection<EvalResult> process(Collection<EvalResult> results) throws EvalException {
        throw new UnsupportedOperationException(
                "This EvalMetaModule do not process the results, but modifies the test setup");
    }
}
