package struclust.eval.modules;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.UUID;

import struclust.eval.EvalException;
import struclust.eval.EvalResult;

import com.google.common.base.Preconditions;

/**
 * Runs a test series. That means it reuses the test data from the first
 * evaluation for all other evaluations. This module is especially helpful for
 * randomized algorithms, if one wants to evaluate different things on the same
 * test data.
 * 
 * @author Till Schäfer
 * 
 * @param <T>
 *            the type of the test data (e.g. a clustering for any test that
 *            evaluates a clustering)
 */
public class SeriesEM<T> extends AbstractEM<T> {

    private List<EvalModule<T>> modules;

    /**
     * Constructor
     * 
     * @param modules
     *            the modules to run in this series. The first module computes
     *            the test data (if no test data is provided) and all other
     *            modules are run on this test data. All other modules must be
     *            able to process precomputed test data. The number of modules
     *            must be larger or equal than one.
     */
    public SeriesEM(List<EvalModule<T>> modules) {
        Preconditions.checkArgument(modules.size() >= 1, "modules.size must be larger or equal to one");

        for (int i = 1; i < modules.size(); i++) {
            Preconditions.checkArgument(modules.get(i).canUsePrecomputedTestData(),
                    "usage of precomputed test data not supported for " + modules.get(i).getClass().getName());
        }

        this.modules = modules;
    }

    @Override
    public Collection<EvalResult> run() throws EvalException {
        Collection<EvalResult> retVal = new LinkedList<>();
        ListIterator<EvalModule<T>> it = modules.listIterator();

        EvalModule<T> first = it.next();
        retVal.addAll(first.run());

        retVal.addAll(run(first.getTestData(), it));
        
        return addSeriesUUID(retVal);
    }

    @Override
    public Collection<EvalResult> run(T testData) throws EvalException {
        Preconditions.checkArgument(modules.get(0).canUsePrecomputedTestData());

        return addSeriesUUID(run(testData, modules.listIterator()));
    }

    private Collection<EvalResult> run(T testData, ListIterator<EvalModule<T>> it) throws EvalException {
        this.testData = testData;
        Collection<EvalResult> retVal = new LinkedList<>();

        while(it.hasNext()) {
            EvalModule<T> module = it.next();
            retVal.addAll(module.run(testData));
        }
        
        return retVal;
    }

    private Collection<EvalResult> addSeriesUUID (Collection<EvalResult> results) {
        String uuid = UUID.randomUUID().toString(); 
        for (EvalResult result : results) {
            result.addMetadata("series uuid", uuid);
        }
        return results;
    }
    
    @Override
    public boolean canUsePrecomputedTestData() {
        return true;
    }

}
