package struclust.eval.modules.clustering;

import java.util.Collection;
import java.util.Collections;
import java.util.List;

import graph.Graph;
import struclust.Cluster;
import struclust.StructuralClustering;
import struclust.convergence.ConvergenceCriteria;
import struclust.eval.EvalException;
import struclust.eval.EvalResult;
import struclust.eval.modules.EvalModule;

/**
 * Evaluation module for a {@link ConvergenceCriteria}. Usually used as
 * injectionModule in each iteration of the algorithm.
 * 
 * Attention: Only ConcergenceCriteria based on the clustering result are
 * supported. Parameter <em>changes</em> is always {@link Integer#MAX_VALUE}.
 * This restriction is <b>NOT</b> enforced by the evaluation Module itself!
 * 
 * Output results contains the key <em>convergence</em> with a boolean value and
 * optionally another key <em>first</em> which is true iff convergence is true
 * for the first time (of the {@link EvalModule} existence).
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class ConvergenceEM<NL, EL, G extends Graph<NL, EL>> extends PrecompClustEM<NL, EL, G> {

    private final ConvergenceCriteria crit;
    private final boolean evalFirst;
    private boolean firstConvergence = true;

    /**
     * Constructor
     * 
     * @param clust
     *            the configured clustering
     * @param crit
     *            the {@link ConvergenceCriteria} to evaluate.
     * @param evalFirst
     *            output a result key <em>first</em> which is true iff
     *            convergence is true for the first time (of the
     *            {@link EvalModule} existence).
     */
    public ConvergenceEM(StructuralClustering<NL, EL, G> clust, ConvergenceCriteria crit, boolean evalFirst) {
        super(clust);
        this.crit = crit;
        this.evalFirst = evalFirst;
    }

    @Override
    protected Collection<EvalResult> runInternal(List<Cluster<NL, EL, G>> clustering) throws EvalException {
        EvalResult retVal = new EvalResult("ConvergenceEM-" + crit.getDescription());
        retVal.addMetadata(clust);
        
        int totalElements = clustering.stream().mapToInt(c -> c.size()).sum();
        boolean convergence = crit.convergent(Integer.MAX_VALUE, totalElements, clustering);
        
        retVal.addResult("convergence", convergence);
        if (evalFirst) {
            retVal.addResult("first", convergence && firstConvergence);
            if (convergence && firstConvergence) {
                firstConvergence = false;
            }
        }
        
        return Collections.singleton(retVal);
    }
    
    @Override
    public void reset() {
        crit.reset();
        firstConvergence = true;
    }
}
