package struclust.eval.modules.clustering;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.apache.commons.io.FilenameUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import graph.Graph;
import struclust.Cluster;
import struclust.StructuralClustering;
import struclust.eval.EvalException;
import struclust.eval.EvalResult;
import struclust.graph.GraphContainer;
import struclust.util.FsOperations;

/**
 * Saves the clustering to disk in a CSV format (cluster_id, graph_id). The
 * graph ID must be stored in the GraphContainer as Property.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class CsvClustSaverEM<NL, EL, G extends Graph<NL, EL>> extends PrecompClustEM<NL, EL, G> {
    private static final Logger logger = LoggerFactory.getLogger(CsvClustSaverEM.class);
    private static final String endl = System.getProperty("line.separator");

    private String idKey;
    private String path;
    private String basename;
    private boolean uniqueFileName;

    /**
     * Constructor
     * 
     * @param clust
     *            The configured clustering algorithm
     * @param idKey
     *            the property key of the property of the graph ID in the
     *            {@link GraphContainer}
     * @param path
     *            the folder path to save the clustering to.
     * @param basename
     *            the basename of the saved file. if null the basename "cluster"
     *            is used.
     * @param uniqueFileName
     *            if true the filename will be always made unique by adding a
     *            number as suffix.
     */
    public CsvClustSaverEM(StructuralClustering<NL, EL, G> clust, String idKey, String path, String basename,
            boolean uniqueFileName) {
        super(clust);
        this.idKey = idKey;
        this.path = path;
        this.basename = basename == null ? "cluster" : basename;
        this.uniqueFileName = uniqueFileName;
    }

    @Override
    protected Collection<EvalResult> runInternal(List<Cluster<NL, EL, G>> clustering) throws EvalException {
        String fullpath = uniqueFileName ? FsOperations.getUniqueFileName(path, basename, "csv", 3)
                : FilenameUtils.concat(path, basename + FilenameUtils.EXTENSION_SEPARATOR + "csv");
        try (FileWriter fw = new FileWriter(fullpath); BufferedWriter outElem = new BufferedWriter(fw)) {
            for (Cluster<NL, EL, G> cluster : clustering) {
                for (GraphContainer<NL, EL, G> gc : cluster) {
                    fw.write(Integer.toString(cluster.getId()));
                    fw.write(",");
                    fw.write(gc.getProperty(idKey));
                    fw.write(endl);
                }
            }
        } catch (IOException e) {
            logger.error("writing csv file failed", e);
        }
        return Collections.emptyList();
    }

}
