package struclust.eval.modules.clustering;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.apache.commons.io.FilenameUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import graph.Graph;
import struclust.Cluster;
import struclust.StructuralClustering;
import struclust.eval.EvalException;
import struclust.eval.EvalResult;
import struclust.graph.GraphContainer;
import struclust.util.FsOperations;
import struclust.util.GraphIO;

/**
 * Saves the clustering to disk in GML format.
 * 
 * The saver has two modes:
 * 
 * 1. It will create two gml files for each cluster. One contains the cluster
 * members and one contains the cluster representatives.
 * 
 * 2. It will create a single file called with a custom GML format containing
 * all clusters and representatives in a structured form.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class GmlClustSaverEM<NL, EL, G extends Graph<NL, EL>> extends PrecompClustEM<NL, EL, G> {
    private static final Logger logger = LoggerFactory.getLogger(GmlClustSaverEM.class);

    private String path;
    private boolean singleFile;
    private boolean createUniqueSubfolder;
    private String basename;
    private boolean uniqueFileName;

    /**
     * Constructor
     * 
     * @param clust
     *            The configured clustering algorithm
     * @param path
     *            the folder path to save the clustering to.
     * @param basename
     *            the basename of the saved file. if null the basename "cluster"
     *            is used.
     * @param singleFile
     *            whether to use a single file in custom GML format or to use
     *            multiple files for each cluster in standard GML format.
     */
    public GmlClustSaverEM(StructuralClustering<NL, EL, G> clust, String path, String basename, boolean singleFile) {
        this(clust, path, basename, singleFile, false, true);
    }

    /**
     * Constructor
     * 
     * @param clust
     *            The configured clustering algorithm
     * @param path
     *            the folder path to save the clustering to.
     * @param basename
     *            the basename of the saved file. if null the basename "cluster"
     *            is used.
     * @param singleFile
     *            whether to use a single file in custom GML format or to use
     *            multiple files for each cluster in standard GML format.
     * @param createUniqueSubfolder
     *            if true a unique subfolder will be created for each call of
     *            {@link #run()}
     * @param uniqueFileName
     *            if true the filename will be always made unique by adding a
     *            number as suffix.
     */
    public GmlClustSaverEM(StructuralClustering<NL, EL, G> clust, String path, String basename, boolean singleFile,
            boolean createUniqueSubfolder, boolean uniqueFileName) {
        super(clust);
        this.path = path;
        this.uniqueFileName = uniqueFileName;
        this.basename = basename == null ? "cluster" : basename;
        this.singleFile = singleFile;
        this.createUniqueSubfolder = createUniqueSubfolder;
    }

    @Override
    protected Collection<EvalResult> runInternal(List<Cluster<NL, EL, G>> testData) throws EvalException {
        String usedPath = path;

        File folder = new File(path);
        if (folder.exists()) {
            if (!folder.isDirectory()) {
                logger.error("path is a non-folder name");
                return Collections.emptyList();
            }
        } else {
            if (!folder.mkdir()) {
                logger.error("non existing folder cannot be created");
                return Collections.emptyList();
            }
        }

        if (createUniqueSubfolder) {
            try {
                usedPath = FsOperations.createUniqueFolder(path, "");
            } catch (IOException e) {
                logger.error("Crating unique folder failed", e);
                return Collections.emptyList();
            }
        }

        if (singleFile) {
            String fullpath = uniqueFileName ? FsOperations.getUniqueFileName(usedPath, basename, "gml", 3)
                    : FilenameUtils.concat(usedPath, basename + FilenameUtils.EXTENSION_SEPARATOR + "gml");
            try (FileWriter fw = new FileWriter(fullpath); BufferedWriter outElem = new BufferedWriter(fw)) {
                GraphIO.writeClustersGML(fw, testData);
            } catch (IOException e) {
                logger.error("Writing clusters failed", e);
            }
        } else {
            // output results
            for (Cluster<NL, EL, G> cluster : testData) {

                String fullPathElem = uniqueFileName ? FsOperations.getUniqueFileName(usedPath,
                        basename + cluster.getId() + "_elem", "gml") : FilenameUtils.concat(usedPath, basename
                        + cluster.getId() + "_elem" + FilenameUtils.EXTENSION_SEPARATOR + "gml");
                
                try (FileWriter fw = new FileWriter(fullPathElem); BufferedWriter outElem = new BufferedWriter(fw)) {
                    for (GraphContainer<NL, EL, G> g : cluster) {
                        GraphIO.writeGMLGraph(outElem, g);
                    }
                } catch (IOException e) {
                    logger.error("Writing cluster elements failed", e);
                }

                String fullPathRep = uniqueFileName ? FsOperations.getUniqueFileName(usedPath,
                        basename + cluster.getId() + "_rep", "gml") : FilenameUtils.concat(usedPath,
                        basename + cluster.getId() + "_rep" + FilenameUtils.EXTENSION_SEPARATOR + "gml");
                
                try (FileWriter fw = new FileWriter(fullPathRep); BufferedWriter outRep = new BufferedWriter(fw);) {
                    for (GraphContainer<NL, EL, G> g : cluster.getRepresentatives()) {
                        GraphIO.writeGMLGraph(outRep, g);
                    }
                } catch (IOException e) {
                    logger.error("Writing Representatives failed", e);
                }
            }
        }

        return Collections.emptyList();
    }
}
