package struclust.eval.modules.clustering;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.apache.commons.io.FilenameUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.base.Preconditions;

import graph.Graph;
import struclust.Cluster;
import struclust.StructuralClustering;
import struclust.eval.EvalException;
import struclust.eval.EvalResult;
import struclust.graph.GraphContainer;
import struclust.util.FsOperations;

/**
 * Saves a clusters id list (separated by newline) in the ordering of a given
 * {@link GraphContainer} list.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class IdListClustSaverEM<NL, EL, G extends Graph<NL, EL>> extends PrecompClustEM<NL, EL, G> {
    private static final Logger logger = LoggerFactory.getLogger(IdListClustSaverEM.class);
    private static final String endl = System.getProperty("line.separator");

    private String path;
    private String basename;
    private boolean uniqueFileName;
    private List<GraphContainer<NL, EL, G>> orderedGraphs;

    /**
     * Constructor
     * 
     * @param clust
     *            The configured clustering algorithm
     * @param orderedGraphs
     *            The cluster ids will be saved in the ordering of this List.
     * @param path
     *            the folder path to save the clustering to.
     * @param basename
     *            the basename of the saved file. if null the basename "cluster"
     *            is used.
     * @param uniqueFileName
     *            if true the filename will be always made unique by adding a
     *            number as suffix.
     */
    public IdListClustSaverEM(StructuralClustering<NL, EL, G> clust, List<GraphContainer<NL, EL, G>> orderedGraphs,
            String path, String basename, boolean uniqueFileName) {
        super(clust);
        this.orderedGraphs = orderedGraphs;
        this.path = path;
        this.basename = basename == null ? "cluster" : basename;
        this.uniqueFileName = uniqueFileName;
    }

    @Override
    protected Collection<EvalResult> runInternal(List<Cluster<NL, EL, G>> clustering) throws EvalException {
        Preconditions.checkArgument(clustering.stream().mapToInt(c -> c.size()).sum() == orderedGraphs.size(),
                "the clustering and the ordered graphs differ in size");

        ArrayList<Integer> idList = new ArrayList<>();
        for (GraphContainer<NL, EL, G> gc : orderedGraphs) {
            Cluster<NL, EL, G> matchedCluster = clustering.stream().filter(c -> c.contains(gc)).findAny()
                    .orElse(null);
            if (matchedCluster == null) {
                throw new IllegalStateException(
                        "graph container from odered list cannot be found in the clustering results");
            }
            idList.add(matchedCluster.getId());
        }
        
        String fullpath = uniqueFileName ? FsOperations.getUniqueFileName(path, basename, "txt", 3)
                : FilenameUtils.concat(path, basename + FilenameUtils.EXTENSION_SEPARATOR + "txt");
        try (FileWriter fw = new FileWriter(fullpath); BufferedWriter outElem = new BufferedWriter(fw)) {
            for (Integer cId : idList) {
                fw.write(cId.toString());
                fw.write(endl);
            }
        } catch (IOException e) {
            logger.error("writing id list file failed", e);
        }
        return Collections.emptyList();
    }

}
