/*
 * Scaffold Hunter
 * Copyright (C) 2012 Till Schäfer
 * See the file README.txt in the root directory of the Scaffold Hunter
 * source tree for details.
 *
 * Scaffold Hunter is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Scaffold Hunter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package struclust.eval.modules.clustering;

import graph.Graph;

import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.Cluster;
import struclust.StructuralClustering;
import struclust.eval.EvalException;
import struclust.eval.EvalResult;
import struclust.eval.modules.AbstractEM;

/**
 * Measures the performance (wall clock time)
 * 
 * @author Till Schäfer
 * 
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class PerfEM<NL, EL, G extends Graph<NL, EL>> extends AbstractEM<List<Cluster<NL, EL, G>>> {
    private static Logger logger = LoggerFactory.getLogger(PerfEM.class);
    private StructuralClustering<NL, EL, G> clust;

    /**
     * Constructor
     * 
     * @param clust
     */
    public PerfEM(StructuralClustering<NL, EL, G> clust) {
        this.clust = clust;
    }

    /**
     * Runs a time measurement
     * 
     * @return the result of the measurement
     */
    @Override
    public Collection<EvalResult> run() {
        logger.info("Running time measurement");

        EvalResult result = new EvalResult("time");
        clust.getConf().addAsMetadata(result);

        /*
         * try to perform a garbage collection before measurement to avoid
         * different conditions for different runs
         */
        Runtime.getRuntime().gc();
        final long startFp = System.nanoTime();
        clust.calcFingerprints();
        final long finishedFp = System.nanoTime();
        long msec = (finishedFp - startFp) / 1000000;
        result.addResult("fingerprint_time", msec);
        logger.info("Finished fingerprint calculation in {} msec", msec);

        final long startClust = System.nanoTime();
        testData = clust.calc();
        final long finishedClust = System.nanoTime();
        msec = (finishedClust - startClust) / 1000000;
        result.addResult("clustering_time", msec);
        logger.info("Finished time measurement in {} msec", msec);

        return Collections.singleton(result);
    }

    @Override
    public boolean canUsePrecomputedTestData() {
        return false;
    }

    @Override
    public Collection<EvalResult> run(List<Cluster<NL, EL, G>> precomputed) throws EvalException {
        throw new UnsupportedOperationException("perf module cannot run a test with precomputed test data");
    }

}
