package struclust.eval.modules.clustering;

import graph.Graph;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.apache.commons.io.FilenameUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.Cluster;
import struclust.StructuralClustering;
import struclust.eval.EvalException;
import struclust.eval.EvalResult;
import struclust.graph.GraphContainer;
import struclust.util.Clusters;
import struclust.util.FsOperations;
import struclust.util.GraphIO;

/**
 * Saves a protein sequence clustering disk.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class ProteinSeqSaverEM<NL, EL, G extends Graph<NL, EL>> extends PrecompClustEM<NL, EL, G> {
    private static final Logger logger = LoggerFactory.getLogger(GmlClustSaverEM.class);

    private String path;
    private boolean createUniqueSubfolder;
    private String basename;
    private boolean uniqueFileName;

    private List<List<GraphContainer<NL, EL, G>>> isomorphicGroups;

    /**
     * Constructor
     * 
     * @param clust
     *            The configured clustering algorithm
     * @param path
     *            the folder path to save the clustering to.
     * @param basename
     *            the basename of the saved file. if null the basename "cluster"
     *            is used.
     */
    public ProteinSeqSaverEM(StructuralClustering<NL, EL, G> clust, String path, String basename) {
        this(clust, path, basename, false, true);
    }

    /**
     * Constructor
     * 
     * @param clust
     *            The configured clustering algorithm
     * @param path
     *            the folder path to save the clustering to.
     * @param basename
     *            the basename of the saved file. if null the basename "cluster"
     *            is used.
     * @param createUniqueSubfolder
     *            if true a unique subfolder will be created for each call of
     *            {@link #run()}
     * @param uniqueFileName
     *            if true the filename will be always made unique by adding a
     *            number as suffix.
     */
    public ProteinSeqSaverEM(StructuralClustering<NL, EL, G> clust, String path, String basename,
            boolean createUniqueSubfolder, boolean uniqueFileName) {
        this(clust, path, basename, createUniqueSubfolder, uniqueFileName, null);
    }

    /**
     * Constructor
     * 
     * @param clust
     *            The configured clustering algorithm
     * @param path
     *            the folder path to save the clustering to.
     * @param basename
     *            the basename of the saved file. if null the basename "cluster"
     *            is used.
     * @param createUniqueSubfolder
     *            if true a unique subfolder will be created for each call of
     *            {@link #run()}
     * @param uniqueFileName
     *            if true the filename will be always made unique by adding a
     *            number as suffix.
     * @param isomorphicGroups
     *            if isomorphic graphs where filtered for the clustering
     *            process, the isomorphic groups can be given here, to re-add
     *            the isomorphic graphs to the clustering result. I.e. each
     *            individual member in a cluster is expanded by its complete
     *            isomorphic group. Each testData cluster member must be
     *            contained in exactly one group.
     */
    public ProteinSeqSaverEM(StructuralClustering<NL, EL, G> clust, String path, String basename,
            boolean createUniqueSubfolder, boolean uniqueFileName,
            List<List<GraphContainer<NL, EL, G>>> isomorphicGroups) {
        super(clust);
        this.path = path;
        this.uniqueFileName = uniqueFileName;
        this.basename = basename == null ? "cluster" : basename;
        this.createUniqueSubfolder = createUniqueSubfolder;
        this.isomorphicGroups = isomorphicGroups;
    }

    @Override
    protected Collection<EvalResult> runInternal(List<Cluster<NL, EL, G>> testData) throws EvalException {
        String usedPath = path;
        List<Cluster<NL, EL, G>> expandedClustering = null;

        File folder = new File(path);
        if (folder.exists()) {
            if (!folder.isDirectory()) {
                logger.error("path is a non-folder name");
                return Collections.emptyList();
            }
        } else {
            if (!folder.mkdir()) {
                logger.error("non existing folder cannot be created");
                return Collections.emptyList();
            }
        }

        if (createUniqueSubfolder) {
            try {
                usedPath = FsOperations.createUniqueFolder(path, "");
            } catch (IOException e) {
                logger.error("Crating unique folder failed", e);
                return Collections.emptyList();
            }
        }

        boolean isomorphicExpansion = isomorphicGroups != null;
        if (isomorphicExpansion) {
            expandedClustering = Clusters.cloneClusteringWithoutMembers(testData);
            for (int i = 0; i < testData.size(); i++) {
                Cluster<NL, EL, G> cluster = testData.get(i);
                Cluster<NL, EL, G> expandedCluster = expandedClustering.get(i);
                for (GraphContainer<NL, EL, G> gc : cluster) {
                    // PARALLEL stream: finding isomorphic group
                    List<GraphContainer<NL, EL, G>> groupOfGc = isomorphicGroups.stream().parallel()
                            .filter(l -> l.contains(gc)).findAny()
                            .orElseThrow(() -> new IllegalArgumentException("no isomorphic group found"));
                    expandedCluster.addAll(groupOfGc);
                }
            }
        }

        String fullpath = uniqueFileName ? FsOperations.getUniqueFileName(usedPath, basename, "txt", 3) : FilenameUtils
                .concat(usedPath, basename + FilenameUtils.EXTENSION_SEPARATOR + "txt");
        try (FileWriter fw = new FileWriter(fullpath); BufferedWriter outElem = new BufferedWriter(fw)) {
            GraphIO.writeClustersProteinSeq(fw, isomorphicExpansion ? expandedClustering : testData);
        } catch (IOException e) {
            logger.error("Writing clusters failed", e);
        }

        return Collections.emptyList();
    }
}
