package struclust.eval.modules.clustering;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.commons.io.FilenameUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import graph.DefaultGraph;
import graph.GraphSize;
import struclust.Cluster;
import struclust.eval.EvalException;
import struclust.eval.EvalResult;
import struclust.eval.modules.AbstractEM;
import struclust.graph.GraphContainer;
import struclust.util.Clusters;
import struclust.util.GraphIO;

/**
 * This is actually a fake module, that does not run any new tests, but reads
 * existing clusterings (in a custom clusterGML format) from disk and uses them
 * as test data. Together with the series module, this allows to apply other
 * evaluation modules on the stored test data.
 * 
 * @author Till Schäfer
 */
public class ReadClusteringGmlEM extends AbstractEM<List<Cluster<String, String, DefaultGraph<String, String>>>> {
    private static Logger logger = LoggerFactory.getLogger(ReadClusteringGmlEM.class);

    private String folderPath;
    private String fileNamePattern;
    private int run;
    private List<GraphContainer<String, String, DefaultGraph<String, String>>> graphs;
    private GraphSize gSize;

    /**
     * Constructor
     * 
     * @param folderPath
     *            the folder path where the gml files are placed
     * @param fileNamePattern
     *            the pattern of the file name for String.format. The run is
     *            passed as integer parameter to the format function.<br>
     *            E.g.: The string ""%03d.gml" would generate file names
     *            001.gml, 002.gml, etc.
     * @param startRun
     *            the initial run value. It is increased with each call of
     *            run();
     * @param graphs
     *            if not null, the read gml graphs are mapped to this graphs
     *            first and the test data will contain the gml cluster
     *            structure, but the graphs graphs from this parameter.
     * @param gSize
     *            the {@link GraphSize} for the clusters
     */
    public ReadClusteringGmlEM(String folderPath, String fileNamePattern, int startRun,
            List<GraphContainer<String, String, DefaultGraph<String, String>>> graphs, GraphSize gSize) {
        this.folderPath = folderPath;
        this.fileNamePattern = fileNamePattern;
        this.run = startRun;
        this.graphs = graphs;
        this.gSize = gSize;
    }

    @Override
    public Collection<EvalResult> run() throws EvalException {
        String fileName = String.format(fileNamePattern, run);
        String path = FilenameUtils.concat(folderPath, fileName);
        try {
            ArrayList<Cluster<String, String, DefaultGraph<String, String>>> gmlClusters = GraphIO
                    .clustersGmlToDefaultGraphs(path, gSize);

            logger.debug("read {} graphs from file {}", gmlClusters.size(), path);
            if (graphs != null) {

                List<Cluster<String, String, DefaultGraph<String, String>>> mappedClusters = Clusters
                        .replaceMembersWithIsomorphicGcs(gmlClusters, graphs);

                assert assertSameElements(graphs, mappedClusters);
                testData = mappedClusters;
            } else {
                testData = gmlClusters;
            }
        } catch (IOException e) {
            throw new EvalException("Reading graphs from gml file '" + path + "' failed", e);
        }
        run++;
        return Collections.emptyList();
    }

    @Override
    public Collection<EvalResult> run(List<Cluster<String, String, DefaultGraph<String, String>>> precomputed)
            throws EvalException {
        throw new UnsupportedOperationException("ReadClusteringGmlEM cannot run a test with precomputed test data");
    }

    @Override
    public boolean canUsePrecomputedTestData() {
        return false;
    }

    private boolean assertSameElements(List<GraphContainer<String, String, DefaultGraph<String, String>>> graphs1,
            Collection<Cluster<String, String, DefaultGraph<String, String>>> clusters2) {
        Set<GraphContainer<String, String, DefaultGraph<String, String>>> set1 = new HashSet<>(graphs1);
        Set<GraphContainer<String, String, DefaultGraph<String, String>>> set2 = clusters2.stream()
                .flatMap(c -> c.stream()).collect(Collectors.toSet());
        return set1.equals(set2);
    }

}
