package struclust.eval.modules.clustering;

import graph.Graph;

import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.commons.math3.stat.descriptive.rank.Median;

import struclust.Cluster;
import struclust.StructuralClustering;
import struclust.eval.EvalException;
import struclust.eval.EvalResult;

/**
 * Abstract class to generate one {@link EvalResult} containing statistics
 * (minimum, maximum, average, median) and another EvalResult with all values.
 * 
 * It is possible to specify a flatMap function from the cluster to another type
 * T and a function to map each T to a double value.
 * 
 * @author Till Schäfer
 * 
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 * @param <T>
 *            the type of the elements over which the values should be computed
 */
public abstract class StatisticsAndValuesEM<NL, EL, G extends Graph<NL, EL>, T> extends PrecompClustEM<NL, EL, G> {

    private String measurement;

    /**
     * Constructor
     * 
     * @param clust
     *            the configured clustering algorithm
     * @param measurement
     *            the measurement name
     */
    public StatisticsAndValuesEM(StructuralClustering<NL, EL, G> clust, String measurement) {
        super(clust);
        this.measurement = measurement;
    }

    @Override
    protected Collection<EvalResult> runInternal(List<Cluster<NL, EL, G>> testData) throws EvalException {
        EvalResult stats = new EvalResult(measurement + "_statistics");
        EvalResult values = new EvalResult(measurement + "_values");

        stats.addRef(values);
        values.addRef(stats);

        clust.getConf().addAsMetadata(stats);
        clust.getConf().addAsMetadata(values);

        stats.addResult("min",
                testData.stream().flatMap(mapper()).mapToDouble(x -> getValue(x)).min().orElse(Double.NaN));
        stats.addResult("max",
                testData.stream().flatMap(mapper()).mapToDouble(x -> getValue(x)).max().orElse(Double.NaN));
        stats.addResult("avg",
                testData.stream().flatMap(mapper()).mapToDouble(x -> getValue(x)).average().orElse(Double.NaN));
        double[] valuesArr = (testData.stream().flatMap(mapper()).mapToDouble(x -> getValue(x))).toArray();
        stats.addResult("med", valuesArr.length == 0 ? Double.NaN : new Median().evaluate(valuesArr));

        for (Cluster<NL, EL, G> cluster : testData) {
            List<? extends T> collectedT = mapper().apply(cluster).collect(Collectors.toList());
            if (collectedT.size() == 1) {
                values.addResult(Integer.toString(cluster.getId()), getValue(collectedT.get(0)));
            } else {
                int i = 0;
                for (T t : collectedT) {
                    values.addResult(Integer.toString(cluster.getId()) + "_" + ++i, getValue(t));
                }
            }
        }

        return new LinkedList<>(Arrays.asList(new EvalResult[] { stats, values }));
    }

    protected abstract Function<? super Cluster<NL, EL, G>, ? extends Stream<? extends T>> mapper();

    protected abstract double getValue(T t);
}
