package struclust.eval.modules.clustering;

import graph.Graph;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import struclust.Cluster;
import struclust.SharedMemorySC;
import struclust.comparison.AccFunction;
import struclust.comparison.AccRepCCD;
import struclust.comparison.ClusterClusterDistance;
import struclust.comparison.ClusterGraphDistance;
import struclust.eval.EvalException;
import struclust.eval.EvalResult;
import struclust.graph.GraphContainer;

import com.google.common.base.Preconditions;

/**
 * Implements an internal clustering validation index called Xie-Beni-Index (XB)
 * 
 * Reference: Xuanli Lisa Xie and Gerardo Beni. A validity measure for fuzzy
 * clustering. IEEE Transactions on Pattern Analysis and Machine Intelligence,
 * 13(8):841-847,1991
 * 
 * @author Till Schäfer
 * 
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class XieBeniIndexEM<NL, EL, G extends Graph<NL, EL>> extends PrecompClustEM<NL, EL, G> {

    private ClusterGraphDistance<NL, EL, G> cgd;
    private ClusterClusterDistance<NL, EL, G> ccd;

    /**
     * Constructor. For compactness calculation, the cluster to graph distance
     * from the clustering is used. For separation calculation the
     * {@link AccRepCCD} is used with the graph to graph distance from the above
     * mentioned cluster to graph distance and the averaging accumulation
     * function.
     * 
     * @param clust
     *            The configured clustering algorithm
     */
    public XieBeniIndexEM(SharedMemorySC<NL, EL, G> clust) {
        this(clust, clust.getConf().cgDist, new AccRepCCD<NL, EL, G>(clust.getConf().cgDist.getGGDist(),
                AccFunction.AVG));
    }

    /**
     * Constructor
     * 
     * @param clust
     *            The configured clustering algorithm
     * @param cgd
     *            cluster to graph distance for cluster width calculation.
     *            distance must have a maximal value.
     * @param ccd
     *            cluster to cluster distance for separation calculation
     */
    public XieBeniIndexEM(SharedMemorySC<NL, EL, G> clust, ClusterGraphDistance<NL, EL, G> cgd,
            ClusterClusterDistance<NL, EL, G> ccd) {
        super(clust);

        Preconditions.checkNotNull(cgd);
        Preconditions.checkNotNull(ccd);
        Preconditions.checkArgument(cgd.getGGDist() != null, "This Constructor must be used in combination "
                + "with a clustering that was configured to use a accmulating ClusterGraphDistance."
                + "Otherwise the GraphGraphDistance must be specified manually");
        Preconditions.checkArgument(!Double.isNaN(cgd.maxDist()));

        this.cgd = cgd;
        this.ccd = ccd;
    }

    @Override
    protected Collection<EvalResult> runInternal(List<Cluster<NL, EL, G>> testData) throws EvalException {
        this.testData = testData;
        EvalResult retVal = new EvalResult("XieBeniIndex");

        int n = 0;
        for (Cluster<NL, EL, G> c : testData) {
            n += c.size();
        }

        // intra cluster width (compactness critera)
        BigDecimal comp = new BigDecimal(0);
        for (Cluster<NL, EL, G> c : testData) {
            if (!c.getRepresentatives().isEmpty()) {
                for (GraphContainer<NL, EL, G> gc : c) {
                    double powDist = Math.pow(cgd.calc(c, gc), 2);
                    comp = comp.add(new BigDecimal(powDist));
                }
            } else {
                /*
                 * clusters without representatives always have maxDist to their
                 * members
                 */
                comp = comp.add(new BigDecimal(Math.pow(cgd.maxDist(), 2) * c.size()));
            }
        }

        // minimal intra cluster distance
        double minCCDist = Double.POSITIVE_INFINITY;
        for (Cluster<NL, EL, G> c1 : testData) {
            if (c1.getRepresentatives().isEmpty())
                continue;
            for (Cluster<NL, EL, G> c2 : testData) {
                if (c1 == c2)
                    break;
                if (c2.getRepresentatives().isEmpty())
                    continue;

                double dist = ccd.calc(c1, c2);
                minCCDist = Math.min(minCCDist, dist);
            }
        }

        // intra cluster separation criteria
        BigDecimal sep = new BigDecimal(minCCDist);
        sep = sep.multiply(sep);
        sep = sep.multiply(new BigDecimal(n));

        double score = sep.intValue() != 0 ? comp.divide(sep, RoundingMode.HALF_UP).doubleValue()
                : Double.POSITIVE_INFINITY;

        clust.getConf().addAsMetadata(retVal);
        retVal.addMetadata("XieBeniIndexEM cluster-graph distance", cgd.getDescription());
        retVal.addMetadata("XieBeniIndexEM cluster-cluster distance", ccd.getDescription());

        retVal.addResult("score", score);

        return Collections.singleton(retVal);
    }

}
