package struclust.eval.modules.comparison;

import graph.Graph;

import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.comparison.GraphGraphDistance;
import struclust.eval.EvalException;
import struclust.eval.EvalResult;
import struclust.eval.modules.AbstractEM;
import struclust.graph.GraphContainer;

/**
 * Performance evaluation for {@link GraphGraphDistance}s. Calculates the
 * pairwise distance between all given graphs and measures the calculation time.
 * 
 * @author Till Schäfer
 * 
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class GGDPerfEM<NL, EL, G extends Graph<NL, EL>> extends AbstractEM<double[][]> {
    private static final Logger logger = LoggerFactory.getLogger(GGDPerfEM.class);

    private GraphGraphDistance<NL, EL, G> dist;
    private List<GraphContainer<NL, EL, G>> graphs;

    /**
     * Constructor
     * 
     * @param dist
     *            distance measure to measure
     * @param graphs
     *            graphs to test the distance measure on
     */
    public GGDPerfEM(GraphGraphDistance<NL, EL, G> dist, List<GraphContainer<NL, EL, G>> graphs) {
        this.dist = dist;
        this.graphs = graphs;
    }

    @Override
    public Collection<EvalResult> run() throws EvalException {
        logger.info("Running distance pairwise time measurement");

        EvalResult result = new EvalResult("distance pairwise time");
        result.addMetadata("distance", dist.getDescription());
        result.addMetadata("number of graphs", graphs.size());

        /*
         * try to perform a garbage collection before measurement to avoid
         * different conditions for different runs
         */
        Runtime.getRuntime().gc();
        Date start = new Date();
        calcDistances();
        Date finish = new Date();
        long msec = finish.getTime() - start.getTime();
        result.addResult("time", msec);
        logger.info("Finished distance pairwise time measurement in {} msec", msec);

        return Collections.singleton(result);
    }

    private void calcDistances() {
        testData = new double[graphs.size()][graphs.size()];
        Iterator<GraphContainer<NL, EL, G>> it1 = graphs.iterator();
        for (int i = 0; i < graphs.size(); i++) {
            GraphContainer<NL, EL, G> gc1 = it1.next();
            
            Iterator<GraphContainer<NL, EL, G>> it2 = graphs.iterator();
            for (int j = 0; j < graphs.size(); j++) {
                GraphContainer<NL, EL, G> gc2 = it2.next();
                testData[j][i] = dist.calc(gc1, gc2);
            }
        }
    }

    @Override
    public Collection<EvalResult> run(double[][] testData) throws EvalException {
        throw new UnsupportedOperationException("perf module cannot run a test with precomputed test data");
    }

    @Override
    public boolean canUsePrecomputedTestData() {
        return false;
    }

}