package struclust.eval.modules.comparison;

import graph.Graph;

import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.comparison.GraphGraphDistance;
import struclust.eval.EvalException;
import struclust.eval.EvalResult;
import struclust.eval.modules.AbstractEM;
import struclust.eval.modules.EvalModule;
import struclust.graph.GraphContainer;

import com.google.common.util.concurrent.ThreadFactoryBuilder;

/**
 * Abstract class for {@link EvalModule}s, that do support precomputed test
 * data. If inherited from this class, only the Method
 * {@link PrecompGGDMatrixEM#runInternal(List)} and a constructor must be
 * implemented.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
@SuppressWarnings("javadoc")
public abstract class PrecompGGDMatrixEM<NL, EL, G extends Graph<NL, EL>> extends AbstractEM<double[][]> {
    private static final Logger logger = LoggerFactory.getLogger(PrecompGGDMatrixEM.class);

    protected GraphGraphDistance<NL, EL, G> dist;
    protected List<GraphContainer<NL, EL, G>> graphs;

    /**
     * Constructor
     * 
     * @param dist
     *            distance measure to measure
     * @param graphs
     *            graphs to test the distance measure on
     */
    public PrecompGGDMatrixEM(GraphGraphDistance<NL, EL, G> dist, List<GraphContainer<NL, EL, G>> graphs) {
        this.dist = dist;
        this.graphs = graphs;
    }

    @Override
    public Collection<EvalResult> run() throws EvalException {
        double[][] matrix = calcDistances();

        return run(matrix);
    }

    @Override
    public Collection<EvalResult> run(double[][] testData) throws EvalException {
        this.testData = testData;

        return runInternal(testData);
    }

    protected abstract Collection<EvalResult> runInternal(double[][] testData) throws EvalException;

    @Override
    public boolean canUsePrecomputedTestData() {
        return true;
    }

    protected double[][] calcDistances() {
        ExecutorService initializer = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors(),
                new ThreadFactoryBuilder().setNameFormat("PrecompGGDMatrixEM-DistComputation-%d").setDaemon(true)
                        .build());
        double[][] matrix = new double[graphs.size()][graphs.size()];
        Iterator<GraphContainer<NL, EL, G>> it1 = graphs.iterator();
        for (int i = 0; i < graphs.size(); i++) {
            GraphContainer<NL, EL, G> gc1 = it1.next();

            initializer.submit(new SingleRowDistInit(matrix, i, gc1));
        }

        initializer.shutdown();
        try {
            initializer.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
        } catch (InterruptedException e) {
            logger.warn("value computation aborted", e);
        }
        return matrix;
    }

    private class SingleRowDistInit implements Runnable {

        private double[][] matrix;
        private int i;
        private GraphContainer<NL, EL, G> gc1;

        public SingleRowDistInit(double[][] matrix, int i, GraphContainer<NL, EL, G> gc1) {
            this.matrix = matrix;
            this.i = i;
            this.gc1 = gc1;
        }

        @Override
        public void run() {
            Iterator<GraphContainer<NL, EL, G>> it2 = graphs.iterator();
            for (int j = 0; j < graphs.size(); j++) {
                GraphContainer<NL, EL, G> gc2 = it2.next();
                matrix[j][i] = dist.calc(gc1, gc2);
            }
        }

    }

    protected double[] values(double[][] testData) {
        int matrixSize = (int) Math.pow(testData.length, 2);
        double[] values = new double[matrixSize];
        int i = 0;
        for (double[] row : testData) {
            for (double val : row) {
                values[i] = val;
                i++;
            }
        }
        return values;
    }
}
