package struclust.eval.testsetups;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;
import java.util.stream.Collectors;

import org.apache.commons.io.FilenameUtils;

import com.google.common.collect.Sets;

import cern.jet.random.AbstractDistribution;
import cern.jet.random.Exponential;
import cern.jet.random.Poisson;
import cern.jet.random.engine.MersenneTwister;
import graph.DefaultGraph;
import graph.DefaultGraphFactory;
import graph.Graph;
import graph.GraphFactory;
import graph.GraphSize;
import graph.MoleculeGraph;
import graph.MoleculeGraphFactory;
import hsahn.comparison.kernel.graph.GraphletKernel;
import index.features.FingerprintBuilder;
import struclust.Cluster;
import struclust.ClusterFactory;
import struclust.FakeSC;
import struclust.SharedMemorySC;
import struclust.SharedMemorySCC;
import struclust.StructuralClustering;
import struclust.comparison.AccCGD;
import struclust.comparison.AccFunction;
import struclust.comparison.GraphGraphDistance;
import struclust.comparison.McsMinSizeGGD;
import struclust.comparison.McsWskrGGD;
import struclust.comparison.WeightedSubIsoFpGGD;
import struclust.competingalgorithms.opensubspace.ExplicitKernelTFVF;
import struclust.competingalgorithms.opensubspace.ProclusConf;
import struclust.competingalgorithms.opensubspace.WrappedProclus;
import struclust.competingalgorithms.trickl.ExplicitKernelKMeansConf;
import struclust.competingalgorithms.trickl.WrappedExplicitKernelKMeans;
import struclust.configurations.DefaultPaths;
import struclust.convergence.ConvergenceCriteria;
import struclust.convergence.FixedIterationNumberCC;
import struclust.convergence.GenericRangeAvgCC;
import struclust.distributions.Dirac;
import struclust.eval.EvalResult;
import struclust.eval.comparison.FowlkesMallowsCC;
import struclust.eval.comparison.NormVariationOfInformationCC;
import struclust.eval.comparison.OverlappingPurityCC;
import struclust.eval.comparison.PurityCC;
import struclust.eval.metamodules.EvalMetaModule;
import struclust.eval.metamodules.FileSaverEMM;
import struclust.eval.metamodules.MetaDataAdderEMM;
import struclust.eval.metamodules.RepeaterEMM;
import struclust.eval.metamodules.StatisticsEMM;
import struclust.eval.modules.EvalModule;
import struclust.eval.modules.SeriesEM;
import struclust.eval.modules.clustering.ClustSizeEM;
import struclust.eval.modules.clustering.ClusteringComparisonEM;
import struclust.eval.modules.clustering.ConvergenceEM;
import struclust.eval.modules.clustering.CoverageEM;
import struclust.eval.modules.clustering.GmlClustSaverEM;
import struclust.eval.modules.clustering.InjectionSaverEM;
import struclust.eval.modules.clustering.NumClustersEM;
import struclust.eval.modules.clustering.PerfEM;
import struclust.eval.modules.clustering.RTMSToNumClustersEM;
import struclust.eval.modules.clustering.RepSizeEM;
import struclust.eval.modules.clustering.RepToMemberSizeEM;
import struclust.eval.modules.clustering.UsePredefinedClusteringEM;
import struclust.granularity.ReMineMS;
import struclust.granularity.RePreClusterSS;
import struclust.granularity.RelRepToMemberSizeSC;
import struclust.granularity.RepThresholdCountFineParallelMC;
import struclust.graph.GraphContainer;
import struclust.graph.generators.ConnectedRGG;
import struclust.graph.generators.SeedCombinationCGG;
import struclust.mcs.MccisKriegeColor2MSA;
import struclust.mcs.MccisKriegeColor2MSDA;
import struclust.mining.ConservativeSCE;
import struclust.mining.CorrectedBinomialSC;
import struclust.mining.RandomGM;
import struclust.mining.RandomGM.RandomGMConf;
import struclust.mining.RelRepToMemberSizeMSS;
import struclust.mining.SimpleSC;
import struclust.preclustering.DissimilarRepPC;
import struclust.repselection.DiscriminativeSupSizeRSS;
import struclust.util.Clusters;
import struclust.util.FsOperations;
import struclust.util.GraphIO;
import struclust.util.Ratings;

/**
 * Test setups for the publication ADMA 2017
 * 
 * @author Till Schäfer
 */
public enum PublicationADMA2017TS implements TestSetup {
    /**
     * Clustering evaluation that generates synthetic graphs from seed patterns.
     * 
     * Quality and performance is measured over a series of different dataset
     * sizes. As a ground truth is given by the generation process, quality can
     * be measured by comparing the clustering results with the ground truth.
     */
    ReproduceMultipleSeeds {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int parallelism = 16;
            int repeats = 30;

            int vertexLabelCount = 10;
            int edgeLabelCount = 3;
            AbstractDistribution labelDistribution = new Exponential(1.0, new MersenneTwister());
            AbstractDistribution seedSizeDistribution = new Poisson(10, new MersenneTwister());
            double seedCycleProbability = 0.1;
            int numClusters = 100;
            AbstractDistribution clusterWeightDistribution = new Dirac(1);
            int clusterSeedPoolSize = 3;
            int commonSeedPoolSize = 10;
            AbstractDistribution usedCommonSeedsDistribution = new Exponential(3, new MersenneTwister());
            AbstractDistribution connectionCountDistribution = new Poisson(5, new MersenneTwister());

            AbstractDistribution noiseSizeDistribution = new Poisson(50, new MersenneTwister());
            double noiseRatio = 0.05;

            SeedCombinationCGG graphGen = new SeedCombinationCGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    seedSizeDistribution, seedCycleProbability, numClusters, clusterWeightDistribution,
                    clusterSeedPoolSize, commonSeedPoolSize, usedCommonSeedsDistribution, connectionCountDistribution);
            ConnectedRGG noiseGen = new ConnectedRGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    noiseSizeDistribution, seedCycleProbability);

            int[] datasetSizes = { 1000, 5000, 10000, 50000, 100000, 500000, 1000000 };

            for (int numGraphs : datasetSizes) {

                /*
                 * generate and filter graphs
                 */
                ArrayList<List<DefaultGraph<Integer, Integer>>> groundTruth = graphGen
                        .createGraphsClustered((int) (numGraphs * (1 - noiseRatio) + 0.5));
                groundTruth.add(noiseGen.createGraphs((int) (numGraphs * noiseRatio + 0.5)));

                ArrayList<GraphContainer<Integer, Integer, DefaultGraph<Integer, Integer>>> graphContainers = TSHelpers
                        .groundTruth2filteredGCs(groundTruth);
                TSHelpers.logDatasetSummary(true, graphContainers);

                /*
                 * clustering configuration
                 */
                SharedMemorySCC<Integer, Integer, DefaultGraph<Integer, Integer>> conf = conf(
                        new DefaultGraphFactory<>(), parallelism);
                SharedMemorySC<Integer, Integer, DefaultGraph<Integer, Integer>> clust = new SharedMemorySC<>(
                        graphContainers, conf);

                ArrayList<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>> gtClusters = Clusters
                        .toClusterList(groundTruth, new ClusterFactory<>(conf));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                conf.evalModuleInjection = false;

                ArrayList<EvalModule<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new CoverageEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new RepSizeEM<>(clust));
                modules.add(new RepToMemberSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));

                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
                modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                        FilenameUtils.getBaseName(folderPath), true, false, true));

                SeriesEM<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>> series = new SeriesEM<>(
                        modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(new RepeaterEMM(repeats, false));
                MetaDataAdderEMM generatorDescription = new MetaDataAdderEMM("graph generator",
                        graphGen.getDescription());
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(generatorDescription);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, generatorDescription,
                        datasetSize);

                TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);

                results.addAll(saver.run(mm, series));

                TSHelpers.saveInjectionResults(conf, clust, folderPath);

            }
            return results;
        }
    },
    /**
     * Clustering evaluation that generates synthetic graphs from seed patterns.
     * 
     * Quality and performance is measured over a series of different dataset
     * sizes. As a ground truth is given by the generation process, quality can
     * be measured by comparing the clustering results with the ground truth.
     */
    ReproduceMultipleSeedsConvergence {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int parallelism = 16;
            int repeats = 30;

            int vertexLabelCount = 10;
            int edgeLabelCount = 3;
            AbstractDistribution labelDistribution = new Exponential(1.0, new MersenneTwister());
            AbstractDistribution seedSizeDistribution = new Poisson(10, new MersenneTwister());
            double seedCycleProbability = 0.1;
            int numClusters = 100;
            AbstractDistribution clusterWeightDistribution = new Dirac(1);
            int clusterSeedPoolSize = 3;
            int commonSeedPoolSize = 10;
            AbstractDistribution usedCommonSeedsDistribution = new Exponential(3, new MersenneTwister());
            AbstractDistribution connectionCountDistribution = new Poisson(5, new MersenneTwister());

            AbstractDistribution noiseSizeDistribution = new Poisson(50, new MersenneTwister());
            double noiseRatio = 0.05;

            SeedCombinationCGG graphGen = new SeedCombinationCGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    seedSizeDistribution, seedCycleProbability, numClusters, clusterWeightDistribution,
                    clusterSeedPoolSize, commonSeedPoolSize, usedCommonSeedsDistribution, connectionCountDistribution);
            ConnectedRGG noiseGen = new ConnectedRGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    noiseSizeDistribution, seedCycleProbability);

            int[] datasetSizes = { 10000 };

            for (int numGraphs : datasetSizes) {

                /*
                 * generate and filter graphs
                 */
                ArrayList<List<DefaultGraph<Integer, Integer>>> groundTruth = graphGen
                        .createGraphsClustered((int) (numGraphs * (1 - noiseRatio) + 0.5));
                groundTruth.add(noiseGen.createGraphs((int) (numGraphs * noiseRatio + 0.5)));

                ArrayList<GraphContainer<Integer, Integer, DefaultGraph<Integer, Integer>>> graphContainers = TSHelpers
                        .groundTruth2filteredGCs(groundTruth);
                TSHelpers.logDatasetSummary(true, graphContainers);

                /*
                 * clustering configuration
                 */
                SharedMemorySCC<Integer, Integer, DefaultGraph<Integer, Integer>> conf = conf(
                        new DefaultGraphFactory<>(), parallelism);
                ConvergenceCriteria origConvCrit = conf.convergence;
                conf.convergence = new FixedIterationNumberCC(50);

                SharedMemorySC<Integer, Integer, DefaultGraph<Integer, Integer>> clust = new SharedMemorySC<>(
                        graphContainers, conf);

                ArrayList<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>> gtClusters = Clusters
                        .toClusterList(groundTruth, new ClusterFactory<>(conf));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                conf.evalModuleInjection = true;
                conf.injectionModules.add(new CoverageEM<>(clust));
                conf.injectionModules.add(new ClustSizeEM<>(clust));
                conf.injectionModules.add(new RepSizeEM<>(clust));
                conf.injectionModules.add(new RepToMemberSizeEM<>(clust));
                conf.injectionModules.add(new NumClustersEM<>(clust));
                conf.injectionModules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                conf.injectionModules
                        .add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                conf.injectionModules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                conf.injectionModules
                        .add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
                conf.injectionModules.add(new ConvergenceEM<>(clust, origConvCrit, true));

                ArrayList<EvalModule<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new CoverageEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new RepSizeEM<>(clust));
                modules.add(new RepToMemberSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));

                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
                modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                        FilenameUtils.getBaseName(folderPath), true, false, true));
                modules.add(new InjectionSaverEM<>(clust, folderPath));

                SeriesEM<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>> series = new SeriesEM<>(
                        modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(new RepeaterEMM(repeats, false));
                MetaDataAdderEMM generatorDescription = new MetaDataAdderEMM("graph generator",
                        graphGen.getDescription());
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(generatorDescription);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, generatorDescription,
                        datasetSize);

                TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);

                results.addAll(saver.run(mm, series));

            }
            return results;
        }
    },
    /**
     * same as {@link PublicationADMA2017TS#ReproduceMultipleSeeds} but with
     * {@link SimpleSC} (exact Support Counting)
     */
    ReproduceMultipleSeedsExactSupportCounting {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int parallelism = 16;
            int repeats = 30;

            int vertexLabelCount = 10;
            int edgeLabelCount = 3;
            AbstractDistribution labelDistribution = new Exponential(1.0, new MersenneTwister());
            AbstractDistribution seedSizeDistribution = new Poisson(10, new MersenneTwister());
            double seedCycleProbability = 0.1;
            int numClusters = 100;
            AbstractDistribution clusterWeightDistribution = new Dirac(1);
            int clusterSeedPoolSize = 3;
            int commonSeedPoolSize = 10;
            AbstractDistribution usedCommonSeedsDistribution = new Exponential(3, new MersenneTwister());
            AbstractDistribution connectionCountDistribution = new Poisson(5, new MersenneTwister());

            AbstractDistribution noiseSizeDistribution = new Poisson(50, new MersenneTwister());
            double noiseRatio = 0.05;

            SeedCombinationCGG graphGen = new SeedCombinationCGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    seedSizeDistribution, seedCycleProbability, numClusters, clusterWeightDistribution,
                    clusterSeedPoolSize, commonSeedPoolSize, usedCommonSeedsDistribution, connectionCountDistribution);
            ConnectedRGG noiseGen = new ConnectedRGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    noiseSizeDistribution, seedCycleProbability);

            int[] datasetSizes = { 1000, 5000, 10000, 50000, 100000 };

            for (int numGraphs : datasetSizes) {

                /*
                 * generate and filter graphs
                 */
                ArrayList<List<DefaultGraph<Integer, Integer>>> groundTruth = graphGen
                        .createGraphsClustered((int) (numGraphs * (1 - noiseRatio) + 0.5));
                groundTruth.add(noiseGen.createGraphs((int) (numGraphs * noiseRatio + 0.5)));

                ArrayList<GraphContainer<Integer, Integer, DefaultGraph<Integer, Integer>>> graphContainers = TSHelpers
                        .groundTruth2filteredGCs(groundTruth);
                TSHelpers.logDatasetSummary(true, graphContainers);

                /*
                 * clustering configuration
                 */
                SharedMemorySCC<Integer, Integer, DefaultGraph<Integer, Integer>> conf = confExactSupportCounting(
                        new DefaultGraphFactory<>(), parallelism);
                SharedMemorySC<Integer, Integer, DefaultGraph<Integer, Integer>> clust = new SharedMemorySC<>(
                        graphContainers, conf);

                ArrayList<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>> gtClusters = Clusters
                        .toClusterList(groundTruth, new ClusterFactory<>(conf));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                conf.evalModuleInjection = false;

                ArrayList<EvalModule<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new CoverageEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new RepSizeEM<>(clust));
                modules.add(new RepToMemberSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
                modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                        FilenameUtils.getBaseName(folderPath), true, false, true));

                SeriesEM<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>> series = new SeriesEM<>(
                        modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(new RepeaterEMM(repeats, false));
                MetaDataAdderEMM generatorDescription = new MetaDataAdderEMM("graph generator",
                        graphGen.getDescription());
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(generatorDescription);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, generatorDescription,
                        datasetSize);

                TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);

                results.addAll(saver.run(mm, series));

                TSHelpers.saveInjectionResults(conf, clust, folderPath);

            }
            return results;
        }
    },
    /**
     * Clustering Evaluation that generates synthetic graphs from seed patterns.
     * 
     * This Test only considers core scaling, in the sense of parallelism.
     * Therefore, a fixed number of iterations is used regardless of the number
     * of iterations.Additionally one should re-use a single dataset over
     * multiple runs.
     * 
     * Note: TurboBoost should be disabled on the testing machine
     */
    ReproduceMultipleSeedsCoreScaling {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int parallelism = 0;
            int repeats = 30;
            boolean generate = false; // otherwise load
            String loadRun = "001";

            int vertexLabelCount = 10;
            int edgeLabelCount = 3;
            AbstractDistribution labelDistribution = new Exponential(1.0, new MersenneTwister());
            AbstractDistribution seedSizeDistribution = new Poisson(10, new MersenneTwister());
            double seedCycleProbability = 0.1;
            int numClusters = 100;
            AbstractDistribution clusterWeightDistribution = new Dirac(1);
            int clusterSeedPoolSize = 3;
            int commonSeedPoolSize = 10;
            AbstractDistribution usedCommonSeedsDistribution = new Exponential(3, new MersenneTwister());
            AbstractDistribution connectionCountDistribution = new Poisson(5, new MersenneTwister());

            AbstractDistribution noiseSizeDistribution = new Poisson(50, new MersenneTwister());
            double noiseRatio = 0.05;

            SeedCombinationCGG graphGen = new SeedCombinationCGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    seedSizeDistribution, seedCycleProbability, numClusters, clusterWeightDistribution,
                    clusterSeedPoolSize, commonSeedPoolSize, usedCommonSeedsDistribution, connectionCountDistribution);
            ConnectedRGG noiseGen = new ConnectedRGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    noiseSizeDistribution, seedCycleProbability);

            int numGraphs = 10000;

            ArrayList<List<DefaultGraph<Integer, Integer>>> groundTruth = null;
            if (generate) {
                groundTruth = graphGen.createGraphsClustered((int) (numGraphs * (1 - noiseRatio) + 0.5));
                groundTruth.add(noiseGen.createGraphs((int) (numGraphs * noiseRatio + 0.5)));

            } else {
                groundTruth = TSHelpers.loadIntLabeledGmlClusters(FilenameUtils
                        .concat(FilenameUtils.concat(baseFolderPath, loadRun), DefaultPaths.datasetSubfolderName));
            }

            ArrayList<GraphContainer<Integer, Integer, DefaultGraph<Integer, Integer>>> graphContainers = TSHelpers
                    .groundTruth2filteredGCs(groundTruth);
            TSHelpers.logDatasetSummary(generate, graphContainers);

            /*
             * clustering configuration
             */
            SharedMemorySCC<Integer, Integer, DefaultGraph<Integer, Integer>> conf = conf(new DefaultGraphFactory<>(),
                    parallelism);
            conf.convergence = new FixedIterationNumberCC(12);

            SharedMemorySC<Integer, Integer, DefaultGraph<Integer, Integer>> clust = new SharedMemorySC<>(
                    graphContainers, conf);

            ArrayList<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>> gtClusters = Clusters
                    .toClusterList(groundTruth, new ClusterFactory<>(conf));

            /*
             * evaluation configuration
             */
            String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

            conf.evalModuleInjection = false;

            ArrayList<EvalModule<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>>> modules = new ArrayList<>();
            modules.add(new PerfEM<>(clust));
            modules.add(new CoverageEM<>(clust));
            modules.add(new ClustSizeEM<>(clust));
            modules.add(new RepSizeEM<>(clust));
            modules.add(new RepToMemberSizeEM<>(clust));
            modules.add(new NumClustersEM<>(clust));
            modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
            modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
            modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
            modules.add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
            modules.add(
                    new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, DefaultPaths.clusteringSubfolderName),
                            FilenameUtils.getBaseName(folderPath), true, false, true));

            SeriesEM<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>> series = new SeriesEM<>(modules);

            FileSaverEMM saver = new FileSaverEMM(
                    FilenameUtils.concat(folderPath,
                            FilenameUtils.concat(DefaultPaths.evalSubfolderName, DefaultPaths.evalResultFileName)),
                    true, false);
            LinkedList<EvalMetaModule> mm = new LinkedList<>();
            mm.push(new RepeaterEMM(repeats, false));
            MetaDataAdderEMM generatorDescription = new MetaDataAdderEMM("graph generator", graphGen.getDescription());
            MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
            mm.push(TSHelpers.hostNameMetaData());
            mm.push(generatorDescription);
            mm.push(datasetSize);
            mm.push(saver);
            mm.push(new StatisticsEMM());

            TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, generatorDescription, datasetSize);

            if (generate) {
                TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);
            }

            results.addAll(saver.run(mm, series));

            TSHelpers.saveInjectionResults(conf, clust, folderPath);

            return results;
        }

    },
    /**
     * same test settings as
     * {@link PublicationADMA2017TS#ReproduceMultipleSeeds} but for ProClus
     * subspace clustering.
     */
    ReproduceMultipleSeedsProclus {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            boolean generate = true; // otherwise load
            String loadRun = "012";

            int vertexLabelCount = 10;
            int edgeLabelCount = 3;
            AbstractDistribution labelDistribution = new Exponential(1.0, new MersenneTwister());
            AbstractDistribution seedSizeDistribution = new Poisson(10, new MersenneTwister());
            double seedCycleProbability = 0.1;
            int numClusters = 100;
            AbstractDistribution clusterWeightDistribution = new Dirac(1);
            int clusterSeedPoolSize = 3;
            int commonSeedPoolSize = 10;
            AbstractDistribution usedCommonSeedsDistribution = new Exponential(3, new MersenneTwister());
            AbstractDistribution connectionCountDistribution = new Poisson(5, new MersenneTwister());

            AbstractDistribution noiseSizeDistribution = new Poisson(50, new MersenneTwister());
            double noiseRatio = 0.05;

            SeedCombinationCGG graphGen = new SeedCombinationCGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    seedSizeDistribution, seedCycleProbability, numClusters, clusterWeightDistribution,
                    clusterSeedPoolSize, commonSeedPoolSize, usedCommonSeedsDistribution, connectionCountDistribution);
            ConnectedRGG noiseGen = new ConnectedRGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    noiseSizeDistribution, seedCycleProbability);

            int[] datasetSizes = { 100, 1000, 2000, 5000, 10000 };

            for (int numGraphs : datasetSizes) {
                ArrayList<List<DefaultGraph<Integer, Integer>>> groundTruth = null;
                if (generate) {
                    groundTruth = graphGen.createGraphsClustered((int) (numGraphs * (1 - noiseRatio) + 0.5));
                    groundTruth.add(noiseGen.createGraphs((int) (numGraphs * noiseRatio + 0.5)));

                } else {
                    groundTruth = TSHelpers.loadIntLabeledGmlClusters(
                            FilenameUtils.concat(FilenameUtils.concat(baseFolderPath, loadRun), "dataset"));
                }

                ArrayList<GraphContainer<Integer, Integer, DefaultGraph<Integer, Integer>>> graphContainers = TSHelpers
                        .groundTruth2filteredGCs(groundTruth);
                TSHelpers.logDatasetSummary(generate, graphContainers);

                /*
                 * clustering configuration
                 */
                ProclusConf<Integer, Integer, DefaultGraph<Integer, Integer>> conf = new ProclusConf<>();
                conf.toFeatureVectorFunction = new ExplicitKernelTFVF<>(new GraphletKernel<>());
                conf.numClusters = numClusters;
                conf.avgDimension = 20;
                StructuralClustering<Integer, Integer, DefaultGraph<Integer, Integer>> clust = new WrappedProclus<>(
                        conf, graphContainers);

                ArrayList<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>> gtClusters = Clusters
                        .toClusterList(groundTruth, new ClusterFactory<>(GraphSize.edges));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                ArrayList<EvalModule<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                        FilenameUtils.getBaseName(folderPath), true, false, true));

                SeriesEM<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>> series = new SeriesEM<>(
                        modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(new RepeaterEMM(15, false));
                MetaDataAdderEMM generatorDescription = new MetaDataAdderEMM("graph generator",
                        graphGen.getDescription());
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(generatorDescription);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, generatorDescription,
                        datasetSize);

                if (generate) {
                    TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                    TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);
                }

                results.addAll(saver.run(mm, series));

            }
            return results;
        }
    },
    /**
     * same test settings as
     * {@link PublicationADMA2017TS#ReproduceMultipleSeeds} but for Kernel
     * K-Means clustering.
     */
    ReproduceMultipleSeedsKernelKMeans {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            boolean generate = true; // otherwise load
            String loadRun = "012";
            int repeats = 30;

            int vertexLabelCount = 10;
            int edgeLabelCount = 3;
            AbstractDistribution labelDistribution = new Exponential(1.0, new MersenneTwister());
            AbstractDistribution seedSizeDistribution = new Poisson(10, new MersenneTwister());
            double seedCycleProbability = 0.1;
            int numClusters = 100;
            AbstractDistribution clusterWeightDistribution = new Dirac(1);
            int clusterSeedPoolSize = 3;
            int commonSeedPoolSize = 10;
            AbstractDistribution usedCommonSeedsDistribution = new Exponential(3, new MersenneTwister());
            AbstractDistribution connectionCountDistribution = new Poisson(5, new MersenneTwister());

            AbstractDistribution noiseSizeDistribution = new Poisson(50, new MersenneTwister());
            double noiseRatio = 0.05;

            SeedCombinationCGG graphGen = new SeedCombinationCGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    seedSizeDistribution, seedCycleProbability, numClusters, clusterWeightDistribution,
                    clusterSeedPoolSize, commonSeedPoolSize, usedCommonSeedsDistribution, connectionCountDistribution);
            ConnectedRGG noiseGen = new ConnectedRGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    noiseSizeDistribution, seedCycleProbability);

            int[] datasetSizes = { 1000, 5000, 10000 };

            for (int numGraphs : datasetSizes) {
                ArrayList<List<DefaultGraph<Integer, Integer>>> groundTruth = null;
                if (generate) {
                    groundTruth = graphGen.createGraphsClustered((int) (numGraphs * (1 - noiseRatio) + 0.5));
                    groundTruth.add(noiseGen.createGraphs((int) (numGraphs * noiseRatio + 0.5)));

                } else {
                    groundTruth = TSHelpers.loadIntLabeledGmlClusters(
                            FilenameUtils.concat(FilenameUtils.concat(baseFolderPath, loadRun), "dataset"));
                }

                ArrayList<GraphContainer<Integer, Integer, DefaultGraph<Integer, Integer>>> graphContainers = TSHelpers
                        .groundTruth2filteredGCs(groundTruth);
                TSHelpers.logDatasetSummary(generate, graphContainers);

                /*
                 * clustering configuration
                 */
                ExplicitKernelKMeansConf<Integer, Integer, DefaultGraph<Integer, Integer>, String> conf = new ExplicitKernelKMeansConf<>();
                conf.kernel = new GraphletKernel<>();
                conf.numCluster = groundTruth.size();
                StructuralClustering<Integer, Integer, DefaultGraph<Integer, Integer>> clust = new WrappedExplicitKernelKMeans<>(
                        conf, graphContainers);

                ArrayList<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>> gtClusters = Clusters
                        .toClusterList(groundTruth, new ClusterFactory<>(GraphSize.edges));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                ArrayList<EvalModule<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                        FilenameUtils.getBaseName(folderPath), true, false, true));

                SeriesEM<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>> series = new SeriesEM<>(
                        modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(new RepeaterEMM(repeats, false));
                MetaDataAdderEMM generatorDescription = new MetaDataAdderEMM("graph generator",
                        graphGen.getDescription());
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(generatorDescription);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, generatorDescription,
                        datasetSize);

                if (generate) {
                    TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                    TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);
                }

                results.addAll(saver.run(mm, series));

            }
            return results;
        }
    },
    /**
     * Clustering Evaluation that generates synthetic graphs from seed patterns.
     * 
     * This Test only considers different settings for the number of
     * representative candidates for a fixed dataset size.
     */
    ReproduceMultipleSeedsRepCandidatesInfluence {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int parallelism = 16;
            int repeats = 30;
            boolean generate = false; // otherwise load
            String loadRun = "001";

            int vertexLabelCount = 10;
            int edgeLabelCount = 3;
            AbstractDistribution labelDistribution = new Exponential(1.0, new MersenneTwister());
            AbstractDistribution seedSizeDistribution = new Poisson(10, new MersenneTwister());
            double seedCycleProbability = 0.1;
            int numClusters = 100;
            AbstractDistribution clusterWeightDistribution = new Dirac(1);
            int clusterSeedPoolSize = 3;
            int commonSeedPoolSize = 10;
            AbstractDistribution usedCommonSeedsDistribution = new Exponential(3, new MersenneTwister());
            AbstractDistribution connectionCountDistribution = new Poisson(5, new MersenneTwister());

            AbstractDistribution noiseSizeDistribution = new Poisson(50, new MersenneTwister());
            double noiseRatio = 0.05;

            SeedCombinationCGG graphGen = new SeedCombinationCGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    seedSizeDistribution, seedCycleProbability, numClusters, clusterWeightDistribution,
                    clusterSeedPoolSize, commonSeedPoolSize, usedCommonSeedsDistribution, connectionCountDistribution);
            ConnectedRGG noiseGen = new ConnectedRGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    noiseSizeDistribution, seedCycleProbability);

            int numGraphs = 10000;

            ArrayList<List<DefaultGraph<Integer, Integer>>> groundTruth = null;
            if (generate) {
                groundTruth = graphGen.createGraphsClustered((int) (numGraphs * (1 - noiseRatio) + 0.5));
                groundTruth.add(noiseGen.createGraphs((int) (numGraphs * noiseRatio + 0.5)));

            } else {
                String path = FilenameUtils.concat(FilenameUtils.concat(baseFolderPath, loadRun),
                        DefaultPaths.datasetSubfolderName);
                groundTruth = TSHelpers.loadIntLabeledGmlClusters(path);
            }

            ArrayList<GraphContainer<Integer, Integer, DefaultGraph<Integer, Integer>>> graphContainers = TSHelpers
                    .groundTruth2filteredGCs(groundTruth);
            TSHelpers.logDatasetSummary(generate, graphContainers);

            int[] numRepCandidatesArr = { 10, 25, 50, 100, 250 };

            for (int numRepCandidates : numRepCandidatesArr) {
                /*
                 * clustering configuration
                 */
                SharedMemorySCC<Integer, Integer, DefaultGraph<Integer, Integer>> conf = conf(
                        new DefaultGraphFactory<>(), parallelism);
                conf.newRepCount = numRepCandidates;
                SharedMemorySC<Integer, Integer, DefaultGraph<Integer, Integer>> clust = new SharedMemorySC<>(
                        graphContainers, conf);

                ArrayList<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>> gtClusters = Clusters
                        .toClusterList(groundTruth, new ClusterFactory<>(conf));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                conf.evalModuleInjection = false;

                ArrayList<EvalModule<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new CoverageEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new RepSizeEM<>(clust));
                modules.add(new RepToMemberSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
                modules.add(new GmlClustSaverEM<>(clust,
                        FilenameUtils.concat(folderPath, DefaultPaths.clusteringSubfolderName),
                        FilenameUtils.getBaseName(folderPath), true, false, true));

                SeriesEM<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>> series = new SeriesEM<>(
                        modules);

                FileSaverEMM saver = new FileSaverEMM(
                        FilenameUtils.concat(folderPath,
                                FilenameUtils.concat(DefaultPaths.evalSubfolderName, DefaultPaths.evalResultFileName)),
                        true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(new RepeaterEMM(repeats, false));
                MetaDataAdderEMM generatorDescription = new MetaDataAdderEMM("graph generator",
                        graphGen.getDescription());
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(generatorDescription);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, generatorDescription,
                        datasetSize);

                if (generate) {
                    TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                    TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);
                }

                results.addAll(saver.run(mm, series));

                TSHelpers.saveInjectionResults(conf, clust, folderPath);
            }

            return results;
        }

    },
    /**
     * This Test only considers different settings for the splitting threshold.
     */
    ReproduceMultipleSeedsSplittingThreshInfluence {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int parallelism = 16;
            int repeats = 30;
            boolean generate = false; // otherwise load
            String loadRun = "001";

            int vertexLabelCount = 10;
            int edgeLabelCount = 3;
            AbstractDistribution labelDistribution = new Exponential(1.0, new MersenneTwister());
            AbstractDistribution seedSizeDistribution = new Poisson(10, new MersenneTwister());
            double seedCycleProbability = 0.1;
            int numClusters = 100;
            AbstractDistribution clusterWeightDistribution = new Dirac(1);
            int clusterSeedPoolSize = 3;
            int commonSeedPoolSize = 10;
            AbstractDistribution usedCommonSeedsDistribution = new Exponential(3, new MersenneTwister());
            AbstractDistribution connectionCountDistribution = new Poisson(5, new MersenneTwister());

            AbstractDistribution noiseSizeDistribution = new Poisson(50, new MersenneTwister());
            double noiseRatio = 0.05;

            SeedCombinationCGG graphGen = new SeedCombinationCGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    seedSizeDistribution, seedCycleProbability, numClusters, clusterWeightDistribution,
                    clusterSeedPoolSize, commonSeedPoolSize, usedCommonSeedsDistribution, connectionCountDistribution);
            ConnectedRGG noiseGen = new ConnectedRGG(vertexLabelCount, edgeLabelCount, labelDistribution,
                    noiseSizeDistribution, seedCycleProbability);

            int numGraphs = 10000;

            ArrayList<List<DefaultGraph<Integer, Integer>>> groundTruth = null;
            if (generate) {
                groundTruth = graphGen.createGraphsClustered((int) (numGraphs * (1 - noiseRatio) + 0.5));
                groundTruth.add(noiseGen.createGraphs((int) (numGraphs * noiseRatio + 0.5)));

            } else {
                String path = FilenameUtils.concat(FilenameUtils.concat(baseFolderPath, loadRun),
                        DefaultPaths.datasetSubfolderName);
                groundTruth = TSHelpers.loadIntLabeledGmlClusters(path);
            }

            ArrayList<GraphContainer<Integer, Integer, DefaultGraph<Integer, Integer>>> graphContainers = TSHelpers
                    .groundTruth2filteredGCs(groundTruth);
            TSHelpers.logDatasetSummary(generate, graphContainers);

            double[] splitThreshs = { 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9 };

            for (double splitThresh : splitThreshs) {
                /*
                 * clustering configuration
                 */
                SharedMemorySCC<Integer, Integer, DefaultGraph<Integer, Integer>> conf = conf(
                        new DefaultGraphFactory<>(), parallelism);
                conf.splitCriteria = new RelRepToMemberSizeSC(splitThresh);
                SharedMemorySC<Integer, Integer, DefaultGraph<Integer, Integer>> clust = new SharedMemorySC<>(
                        graphContainers, conf);

                ArrayList<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>> gtClusters = Clusters
                        .toClusterList(groundTruth, new ClusterFactory<>(conf));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                conf.evalModuleInjection = false;

                ArrayList<EvalModule<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new CoverageEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new RepSizeEM<>(clust));
                modules.add(new RepToMemberSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
                modules.add(new GmlClustSaverEM<>(clust,
                        FilenameUtils.concat(folderPath, DefaultPaths.clusteringSubfolderName),
                        FilenameUtils.getBaseName(folderPath), true, false, true));

                SeriesEM<List<Cluster<Integer, Integer, DefaultGraph<Integer, Integer>>>> series = new SeriesEM<>(
                        modules);

                FileSaverEMM saver = new FileSaverEMM(
                        FilenameUtils.concat(folderPath,
                                FilenameUtils.concat(DefaultPaths.evalSubfolderName, DefaultPaths.evalResultFileName)),
                        true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(new RepeaterEMM(repeats, false));
                MetaDataAdderEMM generatorDescription = new MetaDataAdderEMM("graph generator",
                        graphGen.getDescription());
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(generatorDescription);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, generatorDescription,
                        datasetSize);

                if (generate) {
                    TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                    TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);
                }

                results.addAll(saver.run(mm, series));

                TSHelpers.saveInjectionResults(conf, clust, folderPath);
            }

            return results;
        }

    },
    /**
     * Clustering evaluation that uses the mcr (anchorquery) dataset and tries
     * to reproduce the reaction labels
     */
    ReproduceAnchorQuery {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int parallelism = 16;
            int repeats = 15;

            String sdfPath = "~/AnchorQuery.sdf";
            final String propName = "ReactionName";

            Set<String> propNames = new TreeSet<>();
            propNames.add(propName);
            ArrayList<GraphContainer<String, String, MoleculeGraph>> graphContainers = GraphIO
                    .sdfToMoleculeGraphContainersWithProperties(sdfPath, true, true, Integer.MAX_VALUE, propNames,
                            m -> true);

            int numGraphs = graphContainers.size();
            TSHelpers.logDatasetSummary(false, graphContainers);

            /*
             * clustering configuration
             */
            SharedMemorySCC<String, String, MoleculeGraph> conf = conf(new MoleculeGraphFactory(), parallelism);
            SharedMemorySC<String, String, MoleculeGraph> clust = new SharedMemorySC<>(graphContainers, conf);

            Map<String, List<GraphContainer<String, String, MoleculeGraph>>> groups = graphContainers.stream()
                    .collect(Collectors.groupingBy(
                            (GraphContainer<String, String, MoleculeGraph> gc) -> gc.getProperty(propName)));
            ArrayList<List<GraphContainer<String, String, MoleculeGraph>>> groundTruth = new ArrayList<>();
            for (Entry<String, List<GraphContainer<String, String, MoleculeGraph>>> entry : groups.entrySet()) {
                groundTruth.add(entry.getValue().stream().collect(Collectors.toList()));
            }
            ArrayList<Cluster<String, String, MoleculeGraph>> gtClusters = Clusters.toClusterListGC(groundTruth,
                    new ClusterFactory<>(conf));

            /*
             * evaluation configuration
             */
            String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

            conf.evalModuleInjection = false;

            ArrayList<EvalModule<List<Cluster<String, String, MoleculeGraph>>>> modules = new ArrayList<>();
            modules.add(new PerfEM<>(clust));
            modules.add(new CoverageEM<>(clust));
            modules.add(new ClustSizeEM<>(clust));
            modules.add(new RepSizeEM<>(clust));
            modules.add(new RepToMemberSizeEM<>(clust));
            modules.add(new NumClustersEM<>(clust));
            modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
            modules.add(new ClusteringComparisonEM<>(clust,
                    new OverlappingPurityCC<>(Sets.newHashSet("Name", "ReactionName"), "\\r?\\n"), gtClusters));
            modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
            modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
            modules.add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
            modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                    FilenameUtils.getBaseName(folderPath), true, false, true));

            SeriesEM<List<Cluster<String, String, MoleculeGraph>>> series = new SeriesEM<>(modules);

            FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
            LinkedList<EvalMetaModule> mm = new LinkedList<>();
            mm.push(new RepeaterEMM(repeats, false));
            MetaDataAdderEMM datasetName = new MetaDataAdderEMM("dataset", sdfPath);
            MetaDataAdderEMM labelProperty = new MetaDataAdderEMM("ground truth label property", propName);
            MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
            mm.push(TSHelpers.hostNameMetaData());
            mm.push(datasetName);
            mm.push(labelProperty);
            mm.push(datasetSize);
            mm.push(saver);
            mm.push(new StatisticsEMM());

            TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, datasetName, labelProperty,
                    datasetSize);

            TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
            TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);

            results.addAll(saver.run(mm, series));

            TSHelpers.saveInjectionResults(conf, clust, folderPath);

            return results;
        }
    },
    /**
     * Clustering evaluation that uses the mcr (anchorquery) dataset and tries
     * to reproduce the reaction labels
     */
    ReproduceAnchorQueryConvergence {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int parallelism = 16;
            int repeats = 15;

            String sdfPath = "~/AnchorQuery.sdf";
            final String propName = "ReactionName";

            Set<String> propNames = new TreeSet<>();
            propNames.add(propName);
            ArrayList<GraphContainer<String, String, MoleculeGraph>> graphContainers = GraphIO
                    .sdfToMoleculeGraphContainersWithProperties(sdfPath, true, true, Integer.MAX_VALUE, propNames,
                            m -> true);

            int numGraphs = graphContainers.size();
            TSHelpers.logDatasetSummary(false, graphContainers);

            /*
             * clustering configuration
             */
            SharedMemorySCC<String, String, MoleculeGraph> conf = conf(new MoleculeGraphFactory(), parallelism);
            ConvergenceCriteria origConvCrit = conf.convergence;
            conf.convergence = new FixedIterationNumberCC(30);

            SharedMemorySC<String, String, MoleculeGraph> clust = new SharedMemorySC<>(graphContainers, conf);

            Map<String, List<GraphContainer<String, String, MoleculeGraph>>> groups = graphContainers.stream()
                    .collect(Collectors.groupingBy(
                            (GraphContainer<String, String, MoleculeGraph> gc) -> gc.getProperty(propName)));
            ArrayList<List<GraphContainer<String, String, MoleculeGraph>>> groundTruth = new ArrayList<>();
            for (Entry<String, List<GraphContainer<String, String, MoleculeGraph>>> entry : groups.entrySet()) {
                groundTruth.add(entry.getValue().stream().collect(Collectors.toList()));
            }
            ArrayList<Cluster<String, String, MoleculeGraph>> gtClusters = Clusters.toClusterListGC(groundTruth,
                    new ClusterFactory<>(conf));

            /*
             * evaluation configuration
             */
            String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

            conf.evalModuleInjection = true;
            // conf.injectionModules.add(new CoverageEM<>(clust));
            conf.injectionModules.add(new ClustSizeEM<>(clust));
            conf.injectionModules.add(new RepSizeEM<>(clust));
            conf.injectionModules.add(new RepToMemberSizeEM<>(clust));
            conf.injectionModules.add(new NumClustersEM<>(clust));
            conf.injectionModules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
            conf.injectionModules
                    .add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
            conf.injectionModules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
            conf.injectionModules.add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
            conf.injectionModules.add(new ConvergenceEM<>(clust, origConvCrit, true));

            ArrayList<EvalModule<List<Cluster<String, String, MoleculeGraph>>>> modules = new ArrayList<>();
            modules.add(new PerfEM<>(clust));
            // modules.add(new CoverageEM<>(clust));
            modules.add(new ClustSizeEM<>(clust));
            modules.add(new RepSizeEM<>(clust));
            modules.add(new RepToMemberSizeEM<>(clust));
            modules.add(new NumClustersEM<>(clust));
            modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
            modules.add(new ClusteringComparisonEM<>(clust,
                    new OverlappingPurityCC<>(Sets.newHashSet("Name", "ReactionName"), "\\r?\\n"), gtClusters));
            modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
            modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
            modules.add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
            modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                    FilenameUtils.getBaseName(folderPath), true, false, true));
            modules.add(new InjectionSaverEM<>(clust, folderPath));

            SeriesEM<List<Cluster<String, String, MoleculeGraph>>> series = new SeriesEM<>(modules);

            FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
            LinkedList<EvalMetaModule> mm = new LinkedList<>();
            mm.push(new RepeaterEMM(repeats, false));
            MetaDataAdderEMM datasetName = new MetaDataAdderEMM("dataset", sdfPath);
            MetaDataAdderEMM labelProperty = new MetaDataAdderEMM("ground truth label property", propName);
            MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
            mm.push(TSHelpers.hostNameMetaData());
            mm.push(datasetName);
            mm.push(labelProperty);
            mm.push(datasetSize);
            mm.push(saver);
            mm.push(new StatisticsEMM());

            TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, datasetName, labelProperty,
                    datasetSize);

            TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
            TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);

            results.addAll(saver.run(mm, series));

            return results;
        }
    },
    /**
     * Clustering evaluation that uses the mcr (anchorquery) dataset and tries
     * to reproduce the reaction labels
     * 
     * This Test only considers different settings for the number of
     * representative candidates for a fixed dataset size.
     */
    ReproduceAnchorQueryRepCandidatesInfluence {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int parallelism = 16;
            int repeats = 15;

            String sdfPath = "~/AnchorQuery.sdf";
            final String propName = "ReactionName";

            Set<String> propNames = new TreeSet<>();
            propNames.add(propName);
            ArrayList<GraphContainer<String, String, MoleculeGraph>> graphContainers = GraphIO
                    .sdfToMoleculeGraphContainersWithProperties(sdfPath, true, true, Integer.MAX_VALUE, propNames,
                            m -> true);

            int numGraphs = graphContainers.size();
            TSHelpers.logDatasetSummary(false, graphContainers);

            int[] numRepCandidatesArr = { 10, 25, 50, 100, 250 };

            for (int numRepCandidates : numRepCandidatesArr) {

                /*
                 * clustering configuration
                 */
                SharedMemorySCC<String, String, MoleculeGraph> conf = conf(new MoleculeGraphFactory(), parallelism);
                conf.newRepCount = numRepCandidates;
                SharedMemorySC<String, String, MoleculeGraph> clust = new SharedMemorySC<>(graphContainers, conf);

                Map<String, List<GraphContainer<String, String, MoleculeGraph>>> groups = graphContainers.stream()
                        .collect(Collectors.groupingBy(
                                (GraphContainer<String, String, MoleculeGraph> gc) -> gc.getProperty(propName)));
                ArrayList<List<GraphContainer<String, String, MoleculeGraph>>> groundTruth = new ArrayList<>();
                for (Entry<String, List<GraphContainer<String, String, MoleculeGraph>>> entry : groups.entrySet()) {
                    groundTruth.add(entry.getValue().stream().collect(Collectors.toList()));
                }
                ArrayList<Cluster<String, String, MoleculeGraph>> gtClusters = Clusters.toClusterListGC(groundTruth,
                        new ClusterFactory<>(conf));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                conf.evalModuleInjection = false;

                ArrayList<EvalModule<List<Cluster<String, String, MoleculeGraph>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new CoverageEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new RepSizeEM<>(clust));
                modules.add(new RepToMemberSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust,
                        new OverlappingPurityCC<>(Sets.newHashSet("Name", "ReactionName"), "\\r?\\n"), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
                modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                        FilenameUtils.getBaseName(folderPath), true, false, true));

                SeriesEM<List<Cluster<String, String, MoleculeGraph>>> series = new SeriesEM<>(modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(new RepeaterEMM(repeats, false));
                MetaDataAdderEMM datasetName = new MetaDataAdderEMM("dataset", sdfPath);
                MetaDataAdderEMM labelProperty = new MetaDataAdderEMM("ground truth label property", propName);
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(datasetName);
                mm.push(labelProperty);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, datasetName, labelProperty,
                        datasetSize);

                TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);

                results.addAll(saver.run(mm, series));

                TSHelpers.saveInjectionResults(conf, clust, folderPath);
            }
            return results;
        }
    },
    /**
     * Clustering evaluation that uses the mcr reaction dataset and tries to
     * reproduce the reaction labels
     */
    ReproduceAnchorQueryProclus {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int repeats = 15;

            String datasetFolder = "~";
            String[] sdfPaths = { "AnchorQuery.sdf" };
            final String propName = "ReactionName";

            for (String sdfPath : sdfPaths) {
                sdfPath = FilenameUtils.concat(datasetFolder, sdfPath);

                Set<String> propNames = new TreeSet<>();
                propNames.add(propName);
                ArrayList<GraphContainer<String, String, MoleculeGraph>> graphContainers = GraphIO
                        .sdfToMoleculeGraphContainersWithProperties(sdfPath, true, true, Integer.MAX_VALUE, propNames,
                                m -> m.getProperty(propName) != null);

                int numGraphs = graphContainers.size();
                TSHelpers.logDatasetSummary(false, graphContainers);

                Map<String, List<GraphContainer<String, String, MoleculeGraph>>> groups = graphContainers.stream()
                        .collect(Collectors.groupingBy(
                                (GraphContainer<String, String, MoleculeGraph> gc) -> gc.getProperty(propName)));
                ArrayList<List<GraphContainer<String, String, MoleculeGraph>>> groundTruth = new ArrayList<>();
                for (Entry<String, List<GraphContainer<String, String, MoleculeGraph>>> entry : groups.entrySet()) {
                    groundTruth.add(entry.getValue().stream().collect(Collectors.toList()));
                }

                /*
                 * clustering configuration
                 */
                ProclusConf<String, String, MoleculeGraph> conf = new ProclusConf<>();
                conf.toFeatureVectorFunction = new ExplicitKernelTFVF<>(new GraphletKernel<>());
                conf.numClusters = groundTruth.size();
                conf.avgDimension = 20;
                StructuralClustering<String, String, MoleculeGraph> clust = new WrappedProclus<>(conf, graphContainers);

                ArrayList<Cluster<String, String, MoleculeGraph>> gtClusters = Clusters.toClusterListGC(groundTruth,
                        new ClusterFactory<>(GraphSize.edges));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                ArrayList<EvalModule<List<Cluster<String, String, MoleculeGraph>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                        FilenameUtils.getBaseName(folderPath), true, false, true));

                SeriesEM<List<Cluster<String, String, MoleculeGraph>>> series = new SeriesEM<>(modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(new RepeaterEMM(repeats, false));
                MetaDataAdderEMM generatorDescription = new MetaDataAdderEMM("dataset path", sdfPath);
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(generatorDescription);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, generatorDescription,
                        datasetSize);

                results.addAll(saver.run(mm, series));
            }
            return results;
        }
    },
    /**
     * Clustering evaluation that uses the mcr reaction dataset and tries to
     * reproduce the reaction labels
     */
    ReproduceAnchorQueryKernelKMeans {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int repeats = 15;

            String datasetFolder = "~";
            String[] sdfPaths = { "AnchorQuery.sdf" };
            final String propName = "ReactionName";

            for (String sdfPath : sdfPaths) {
                sdfPath = FilenameUtils.concat(datasetFolder, sdfPath);
                Set<String> propNames = new TreeSet<>();
                propNames.add(propName);
                ArrayList<GraphContainer<String, String, MoleculeGraph>> graphContainers = GraphIO
                        .sdfToMoleculeGraphContainersWithProperties(sdfPath, true, true, Integer.MAX_VALUE, propNames,
                                m -> m.getProperty(propName) != null);

                int numGraphs = graphContainers.size();
                TSHelpers.logDatasetSummary(false, graphContainers);

                Map<String, List<GraphContainer<String, String, MoleculeGraph>>> groups = graphContainers.stream()
                        .collect(Collectors.groupingBy(
                                (GraphContainer<String, String, MoleculeGraph> gc) -> gc.getProperty(propName)));
                ArrayList<List<GraphContainer<String, String, MoleculeGraph>>> groundTruth = new ArrayList<>();
                for (Entry<String, List<GraphContainer<String, String, MoleculeGraph>>> entry : groups.entrySet()) {
                    groundTruth.add(entry.getValue().stream().collect(Collectors.toList()));
                }

                /*
                 * clustering configuration
                 */
                ExplicitKernelKMeansConf<String, String, MoleculeGraph, String> conf = new ExplicitKernelKMeansConf<>();
                conf.kernel = new GraphletKernel<>();
                conf.numCluster = groundTruth.size();
                StructuralClustering<String, String, MoleculeGraph> clust = new WrappedExplicitKernelKMeans<>(conf,
                        graphContainers);

                ArrayList<Cluster<String, String, MoleculeGraph>> gtClusters = Clusters.toClusterListGC(groundTruth,
                        new ClusterFactory<>(GraphSize.edges));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                ArrayList<EvalModule<List<Cluster<String, String, MoleculeGraph>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                        FilenameUtils.getBaseName(folderPath), true, false, true));

                SeriesEM<List<Cluster<String, String, MoleculeGraph>>> series = new SeriesEM<>(modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(new RepeaterEMM(repeats, false));
                MetaDataAdderEMM generatorDescription = new MetaDataAdderEMM("dataset path", sdfPath);
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(generatorDescription);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, generatorDescription,
                        datasetSize);

                results.addAll(saver.run(mm, series));
            }
            return results;
        }
    },
    /**
     * Reads in the SCAP results and compares them to a ground truth
     */
    EvaluateScapRmsResults {

        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            // String runs[] = { "001", "002", "003", "004", "005" };
            String runs[] = { "008" };

            for (String run : runs) {

                String gmlSourceFile = "/home/schaefer/data/promotion/projekte/spp-big_data/00_struclust/eval_uncompressed/datasets/PublicationADMA2017TS.ReproduceMultipleSeeds/"
                        + run + "/all.gml";
                String groundTruthFolder = "/home/schaefer/data/promotion/projekte/spp-big_data/00_struclust/eval_uncompressed/datasets/PublicationADMA2017TS.ReproduceMultipleSeeds/"
                        + run;
                String scapResultFile = "/home/schaefer/data/promotion/projekte/spp-big_data/00_struclust/eval_results/snail01.cs.tu-dortmund.de/scap_comparison/PublicationADMA2017TS.ReproduceMultipleSeeds/"
                        + run + "/clusters_final.txt";
                GraphSize gSize = GraphSize.edges;

                ArrayList<List<DefaultGraph<String, String>>> gtClusterList = TSHelpers
                        .loadStringLabeledGmlClusters(groundTruthFolder);
                ArrayList<Cluster<String, String, DefaultGraph<String, String>>> groundTruth = Clusters
                        .toClusterList(gtClusterList, new ClusterFactory<>(gSize));
                List<GraphContainer<String, String, DefaultGraph<String, String>>> gtGraphContainers = groundTruth
                        .stream().flatMap(c -> c.stream()).collect(Collectors.toList());

                ArrayList<ArrayList<DefaultGraph<String, String>>> scapResults = TSHelpers
                        .matchScapResults(gmlSourceFile, scapResultFile);
                ArrayList<Cluster<String, String, DefaultGraph<String, String>>> scapClusters = Clusters
                        .toClusterList(scapResults, new ClusterFactory<>(GraphSize.edges));
                List<Cluster<String, String, DefaultGraph<String, String>>> scapReplacedClusters = Clusters
                        .replaceMembersWithMultiIsomorphicGcs(scapClusters, gtGraphContainers);

                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                ArrayList<EvalModule<List<Cluster<String, String, DefaultGraph<String, String>>>>> modules = new ArrayList<>();
                FakeSC<String, String, DefaultGraph<String, String>> fakeSC = new FakeSC<>();
                modules.add(new UsePredefinedClusteringEM<>(scapReplacedClusters, gSize));
                modules.add(new ClusteringComparisonEM<>(fakeSC, new PurityCC<>(), groundTruth));
                modules.add(new ClustSizeEM<>(fakeSC));

                SeriesEM<List<Cluster<String, String, DefaultGraph<String, String>>>> series = new SeriesEM<>(modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                MetaDataAdderEMM datasetSizeAdder = new MetaDataAdderEMM("dataset size",
                        String.valueOf(gtGraphContainers.size()));
                MetaDataAdderEMM gmlSourceFileAdder = new MetaDataAdderEMM("gmlSourceFile", gmlSourceFile);
                MetaDataAdderEMM groundTruthFolderAdder = new MetaDataAdderEMM("groundTruthFolder", groundTruthFolder);
                MetaDataAdderEMM scapResultFileAdder = new MetaDataAdderEMM("scapResultFile", scapResultFile);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(datasetSizeAdder);
                mm.push(gmlSourceFileAdder);
                mm.push(groundTruthFolderAdder);
                mm.push(scapResultFileAdder);

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, fakeSC.getConf(), folderPath, datasetSizeAdder,
                        gmlSourceFileAdder, groundTruthFolderAdder, scapResultFileAdder);

                results.addAll(saver.run(mm, series));
            }

            return results;
        }

    },
    /**
     * Reads in the SCAP results and compares them to a ground truth
     */
    EvaluateScapHeterocyclenResults {

        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            // String runs[] = { "001", "002", "003", "004", "005" };
            // String runs[] = { "001", "002", "003" };
            String runs[] = { "001" };

            for (String run : runs) {

                String gmlSourceFile = "/home/schaefer/data/promotion/projekte/spp-big_data/00_struclust/eval_uncompressed/datasets/PublicationADMA2017TS.ReproduceHeterocyclen/"
                        + run + "/all.gml";
                String groundTruthFolder = "/home/schaefer/data/promotion/projekte/spp-big_data/00_struclust/eval_uncompressed/datasets/PublicationADMA2017TS.ReproduceHeterocyclen/"
                        + run;
                String scapResultFile = "/home/schaefer/data/promotion/projekte/spp-big_data/00_struclust/eval_results/snail01.cs.uni-dortmund.de/PublicationADMA2017TS.ReproduceHeterocyclenScap/"
                        + run + "/clusters_final.txt";
                GraphSize gSize = GraphSize.edges;

                ArrayList<List<DefaultGraph<String, String>>> gtClusterList = TSHelpers
                        .loadStringLabeledGmlClusters(groundTruthFolder);
                ArrayList<Cluster<String, String, DefaultGraph<String, String>>> groundTruth = Clusters
                        .toClusterList(gtClusterList, new ClusterFactory<>(gSize));
                List<GraphContainer<String, String, DefaultGraph<String, String>>> gtGraphContainers = groundTruth
                        .stream().flatMap(c -> c.stream()).collect(Collectors.toList());

                ArrayList<ArrayList<DefaultGraph<String, String>>> scapResults = TSHelpers
                        .matchScapResults(gmlSourceFile, scapResultFile);
                ArrayList<Cluster<String, String, DefaultGraph<String, String>>> scapClusters = Clusters
                        .toClusterList(scapResults, new ClusterFactory<>(GraphSize.edges));
                List<Cluster<String, String, DefaultGraph<String, String>>> scapReplacedClusters = Clusters
                        .replaceMembersWithMultiIsomorphicGcs(scapClusters, gtGraphContainers);

                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                ArrayList<EvalModule<List<Cluster<String, String, DefaultGraph<String, String>>>>> modules = new ArrayList<>();
                FakeSC<String, String, DefaultGraph<String, String>> fakeSC = new FakeSC<>();
                modules.add(new UsePredefinedClusteringEM<>(scapReplacedClusters, gSize));
                modules.add(new ClusteringComparisonEM<>(fakeSC, new PurityCC<>(), groundTruth));
                modules.add(new ClustSizeEM<>(fakeSC));

                SeriesEM<List<Cluster<String, String, DefaultGraph<String, String>>>> series = new SeriesEM<>(modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                MetaDataAdderEMM datasetSizeAdder = new MetaDataAdderEMM("dataset size",
                        String.valueOf(gtGraphContainers.size()));
                MetaDataAdderEMM gmlSourceFileAdder = new MetaDataAdderEMM("gmlSourceFile", gmlSourceFile);
                MetaDataAdderEMM groundTruthFolderAdder = new MetaDataAdderEMM("groundTruthFolder", groundTruthFolder);
                MetaDataAdderEMM scapResultFileAdder = new MetaDataAdderEMM("scapResultFile", scapResultFile);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(datasetSizeAdder);
                mm.push(gmlSourceFileAdder);
                mm.push(groundTruthFolderAdder);
                mm.push(scapResultFileAdder);

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, fakeSC.getConf(), folderPath, datasetSizeAdder,
                        gmlSourceFileAdder, groundTruthFolderAdder, scapResultFileAdder);

                results.addAll(saver.run(mm, series));
            }

            return results;
        }

    },
    /**
     * Reads in the SCAP results and compares them to a ground truth
     */
    EvaluateScapMcrResults {

        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            // String runs[] = { "001", "002", "003", "004", "005" };
            String runs[] = { "001" };

            for (String run : runs) {

                String gmlSourceFile = "/home/schaefer/data/promotion/projekte/spp-big_data/00_struclust/eval_uncompressed/datasets/PublicationADMA2017TS.ReproduceMcrReactions/"
                        + run + "/all.gml";
                String groundTruthFolder = "/home/schaefer/data/promotion/projekte/spp-big_data/00_struclust/eval_uncompressed/datasets/PublicationADMA2017TS.ReproduceMcrReactions/"
                        + run;
                String scapResultFile = "/home/schaefer/data/promotion/projekte/spp-big_data/00_struclust/eval_results/snail01.cs.tu-dortmund.de/scap_comparison/PublicationADMA2017TS.ReproduceMcrReactions/"
                        + run + "/clusters_final.txt";
                GraphSize gSize = GraphSize.edges;

                ArrayList<List<DefaultGraph<String, String>>> gtClusterList = TSHelpers
                        .loadStringLabeledGmlClusters(groundTruthFolder);
                ArrayList<Cluster<String, String, DefaultGraph<String, String>>> groundTruth = Clusters
                        .toClusterList(gtClusterList, new ClusterFactory<>(gSize));
                List<GraphContainer<String, String, DefaultGraph<String, String>>> gtGraphContainers = groundTruth
                        .stream().flatMap(c -> c.stream()).collect(Collectors.toList());

                ArrayList<ArrayList<DefaultGraph<String, String>>> scapResults = TSHelpers
                        .matchScapResults(gmlSourceFile, scapResultFile);
                ArrayList<Cluster<String, String, DefaultGraph<String, String>>> scapClusters = Clusters
                        .toClusterList(scapResults, new ClusterFactory<>(GraphSize.edges));
                List<Cluster<String, String, DefaultGraph<String, String>>> scapReplacedClusters = Clusters
                        .replaceMembersWithMultiIsomorphicGcs(scapClusters, gtGraphContainers);

                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                ArrayList<EvalModule<List<Cluster<String, String, DefaultGraph<String, String>>>>> modules = new ArrayList<>();
                FakeSC<String, String, DefaultGraph<String, String>> fakeSC = new FakeSC<>();
                modules.add(new UsePredefinedClusteringEM<>(scapReplacedClusters, gSize));
                modules.add(new ClusteringComparisonEM<>(fakeSC, new PurityCC<>(), groundTruth));
                modules.add(new ClustSizeEM<>(fakeSC));

                SeriesEM<List<Cluster<String, String, DefaultGraph<String, String>>>> series = new SeriesEM<>(modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                MetaDataAdderEMM datasetSizeAdder = new MetaDataAdderEMM("dataset size",
                        String.valueOf(gtGraphContainers.size()));
                MetaDataAdderEMM gmlSourceFileAdder = new MetaDataAdderEMM("gmlSourceFile", gmlSourceFile);
                MetaDataAdderEMM groundTruthFolderAdder = new MetaDataAdderEMM("groundTruthFolder", groundTruthFolder);
                MetaDataAdderEMM scapResultFileAdder = new MetaDataAdderEMM("scapResultFile", scapResultFile);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(datasetSizeAdder);
                mm.push(gmlSourceFileAdder);
                mm.push(groundTruthFolderAdder);
                mm.push(scapResultFileAdder);

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, fakeSC.getConf(), folderPath, datasetSizeAdder,
                        gmlSourceFileAdder, groundTruthFolderAdder, scapResultFileAdder);

                results.addAll(saver.run(mm, series));
            }

            return results;
        }

    },
    /**
     * Reproduce a-priory classes on heterocyclen dataset
     */
    ReproduceHeterocyclen {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int parallelism = 16;
            int repeats = 30;

            String datasetFolder = "~";
            String[] sdfPaths = { "Heterocyclic.sdf" };
            final String propName = "ReactionName";

            for (String sdfPath : sdfPaths) {
                sdfPath = FilenameUtils.concat(datasetFolder, sdfPath);
                Set<String> propNames = new TreeSet<>();
                propNames.add(propName);
                ArrayList<GraphContainer<String, String, MoleculeGraph>> graphContainers = GraphIO
                        .sdfToMoleculeGraphContainersWithProperties(sdfPath, true, true, Integer.MAX_VALUE, propNames,
                                m -> true);

                int numGraphs = graphContainers.size();
                TSHelpers.logDatasetSummary(false, graphContainers);

                /*
                 * clustering configuration
                 */
                SharedMemorySCC<String, String, MoleculeGraph> conf = conf(new MoleculeGraphFactory(), parallelism);
                SharedMemorySC<String, String, MoleculeGraph> clust = new SharedMemorySC<>(graphContainers, conf);

                Map<String, List<GraphContainer<String, String, MoleculeGraph>>> groups = graphContainers.stream()
                        .collect(Collectors.groupingBy(
                                (GraphContainer<String, String, MoleculeGraph> gc) -> gc.getProperty(propName)));
                ArrayList<List<GraphContainer<String, String, MoleculeGraph>>> groundTruth = new ArrayList<>();
                for (Entry<String, List<GraphContainer<String, String, MoleculeGraph>>> entry : groups.entrySet()) {
                    groundTruth.add(entry.getValue().stream().collect(Collectors.toList()));
                }
                ArrayList<Cluster<String, String, MoleculeGraph>> gtClusters = Clusters.toClusterListGC(groundTruth,
                        new ClusterFactory<>(conf));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                conf.evalModuleInjection = false;

                ArrayList<EvalModule<List<Cluster<String, String, MoleculeGraph>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new CoverageEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new RepSizeEM<>(clust));
                modules.add(new RepToMemberSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust,
                        new OverlappingPurityCC<>(Sets.newHashSet("Name", "ReactionName"), "\\r?\\n"), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
                modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                        FilenameUtils.getBaseName(folderPath), true, false, true));

                SeriesEM<List<Cluster<String, String, MoleculeGraph>>> series = new SeriesEM<>(modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                // mm.push(new RepeaterEMM(numGraphs <= 10000 ? 30 : 15,
                // false));
                mm.push(new RepeaterEMM(repeats, false));
                MetaDataAdderEMM datasetName = new MetaDataAdderEMM("dataset", sdfPath);
                MetaDataAdderEMM labelProperty = new MetaDataAdderEMM("ground truth label property", propName);
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(datasetName);
                mm.push(labelProperty);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, datasetName, labelProperty,
                        datasetSize);

                TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);

                results.addAll(saver.run(mm, series));

                TSHelpers.saveInjectionResults(conf, clust, folderPath);

            }

            return results;
        }
    },
    /**
     * Reproduce a-priory classes on heterocyclen dataset
     */
    ReproduceHeterocyclenConvergence {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int parallelism = 16;
            int repeats = 15;

            String datasetFolder = "~";
            String[] sdfPaths = { "Heterocyclic.sdf" };
            final String propName = "ReactionName";

            for (String sdfPath : sdfPaths) {
                sdfPath = FilenameUtils.concat(datasetFolder, sdfPath);
                Set<String> propNames = new TreeSet<>();
                propNames.add(propName);
                ArrayList<GraphContainer<String, String, MoleculeGraph>> graphContainers = GraphIO
                        .sdfToMoleculeGraphContainersWithProperties(sdfPath, true, true, Integer.MAX_VALUE, propNames,
                                m -> true);

                int numGraphs = graphContainers.size();
                TSHelpers.logDatasetSummary(false, graphContainers);

                /*
                 * clustering configuration
                 */
                SharedMemorySCC<String, String, MoleculeGraph> conf = conf(new MoleculeGraphFactory(), parallelism);
                ConvergenceCriteria origConvCrit = conf.convergence;
                conf.convergence = new FixedIterationNumberCC(50);

                SharedMemorySC<String, String, MoleculeGraph> clust = new SharedMemorySC<>(graphContainers, conf);

                Map<String, List<GraphContainer<String, String, MoleculeGraph>>> groups = graphContainers.stream()
                        .collect(Collectors.groupingBy(
                                (GraphContainer<String, String, MoleculeGraph> gc) -> gc.getProperty(propName)));
                ArrayList<List<GraphContainer<String, String, MoleculeGraph>>> groundTruth = new ArrayList<>();
                for (Entry<String, List<GraphContainer<String, String, MoleculeGraph>>> entry : groups.entrySet()) {
                    groundTruth.add(entry.getValue().stream().collect(Collectors.toList()));
                }
                ArrayList<Cluster<String, String, MoleculeGraph>> gtClusters = Clusters.toClusterListGC(groundTruth,
                        new ClusterFactory<>(conf));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                conf.evalModuleInjection = true;
                conf.injectionModules.add(new CoverageEM<>(clust));
                conf.injectionModules.add(new ClustSizeEM<>(clust));
                conf.injectionModules.add(new RepSizeEM<>(clust));
                conf.injectionModules.add(new RepToMemberSizeEM<>(clust));
                conf.injectionModules.add(new NumClustersEM<>(clust));
                conf.injectionModules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                conf.injectionModules
                        .add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                conf.injectionModules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                conf.injectionModules
                        .add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
                conf.injectionModules.add(new ConvergenceEM<>(clust, origConvCrit, true));

                ArrayList<EvalModule<List<Cluster<String, String, MoleculeGraph>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new CoverageEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new RepSizeEM<>(clust));
                modules.add(new RepToMemberSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust,
                        new OverlappingPurityCC<>(Sets.newHashSet("Name", "ReactionName"), "\\r?\\n"), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
                modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                        FilenameUtils.getBaseName(folderPath), true, false, true));
                modules.add(new InjectionSaverEM<>(clust, folderPath));

                SeriesEM<List<Cluster<String, String, MoleculeGraph>>> series = new SeriesEM<>(modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                // mm.push(new RepeaterEMM(numGraphs <= 10000 ? 30 : 15,
                // false));
                mm.push(new RepeaterEMM(repeats, false));
                MetaDataAdderEMM datasetName = new MetaDataAdderEMM("dataset", sdfPath);
                MetaDataAdderEMM labelProperty = new MetaDataAdderEMM("ground truth label property", propName);
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(datasetName);
                mm.push(labelProperty);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, datasetName, labelProperty,
                        datasetSize);

                TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);

                results.addAll(saver.run(mm, series));
            }

            return results;
        }
    },
    /**
     * This Test only considers different settings for the number of
     * representative candidates for a fixed dataset size.
     */
    ReproduceHeterocyclenRepCandidatesInfluence {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int parallelism = 16;
            int repeats = 30;

            String sdfPath = "~/Heterocyclic.sdf";

            final String propName = "ReactionName";

            Set<String> propNames = new TreeSet<>();
            propNames.add(propName);
            ArrayList<GraphContainer<String, String, MoleculeGraph>> graphContainers = GraphIO
                    .sdfToMoleculeGraphContainersWithProperties(sdfPath, true, true, Integer.MAX_VALUE, propNames,
                            m -> m.getAtomCount() < 90 && m.getBondCount() < 100);

            int numGraphs = graphContainers.size();
            TSHelpers.logDatasetSummary(false, graphContainers);

            int[] numRepCandidatesArr = { 10, 25, 50, 100, 250 };

            for (int numRepCandidates : numRepCandidatesArr) {
                /*
                 * clustering configuration
                 */
                SharedMemorySCC<String, String, MoleculeGraph> conf = conf(new MoleculeGraphFactory(), parallelism);
                conf.newRepCount = numRepCandidates;
                SharedMemorySC<String, String, MoleculeGraph> clust = new SharedMemorySC<>(graphContainers, conf);

                Map<String, List<GraphContainer<String, String, MoleculeGraph>>> groups = graphContainers.stream()
                        .collect(Collectors.groupingBy(
                                (GraphContainer<String, String, MoleculeGraph> gc) -> gc.getProperty(propName)));
                ArrayList<List<GraphContainer<String, String, MoleculeGraph>>> groundTruth = new ArrayList<>();
                for (Entry<String, List<GraphContainer<String, String, MoleculeGraph>>> entry : groups.entrySet()) {
                    groundTruth.add(entry.getValue().stream().collect(Collectors.toList()));
                }
                ArrayList<Cluster<String, String, MoleculeGraph>> gtClusters = Clusters.toClusterListGC(groundTruth,
                        new ClusterFactory<>(conf));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                conf.evalModuleInjection = false;

                ArrayList<EvalModule<List<Cluster<String, String, MoleculeGraph>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new CoverageEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new RepSizeEM<>(clust));
                modules.add(new RepToMemberSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
                modules.add(new GmlClustSaverEM<>(clust,
                        FilenameUtils.concat(folderPath, DefaultPaths.clusteringSubfolderName),
                        FilenameUtils.getBaseName(folderPath), true, false, true));

                SeriesEM<List<Cluster<String, String, MoleculeGraph>>> series = new SeriesEM<>(modules);

                FileSaverEMM saver = new FileSaverEMM(
                        FilenameUtils.concat(folderPath,
                                FilenameUtils.concat(DefaultPaths.evalSubfolderName, DefaultPaths.evalResultFileName)),
                        true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(new RepeaterEMM(repeats, false));
                MetaDataAdderEMM datasetName = new MetaDataAdderEMM("dataset", sdfPath);
                MetaDataAdderEMM labelProperty = new MetaDataAdderEMM("ground truth label property", propName);
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(datasetName);
                mm.push(labelProperty);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, datasetName, labelProperty,
                        datasetSize);

                TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);

                results.addAll(saver.run(mm, series));

                TSHelpers.saveInjectionResults(conf, clust, folderPath);
            }

            return results;
        }

    },
    /**
     * Reproduce a-priory classes on heterocyclen dataset with ProClus subspace
     * clustering.
     */
    ReproduceHeterocyclenProclus {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int repeats = 15;

            String datasetFolder = "~";
            String[] sdfPaths = { "Heterocyclic.sdf" };
            final String propName = "ReactionName";

            for (String sdfPath : sdfPaths) {
                sdfPath = FilenameUtils.concat(datasetFolder, sdfPath);

                Set<String> propNames = new TreeSet<>();
                propNames.add(propName);
                ArrayList<GraphContainer<String, String, MoleculeGraph>> graphContainers = GraphIO
                        .sdfToMoleculeGraphContainersWithProperties(sdfPath, true, true, Integer.MAX_VALUE, propNames,
                                m -> m.getAtomCount() < 90 && m.getBondCount() < 100
                                        && m.getProperty(propName) != null);

                int numGraphs = graphContainers.size();
                TSHelpers.logDatasetSummary(false, graphContainers);

                Map<String, List<GraphContainer<String, String, MoleculeGraph>>> groups = graphContainers.stream()
                        .collect(Collectors.groupingBy(
                                (GraphContainer<String, String, MoleculeGraph> gc) -> gc.getProperty(propName)));
                ArrayList<List<GraphContainer<String, String, MoleculeGraph>>> groundTruth = new ArrayList<>();
                for (Entry<String, List<GraphContainer<String, String, MoleculeGraph>>> entry : groups.entrySet()) {
                    groundTruth.add(entry.getValue().stream().collect(Collectors.toList()));
                }

                /*
                 * clustering configuration
                 */
                ProclusConf<String, String, MoleculeGraph> conf = new ProclusConf<>();
                conf.toFeatureVectorFunction = new ExplicitKernelTFVF<>(new GraphletKernel<>());
                conf.numClusters = groundTruth.size();
                conf.avgDimension = 20;
                StructuralClustering<String, String, MoleculeGraph> clust = new WrappedProclus<>(conf, graphContainers);

                ArrayList<Cluster<String, String, MoleculeGraph>> gtClusters = Clusters.toClusterListGC(groundTruth,
                        new ClusterFactory<>(GraphSize.edges));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                ArrayList<EvalModule<List<Cluster<String, String, MoleculeGraph>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                        FilenameUtils.getBaseName(folderPath), true, false, true));

                SeriesEM<List<Cluster<String, String, MoleculeGraph>>> series = new SeriesEM<>(modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(new RepeaterEMM(repeats, false));
                MetaDataAdderEMM generatorDescription = new MetaDataAdderEMM("dataset path", sdfPath);
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(generatorDescription);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, generatorDescription,
                        datasetSize);

                results.addAll(saver.run(mm, series));
            }
            return results;
        }
    },
    /**
     * Reproduce a-priory classes on heterocyclen dataset with Kernel K-Means
     * clustering.
     */
    ReproduceHeterocyclenKernelKMeans {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int repeats = 15;

            String datasetFolder = "~";
            String[] sdfPaths = { "Heterocyclic.sdf" };
            final String propName = "ReactionName";

            for (String sdfPath : sdfPaths) {
                sdfPath = FilenameUtils.concat(datasetFolder, sdfPath);
                Set<String> propNames = new TreeSet<>();
                propNames.add(propName);
                ArrayList<GraphContainer<String, String, MoleculeGraph>> graphContainers = GraphIO
                        .sdfToMoleculeGraphContainersWithProperties(sdfPath, true, true, Integer.MAX_VALUE, propNames,
                                m -> m.getAtomCount() < 90 && m.getBondCount() < 100
                                        && m.getProperty(propName) != null);

                int numGraphs = graphContainers.size();
                TSHelpers.logDatasetSummary(false, graphContainers);

                Map<String, List<GraphContainer<String, String, MoleculeGraph>>> groups = graphContainers.stream()
                        .collect(Collectors.groupingBy(
                                (GraphContainer<String, String, MoleculeGraph> gc) -> gc.getProperty(propName)));
                ArrayList<List<GraphContainer<String, String, MoleculeGraph>>> groundTruth = new ArrayList<>();
                for (Entry<String, List<GraphContainer<String, String, MoleculeGraph>>> entry : groups.entrySet()) {
                    groundTruth.add(entry.getValue().stream().collect(Collectors.toList()));
                }

                /*
                 * clustering configuration
                 */
                ExplicitKernelKMeansConf<String, String, MoleculeGraph, String> conf = new ExplicitKernelKMeansConf<>();
                conf.kernel = new GraphletKernel<>();
                conf.numCluster = groundTruth.size();
                StructuralClustering<String, String, MoleculeGraph> clust = new WrappedExplicitKernelKMeans<>(conf,
                        graphContainers);

                ArrayList<Cluster<String, String, MoleculeGraph>> gtClusters = Clusters.toClusterListGC(groundTruth,
                        new ClusterFactory<>(GraphSize.edges));

                /*
                 * evaluation configuration
                 */
                String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

                ArrayList<EvalModule<List<Cluster<String, String, MoleculeGraph>>>> modules = new ArrayList<>();
                modules.add(new PerfEM<>(clust));
                modules.add(new ClustSizeEM<>(clust));
                modules.add(new NumClustersEM<>(clust));
                modules.add(new ClusteringComparisonEM<>(clust, new PurityCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new NormVariationOfInformationCC<>(), gtClusters));
                modules.add(new ClusteringComparisonEM<>(clust, new FowlkesMallowsCC<>(), gtClusters));
                modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                        FilenameUtils.getBaseName(folderPath), true, false, true));

                SeriesEM<List<Cluster<String, String, MoleculeGraph>>> series = new SeriesEM<>(modules);

                FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
                LinkedList<EvalMetaModule> mm = new LinkedList<>();
                mm.push(new RepeaterEMM(repeats, false));
                MetaDataAdderEMM generatorDescription = new MetaDataAdderEMM("dataset path", sdfPath);
                MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", String.valueOf(numGraphs));
                mm.push(TSHelpers.hostNameMetaData());
                mm.push(generatorDescription);
                mm.push(datasetSize);
                mm.push(saver);
                mm.push(new StatisticsEMM());

                TSHelpers.clusteredGraphsToGml(groundTruth, folderPath);
                TSHelpers.clusteredGraphsToSdf(groundTruth, folderPath);

                TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, generatorDescription,
                        datasetSize);

                results.addAll(saver.run(mm, series));
            }
            return results;
        }
    },
    /**
     * The default test setup for clustering evaluation.
     */
    ChemDB {
        @Override
        public LinkedList<EvalResult> run() throws Exception {
            LinkedList<EvalResult> results = new LinkedList<>();
            String canonicalDescription = TSHelpers.canonicalTestDescription(this);
            String baseFolderPath = TSHelpers.effectiveEvalResultFolder(DefaultPaths.evalResultFolder,
                    canonicalDescription);

            int parallelism = 16;
            int repeats = 3;

            String sdfPath = "/home/schaefer/data/promotion/projekte/spp-big_data/00_struclust/datasets/chemdb/isomer3d.all.sdf";
            List<MoleculeGraph> graphs = GraphIO.sdfToMoleculeGraphs(sdfPath, true, true, 5000000);

            ArrayList<GraphContainer<String, String, MoleculeGraph>> graphContainers = TSHelpers
                    .graphs2filteredGCs(graphs);
            TSHelpers.logDatasetSummary(false, graphContainers);

            SharedMemorySCC<String, String, MoleculeGraph> conf = conf(new MoleculeGraphFactory(), parallelism);
            SharedMemorySC<String, String, MoleculeGraph> clust = new SharedMemorySC<>(graphContainers, conf);

            String folderPath = FsOperations.createUniqueFolder(baseFolderPath, "", 3);

            ArrayList<EvalModule<List<Cluster<String, String, MoleculeGraph>>>> modules = new ArrayList<>();
            modules.add(new PerfEM<>(clust));
            modules.add(new ClustSizeEM<>(clust));
            modules.add(new RepSizeEM<>(clust));
            modules.add(new RepToMemberSizeEM<>(clust));
            modules.add(new NumClustersEM<>(clust));
            modules.add(new RTMSToNumClustersEM<>(clust, new double[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 }));
            modules.add(new CoverageEM<>(clust));
            modules.add(new GmlClustSaverEM<>(clust, FilenameUtils.concat(folderPath, "gml"),
                    FilenameUtils.getBaseName(folderPath), true, false, false));

            SeriesEM<List<Cluster<String, String, MoleculeGraph>>> series = new SeriesEM<>(modules);

            FileSaverEMM saver = new FileSaverEMM(FilenameUtils.concat(folderPath, "eval/eval.txt"), true, false);
            LinkedList<EvalMetaModule> mm = new LinkedList<>();
            mm.push(new RepeaterEMM(repeats, false));
            MetaDataAdderEMM metaAdderDataset = new MetaDataAdderEMM("dataset", sdfPath);
            MetaDataAdderEMM datasetSize = new MetaDataAdderEMM("dataset size", Integer.toString(graphs.size()));
            mm.push(metaAdderDataset);
            mm.push(TSHelpers.hostNameMetaData());
            mm.push(datasetSize);
            mm.push(saver);
            mm.push(new StatisticsEMM());

            TSHelpers.logAndSaveMetadataInfo(canonicalDescription, conf, folderPath, metaAdderDataset, datasetSize);

            results.addAll(saver.run(mm, series));

            return results;
        }
    };

    /**
     * Configuration for the TestSetups with multiple hypothesis testing
     * correction (support counting)
     * 
     * @param gf
     *            the {@link GraphFactory} for G
     * @param parallelism
     *            the parallelism (see {@link SharedMemorySCC#parallelism}
     * @return the configuration
     */
    public static <NL, EL, G extends Graph<NL, EL>> SharedMemorySCC<NL, EL, G> conf(GraphFactory<NL, EL, G> gf,
            int parallelism) {
        SharedMemorySCC<NL, EL, G> conf = new SharedMemorySCC<>(gf, parallelism);

        FingerprintBuilder fingerprintBuilder = new FingerprintBuilder(2048);
        GraphGraphDistance<NL, EL, G> assignmentDist = new WeightedSubIsoFpGGD<>(2, conf);
        McsWskrGGD<NL, EL, G> mcsDist = new McsWskrGGD<>(false, null, new MccisKriegeColor2MSA<>());

        conf.gSize = GraphSize.edges;
        conf.supportCounting = new CorrectedBinomialSC<>(0.1, 0.5, conf.parallelism);
        conf.preClusterer = new DissimilarRepPC<>(conf, 3000, 0.01, 5, 30, 0.3, mcsDist, assignmentDist, gf,
                fingerprintBuilder);
        conf.convergence = new GenericRangeAvgCC(Ratings.wAvgRTMS2NumClust(1), 0.01, 4, 3);
        conf.maxIterations = 50;
        conf.clusterSplitting = true;
        conf.splitCriteria = new RelRepToMemberSizeSC(0.6);
        conf.splitStrategy = new RePreClusterSS<>(
                new DissimilarRepPC<>(conf, 1000, 0.05, 5, 30, 0.3, mcsDist, assignmentDist, gf, fingerprintBuilder),
                4);
        conf.splitCooldownRounds = 2;
        conf.clusterMerging = true;
        conf.mergeCriteria = new RepThresholdCountFineParallelMC<>(conf.parallelism,
                new McsMinSizeGGD<>(false, null, new MccisKriegeColor2MSDA<>(), 0.7), 0.5, 1);
        conf.mergeStrategy = new ReMineMS<>();
        conf.startupRounds = 3;
        conf.assignToNewClusterOnMaxDissimilarity = true;
        conf.minimalClusterSize = 5;
        conf.cgDist = new AccCGD<>(assignmentDist, AccFunction.AVG);

        RandomGMConf<NL, EL, G> randGmConf = new RandomGMConf<>(conf.gFactory, conf);
        randGmConf.useSupSizeProduct = false;
        randGmConf.parallel = false;
        randGmConf.supCallsEstimator = new ConservativeSCE<>();
        conf.graphMiner = new RandomGM<>(randGmConf);

        conf.minSupStrategy = new RelRepToMemberSizeMSS<>(0.0, 0.4, 0.9, 0.99);
        conf.repCount = 3;
        conf.newRepCount = 25;
        conf.repSelectionStrategy = new DiscriminativeSupSizeRSS<>(conf, false, true);
        conf.defFpBuilder = fingerprintBuilder;

        return conf;
    }

    /**
     * Configuration for the TestSetups with exact support counting
     * 
     * @param gf
     *            the {@link GraphFactory} for G
     * @param parallelism
     *            the parallelism (see {@link SharedMemorySCC#parallelism}
     * @return the configuration
     */
    public static <NL, EL, G extends Graph<NL, EL>> SharedMemorySCC<NL, EL, G> confExactSupportCounting(
            GraphFactory<NL, EL, G> gf, int parallelism) {
        SharedMemorySCC<NL, EL, G> conf = new SharedMemorySCC<>(gf, parallelism);

        FingerprintBuilder fingerprintBuilder = new FingerprintBuilder(2048);
        GraphGraphDistance<NL, EL, G> assignmentDist = new WeightedSubIsoFpGGD<>(2, conf);
        McsWskrGGD<NL, EL, G> mcsDist = new McsWskrGGD<>(false, null, new MccisKriegeColor2MSA<>());

        conf.gSize = GraphSize.edges;
        conf.supportCounting = new SimpleSC<>(conf.parallelism);
        conf.preClusterer = new DissimilarRepPC<>(conf, 3000, 0.01, 5, 30, 0.3, mcsDist, assignmentDist, gf,
                fingerprintBuilder);
        conf.convergence = new GenericRangeAvgCC(Ratings.wAvgRTMS2NumClust(1), 0.01, 4, 3);
        conf.maxIterations = 50;
        conf.clusterSplitting = true;
        conf.splitCriteria = new RelRepToMemberSizeSC(0.6);
        conf.splitStrategy = new RePreClusterSS<>(
                new DissimilarRepPC<>(conf, 1000, 0.05, 5, 30, 0.3, mcsDist, assignmentDist, gf, fingerprintBuilder),
                4);
        conf.splitCooldownRounds = 2;
        conf.clusterMerging = true;
        conf.mergeCriteria = new RepThresholdCountFineParallelMC<>(conf.parallelism,
                new McsMinSizeGGD<>(false, null, new MccisKriegeColor2MSDA<>(), 0.7), 0.5, 1);
        conf.mergeStrategy = new ReMineMS<>();
        conf.startupRounds = 3;
        conf.assignToNewClusterOnMaxDissimilarity = true;
        conf.minimalClusterSize = 5;
        conf.cgDist = new AccCGD<>(assignmentDist, AccFunction.AVG);

        RandomGMConf<NL, EL, G> randGmConf = new RandomGMConf<>(conf.gFactory, conf);
        randGmConf.useSupSizeProduct = false;
        randGmConf.parallel = false;
        randGmConf.supCallsEstimator = new ConservativeSCE<>();
        conf.graphMiner = new RandomGM<>(randGmConf);

        conf.minSupStrategy = new RelRepToMemberSizeMSS<>(0.0, 0.4, 0.9, 0.99);
        conf.repCount = 3;
        conf.newRepCount = 25;
        conf.repSelectionStrategy = new DiscriminativeSupSizeRSS<>(conf, false, true);
        conf.defFpBuilder = fingerprintBuilder;

        return conf;
    }

}
