package struclust.eval.util;

import java.util.Collection;

import struclust.Cluster;
import struclust.eval.comparison.ContingencyMatrix;

/**
 * Convenience methods to calculate information theoretic clustering measures
 * 
 * @author Till Schäfer
 */
public class InformationTheoreticClusteringMeasures {
    /**
     * Calculates the entropy of a clustering (regarding its size distribution)
     * 
     * <br>
     * from: Nguyen Xuan Vinh, Julien Epps, and James Bailey. 2010. Information
     * Theoretic Measures for Clusterings Comparison: Variants, Properties,
     * Normalization and Correction for Chance. J. Mach. Learn. Res. 9999
     * (December 2010), 2837-2854.
     * 
     * @param clustering
     *            the clustering
     * @return the entropy
     */
    public static double entropy(Collection<Cluster<?, ?, ?>> clustering) {
        int overallSize = clustering.stream().mapToInt(c -> c.size()).sum();
        double retVal = 0;

        for (Cluster<?, ?, ?> cluster : clustering) {
            int ai = cluster.size();
            double normalisedAi = (double) ai / overallSize;

            retVal += normalisedAi * Math.log(normalisedAi);
        }

        return -retVal;
    }

    /**
     * Calculates the joint entropy between two clusterings (regarding its size
     * distribution)
     * 
     * <br>
     * from: Nguyen Xuan Vinh, Julien Epps, and James Bailey. 2010. Information
     * Theoretic Measures for Clusterings Comparison: Variants, Properties,
     * Normalization and Correction for Chance. J. Mach. Learn. Res. 9999
     * (December 2010), 2837-2854.
     * 
     * @param matrix
     *            the {@link ContingencyMatrix}
     * @return the joint entropy
     */
    public static double jointEntropy(ContingencyMatrix<?, ?, ?> matrix) {
        int overallSize = matrix.numElements();
        double retVal = 0;

        for (int i = 0; i < matrix.numClusters1(); i++) {
            for (int j = 0; j < matrix.numClusters2(); j++) {
                double normalizedNij = (double) matrix.getContingencyValue(i, j) / overallSize;
                if (normalizedNij != 0) {
                    retVal += normalizedNij * Math.log(normalizedNij);
                }
            }
        }

        return -retVal;
    }

    /**
     * Calculates the conditional entropy (regarding its size distribution)
     * 
     * <br>
     * from: Nguyen Xuan Vinh, Julien Epps, and James Bailey. 2010. Information
     * Theoretic Measures for Clusterings Comparison: Variants, Properties,
     * Normalization and Correction for Chance. J. Mach. Learn. Res. 9999
     * (December 2010), 2837-2854.
     * 
     * @param matrix
     *            the {@link ContingencyMatrix}
     * @param invertCondition
     *            false=H(C1|C2) and true=H(C2|C1) -> see:
     *            {@link ContingencyMatrix}
     * @return the conditional entropy
     */
    public static double conditionalEntropy(ContingencyMatrix<?, ?, ?> matrix, boolean invertCondition) {
        double overallSize = matrix.numElements();
        double retVal = 0;

        for (int i = 0; i < matrix.numClusters1(); i++) {
            for (int j = 0; j < matrix.numClusters2(); j++) {
                double nij = matrix.getContingencyValue(i, j);
                double clusterSize = invertCondition ? matrix.getClusterSize1(i) : matrix.getClusterSize2(j);

                assert clusterSize != 0;
                if (nij != 0) {
                    retVal += nij / overallSize * Math.log((nij / overallSize) / (clusterSize / overallSize));
                }
            }
        }

        return -retVal;
    }

    /**
     * Calculates the mutual information (regarding its size distribution)
     * 
     * <br>
     * from: Nguyen Xuan Vinh, Julien Epps, and James Bailey. 2010. Information
     * Theoretic Measures for Clusterings Comparison: Variants, Properties,
     * Normalization and Correction for Chance. J. Mach. Learn. Res. 9999
     * (December 2010), 2837-2854.
     * 
     * @param matrix
     *            the {@link ContingencyMatrix}
     * @return the mutual information
     */
    public static double mutualInformation(ContingencyMatrix<?, ?, ?> matrix) {
        double overallSize = matrix.numElements();
        double retVal = 0;

        for (int i = 0; i < matrix.numClusters1(); i++) {
            for (int j = 0; j < matrix.numClusters2(); j++) {
                double nij = matrix.getContingencyValue(i, j);
                double multClusterSizes = matrix.getClusterSize1(i) * matrix.getClusterSize2(j);

                assert multClusterSizes != 0;
                if (nij != 0) {
                    retVal += nij / overallSize
                            * Math.log((nij / overallSize) / (multClusterSizes / Math.pow(overallSize, 2)));
                }
            }
        }

        return retVal;
    }
}
