package hsahn.algorithm.graph.coloring;

import hsahn.graph.Graph;
import hsahn.graph.Graph.Vertex;
import hsahn.graph.properties.VertexArray;

import java.util.ArrayList;

/**
 * Computes a graph coloring, i.e. a color is assigned to
 * each vertex such that all adjacent vertices have a different 
 * colors.
 */
public abstract class GraphColorer {

	/**
	 * Computes a color assignment and the number of color classes used.
	 * @param g a graph that is colored
	 * @param c an assignment of vertices to color classes by an integer is stored 
	 * in this parameter
	 * @return the number of color classes, i.e. an upper bound for the chromatic
	 * number which is tight when the implementing GraphColorer is exact.
	 */
	public abstract int color(Graph g, VertexArray<Integer> c);

	/**
	 * Computes a color assignment.
	 * @param g a graph
	 * @return an assignment of vertices to color classes identified by an integer
	 */
	public VertexArray<Integer> color(Graph g) {
		VertexArray<Integer> c = new VertexArray<Integer>(g);
		color(g, c);
		return c;
	}

	
	/**
	 * Computes a color class partition of the vertices.
	 * @param g a graph
	 * @return a list of color classes, each containing a list of the vertices in the class
	 */
	public ArrayList<ArrayList<Vertex>> colorClassPartition(Graph g) {
		
		ArrayList<ArrayList<Vertex>> r = new ArrayList<ArrayList<Vertex>>();
		VertexArray<Integer> c = color(g);

		for (Vertex v : g.vertices()) {
			int classId = c.get(v);
			while (r.size()<= classId) {
				r.add(new ArrayList<Vertex>());
			}
			r.get(classId).add(v);
		}
		
		return r;
	}
}
