package hsahn.comparison.kernel;

import hsahn.concepts.Transformation;

/**
 * Interface for classes merging the vectorial values in a gram matrix of composite 
 * kernels into a single kernel value.
 * 
 * @author kriege
 *
 */
public interface CompositeMerger extends Transformation<double[][][], double[][]> {
	
	public String getID();
	

	/**
	 * Computes element-wise normalized kernel values and adds them up to a
	 * single kernel value.
	 * 
	 * @author kriege
	 *
	 */
	public static class CompositeNormalization implements CompositeMerger {
	
		@Override
        public double[][] transform(double[][][] in) {
			int n = in.length;
			double[][] out = new double[n][n];
			for (int i=0; i<n; i++) {
				for (int j=i; j<n; j++) {
					out[i][j] = out[j][i] = compute(in[i][j], in[i][i], in[j][j]); 
				}
			}
			return out;
		}
		
		public double compute(double[] k, double[] selfsim1, double[] selfsim2) {
			int n = k.length;
			double r = 0;
			for (int i=0; i<n; i++) {
				// this should only happen with zero-extended components and k[i] == 0 
				if (selfsim1[i] == 0.0d || selfsim2[i]== 0.0d) continue; 
				r += k[i]/Math.sqrt(selfsim1[i]*selfsim2[i]);
			}
			return r;
		}

		@Override
        public String getID() {
			return "CN";
		}
		
	}
	
	/**
	 * Adds up the elements in the kernel value vectors to a single kernel value.
	 * 
	 * @author kriege
	 *
	 */
	public static class Addition implements CompositeMerger {

		@Override
        public double[][] transform(double[][][] in) {
			int n = in.length;
			double[][] out = new double[n][n];
			for (int i=0; i<n; i++) {
				for (int j=i; j<n; j++) {
					double[] k = in[i][j];
					double v = 0;
					for (double d : k) v += d;
					out[i][j] = out[j][i] = v; 
				}
			}
			return out;
		}

		@Override
        public String getID() {
			return "AD";
		}
		
	}

}