package hsahn.comparison.kernel.graph;

import hsahn.algorithm.graph.vectorization.EdgeVectorizer;
import hsahn.comparison.kernel.ExplicitMappingKernel;
import hsahn.comparison.kernel.Kernel;
import hsahn.comparison.kernel.basic.DiracKernel;
import hsahn.concepts.TransformationTools;
import hsahn.datastructure.SparseFeatureVector;
import hsahn.datastructure.Triple;
import hsahn.graph.LGraph;

import java.util.ArrayList;
import java.util.Collection;

/**
 * Compares the edges of two graphs taking the two vertex labels as
 * well as the edge label into account. Note: For two edges (u1, v1)
 * and (u2, v2) both mapping are considered, i.e. u1 to u2, v1 to v2
 * and u1 to v2, v1 to u2.
 * 
 * k(G1, G2) = \sum_(u1,v1) \sum_(u2,v2) kv(u1,u2)*kv(v1,v2)*ke((u1,v1),(u2,v2))
 * 
 * where ke(e1, e2) = 0 if e1 not in G1 or e2 not in G2.
 * 
 * Note: This implementation internally computes a random walk kernel 
 * with fixed length 1. This is equivalent to the EdgeKernel defined 
 * above and has the advantage that the product graph caches the results
 * of vertex and edge kernel evaluations.
 * 
 * @author kriege
 *
 * @param <V> vertex label type
 * @param <E> edge label type
 */
public class EdgeKernel<V, E> extends FixedLengthRandomWalkKernel<V, E> implements 
	ExplicitMappingKernel<LGraph<V, E>, SparseFeatureVector<Triple<V,E,V>>>
{
	
	public EdgeKernel(Kernel<? super E> edgeKernel, Kernel<? super V> vertexKernel) {
		super(1, edgeKernel, vertexKernel);
	}

// Note: This was found to be slower in general than the method using the
// a product graph in combination with a fixed length random walk.
// However, for sparse unlabeled graphs and dense vertex kernels that
// can be computed efficiently the other implementation might be 
// preferable.
//
//	public double evaluate(LabeledGraph<V, E> lg1, LabeledGraph<V, E> lg2) {
//		computations++;
//		
//		double totalValue = 0;
//		
//		long startTime = System.nanoTime();		
//		Graph g1 = lg1.getGraph();
//		Graph g2 = lg2.getGraph();
//		VertexArray<V> va1 = lg1.getVertexLabel();
//		VertexArray<V> va2 = lg2.getVertexLabel();
//		EdgeArray<E> ea1 = lg1.getEdgeLabel();
//		EdgeArray<E> ea2 = lg2.getEdgeLabel();
//		
//		for (Edge e1 : g1.edges()) {
//			for (Edge e2 : g2.edges()) {
//				Vertex u1 = e1.getFirstVertex();
//				Vertex v1 = e1.getSecondVertex();
//				Vertex u2 = e2.getFirstVertex();
//				Vertex v2 = e2.getSecondVertex();
//				double edgeValue = edgeKernel.evaluate(ea1.get(e1), ea2.get(e2));
//				// first mapping
//				double val = edgeValue;
//				val *= vertexKernel.evaluate(va1.get(u1), va2.get(u2));
//				val *= vertexKernel.evaluate(va1.get(v1), va2.get(v2));
//				totalValue += val;
//				// second mapping
//				val = edgeValue;
//				val *= vertexKernel.evaluate(va1.get(u1), va2.get(v2));
//				val *= vertexKernel.evaluate(va1.get(v1), va2.get(u2));
//				totalValue += val;
//			}
//		}
//		kernelComputationTime += System.nanoTime() - startTime;
//		
//		return totalValue;
//	}

	@Override
    public ArrayList<SparseFeatureVector<Triple<V, E, V>>> computeExplicitMapping(Collection<? extends LGraph<V, E>> graphs) {
		if (!(pgb.vertexKernel instanceof DiracKernel) || !(pgb.edgeKernel instanceof DiracKernel)) {
			throw new IllegalStateException("Explicit mapping requires both, vertex and edge kernel, to be dirac kernels!");
		}
		
		EdgeVectorizer<V, E> ev = new EdgeVectorizer<V, E>();
		
		return TransformationTools.transformAll(ev, graphs);
	}
	
	@Override
    public String getID() {
		return "E_"+pgb.getVertexKernel().getID()+"_"+pgb.getEdgeKernel().getID();
	}
	
}
