package hsahn.comparison.kernel.graph;

import hsahn.comparison.kernel.Kernel;
import hsahn.comparison.kernel.graph.ProductGraph.PEdge;
import hsahn.comparison.kernel.graph.ProductGraph.PVertex;
import hsahn.graph.Graph.Edge;

/**
 * Computes a random walk kernel taking all walks with the specified length 
 * into account, i.e. \sum_i,j E^k.
 * 
 * @author kriege
 *
 * @param <V>
 * @param <E>
 */
// TODO count all path up to size k and implement weight function support
// TODO for certain weight functions the kernel can be implemented in a more 
// efficient way, for example, when only walks of size k should be taken into 
// account directly computing E^k is much more efficient than multiplying with
// E  k times.
// Possible better solutions:
//   - Diagonalisation (by eigenvalue computation)
//   - http://www.jstor.org/pss/30037495
public class FixedLengthRandomWalkKernel<V,E> extends RandomWalkKernel<V, E> {
	
	int k;
	
	/**
	 * @param k maximum size of the random walk
	 * @param edgeKernel
	 * @param vertexKernel
	 */
	public FixedLengthRandomWalkKernel(int k, Kernel<? super E> edgeKernel, Kernel<? super V> vertexKernel) {
		super(edgeKernel, vertexKernel);
		this.k = k;
	}

	@Override
	double computeKernel(ProductGraph pg) {
		
		if (k == 0) {
			double r = 0;
			for (PVertex v : pg.vertices()) {
				r += v.getWeight();
			}
			return r;
		}
		
		if (k == 1) {
			// we do not need to create a matrix in this case,
			// it is sufficient iterate over all edges, which is
			// preferable especially for sparse product graphs
			double r = 0;
			for (PEdge e : pg.edges()) {
				r += e.getWeight() * 
						e.getFirstVertex().getWeight() *
						e.getSecondVertex().getWeight();
			}
			// multiply by two since each edge allows
			// two walks.
			return 2*r;
		}
		
		// k > 1
		int n = pg.getVertexCount();
		double[][] W = pg.getEdgeWeights();
		
		// encode vertex kernel results
		double[] vw = pg.getVertexWeights();
		for (Edge e : pg.edges()) {
			int iU = e.getFirstVertex().getIndex();
			int iV = e.getSecondVertex().getIndex();
			W[iU][iV] *= (vw[iU] * vw[iV]); 
			W[iV][iU] = W[iU][iV];
		}
		
		// walks of length 1 are stored in W
		// clone W (clone method does not work for 2d arrays)
		double[][] R = new double[n][n];
		for (int i=0; i<n; i++)	R[i] = W[i].clone();
		
		double[][] tmp = new double[n][n];
		for (int i=1; i<k; i++) {
			mult(tmp, R, W, n);
			double[][] tmp2 = R;
			R = tmp;
			tmp = tmp2;
		}

		double r = 0;
		for (int i=0; i<n; i++) {
			for (int j=0; j<n; j++) {
				r += R[i][j];
			}
		}

		return r;
	}
	
	
	/**
	 * Computes the product P*F and stores the result in R. Implementation 
	 * requires time O(n^3). Note: All matrices must be symmetric!
	 */
	private static void mult(double[][] R, double[][] P, double[][] F, int n) {
		
		for (int i=0; i<n; i++) {
			for (int j=i; j<n; j++) {
				double sum = 0d;
				for (int k=0; k<n; k++) {
					sum += (P[i][k] * F[k][j]);
				}
				R[i][j] = R[j][i] = sum;
			}
		}
	}
	
//	static void printRowSum(double[][] d) {
//		double total = 0;
//		for (double[] row : d) {
//			double val = 0;
//			for (double e : row) {
//				val += e;
//			}
//			System.out.println(val);
//			total += val;
//		}
//		System.out.println("full sum "+total);
//	}
//	
//	static void printFullSum(double[][] d) {
//		double total = 0;
//		for (double[] row : d)
//			for (double e : row)
//				total += e;
//		System.out.println("full sum "+total);
//	}
//	
//	static void printMatrix(double[][] d) {
//		for (int i=0; i<d.length; i++) {
//			for (int j=0; j<d[i].length; j++) {
//				System.out.print((int)d[i][j]+" ");
//			}
//			System.out.println();
//		}
//	}
//	
//	public static void main(String[] args) {
//		double[][] W = {{0,1,1},{1,0,1},{1,1,0}};
//		double[][] R = new double[W.length][W.length];
//		mult(R, W, W, W.length);
//		printMatrix(W);
//		System.out.println();
//		printMatrix(R);
//	}
	
	@Override
    public String getID() {
		return "FLRW_"+k+"_"+pgb.getVertexKernel().getID()+"_"+pgb.getEdgeKernel().getID();
	}

}
