package hsahn.comparison.kernel.graph;

import hsahn.comparison.kernel.Kernel;
import hsahn.comparison.kernel.graph.ProductGraph.PEdge;
import hsahn.comparison.kernel.graph.ProductGraph.PVertex;

/**
 * Implementation of a fixed-length random walk kernel especially suitable for
 * sparse product graphs. A walk (v0, e0, v1, ..., en-1, vn) contributes weight
 * vk(v0) * ek(e0) * vk(v1) * ... * ek(en-1) + vk(vn) to the kernel value, where
 * vertices and edges of the product graph correspond to matched vertices and
 * edges of the two factor graphs under comparison. Please note that this
 * implementation is equivalent to the {@link FixedLengthRandomWalkKernel} when
 * the vertex kernel is a Dirac kernel, but differs since inner vertices of the
 * product graph only contribute once to the weight of a walk.
 * 
 * @author kriege
 *
 * @param <V>
 *            vertex label type
 * @param <E>
 *            edge label type
 */
public class FixedLengthRandomWalkKernel2<V, E> extends RandomWalkKernel<V, E> {

    int k;

    /**
     * @param k
     *            maximum size of the random walk
     * @param edgeKernel
     * @param vertexKernel
     */
    public FixedLengthRandomWalkKernel2(int k, Kernel<? super E> edgeKernel, Kernel<? super V> vertexKernel) {
        super(edgeKernel, vertexKernel);
        this.k = k;
    }

    @Override
    double computeKernel(ProductGraph pg) {

        // no need to copy values in this case
        if (k == 0) {
            double r = 0;
            for (PVertex v : pg.vertices()) {
                r += v.getWeight();
            }
            return r;
        }

        double[] weights = new double[pg.getVertexCount()];
        double[] newWeights = new double[pg.getVertexCount()];
        for (PVertex v : pg.vertices()) {
            weights[v.getIndex()] = v.getWeight();
        }

        for (int i = 1; i <= k; i++) {
            // compute new weights
            for (PVertex v : pg.vertices()) {
                newWeights[v.getIndex()] = computeNewWeight(v, weights);
            }
            // swap weight arrays
            double[] tmp = weights;
            weights = newWeights;
            newWeights = tmp;
        }

        double r = 0;
        for (PVertex v : pg.vertices()) {
            r += weights[v.getIndex()];
        }

        return r;
    }

    private double computeNewWeight(PVertex v, double[] weights) {
        double r = 0;
        for (PEdge e : v.edges()) {
            PVertex w = e.getOppositeVertex(v);
            // TODO edge vertex product can be computed once and then stored
            // use a directed graph as product graph
            r += weights[w.getIndex()] * e.getWeight() * v.getWeight();
        }

        return r;
    }

    @Override
    public String getID() {
        return "FLRWTWO_" + k + "_" + pgb.getVertexKernel().getID() + "_" + pgb.getEdgeKernel().getID();
    }

}
