package hsahn.comparison.kernel.graph;

import hsahn.comparison.kernel.Kernel;

/**
 * Base class for random walk kernels inspired by the approach proposed by 
 * G�rtner et al. (2003), which originally requires vertex and edge labels 
 * to be identical.
 * This version is extended to arbitrary vertex and edge kernels by using
 * a weight matrix instead of the adjacency matrix of the directed product
 * graph, such that an entry a_i,j represents the weight of the edge (v_i, v_j).
 * Each edge ((u1, u2), (v1, v2)) passed by a random walk in the product graph 
 * is considered to have weight c(u1,u2) * c(v1,v2) * c((u1,v1),(u2,v2))
 * determined by the provided vertex and edge kernel, respectively. This
 * straightforward extension corresponds to the approach taken in Borgwardt (2005).
 * Note that each inner vertex of a random walk contributes twice to the
 * value of the walk, while the end points only appear once in the product.
 * If the specified vertex and edge kernels are {@link DiracKernel}s, this 
 * kernel is equivalent to kernel proposed by G�rtner et al. (2003).
 * 
 * @author kriege
 *
 * @param <V> vertex label type
 * @param <E> edge label type
 */
// TODO javadoc needs update, move this description to subclass
// when Tsudo kernel is implemented
public abstract class RandomWalkKernel<V,E> extends ProductGraphKernel<V, E, ProductGraph> {
	
	/**
	 * @param edgeKernel edge comparison function used to compute weight matrix
	 * @param vertexKernel vertex comparison function used to compute weight matrix
	 */
	public RandomWalkKernel(Kernel<? super E> edgeKernel, Kernel<? super V> vertexKernel) {
		super(new WeightedDirectProductGraphBuilderDense<V, E>(edgeKernel, vertexKernel));
	}

	/**
	 * {@inheritDoc}
	 *  
	 * Note: Implementing classes should use both, vertex and edge
	 * weights, to compute a weight matrix, where each element
	 * encodes the product of an edge kernel result and two vertex
	 * kernel results.
	 */
	// TODO javadoc needs update
	@Override
    abstract double computeKernel(ProductGraph pg);
	

}
