package hsahn.comparison.kernel.graph;

import hsahn.algorithm.graph.vectorization.EdgeVectorizer;
import hsahn.algorithm.shortestpath.SPTools;
import hsahn.comparison.kernel.ExplicitMappingKernel;
import hsahn.comparison.kernel.Kernel;
import hsahn.comparison.kernel.basic.DiracKernel;
import hsahn.concepts.Transformation;
import hsahn.datastructure.SparseFeatureVector;
import hsahn.datastructure.Triple;
import hsahn.graph.LGraph;

import java.util.ArrayList;
import java.util.Collection;

/**
 * Computes the shortest path kernel (Borgwardt, 2005). In addition this 
 * class supports fast computation for large datasets by explicit mapping
 * provided that vertex and edge kernels both are {@link DiracKernel}s.
 * 
 * Please note that each call of {@link #compute(LGraph, LGraph)}
 * computes the {@link FloydWarshallTransformation} for both graphs. A 
 * significant runtime improvement can be achieved using the following 
 * equivalent code, which only computes the transformation once for each 
 * graph:
 * 
 * <code>
 * 		Collection<LabeledGraph<String,String>> graphs = [graphs];
 *		List<LabeledGraph<String, Integer>> transformedGraphs = 
 *				TransformationTools.transformAll(new FloydWarshallTransformation<String>(), graphs);
 *		EdgeKernel<String, Integer> ek = 
 *				new EdgeKernel<String, Integer>([edge kernel], [vertex kernel]);
 *		ek.evaluate(transformedGraphs.get([x]), transformedGraphs.get([y]));
 * </code>
 * 
 * @author kriege
 *
 * @param <V> vertex label type
 * @param <E> edge label type; note that edge labels are completely ignored by this kernel
 */
public class ShortestPathKernel<V, E> implements 
	Kernel<LGraph<V, E>>,
	ExplicitMappingKernel<LGraph<V, E>, SparseFeatureVector<Triple<V, Integer, V>>> 
{
	
	EdgeKernel<V, Integer> ek;
	FloydWarshallTransformation<V> fwt;
	Kernel<? super Integer> edgeKernel;
	Kernel<? super V> vertexKernel;
	
	/**
	 * Instantiates a new shortest path kernel.
	 * 
	 * @param edgeKernel kernel used to compare shortest path distances
	 * @param vertexKernel kernel used to compare vertex labels
	 */
	public ShortestPathKernel(Kernel<? super Integer> edgeKernel, Kernel<? super V> vertexKernel) {
		this.edgeKernel = edgeKernel;
		this.vertexKernel = vertexKernel;
		fwt = new FloydWarshallTransformation<V>();
		ek = new EdgeKernel<V, Integer>(edgeKernel, vertexKernel);
	}

	@Override
    public double compute(LGraph<V, E> g1, LGraph<V, E> g2) {
		LGraph<V, Integer> fwtG1 = fwt.transform(g1);
		LGraph<V, Integer> fwtG2 = fwt.transform(g2);
		
		return ek.compute(fwtG1, fwtG2);
	};

	/**
	 * Note: Vertex and edge labels are compared by Object.equals().
	 */
	@Override
    public ArrayList<SparseFeatureVector<Triple<V, Integer, V>>> computeExplicitMapping(Collection<? extends LGraph<V, E>> graphs) {
		if (!(vertexKernel instanceof DiracKernel) || !(edgeKernel instanceof DiracKernel)) {
			throw new IllegalStateException("Explicit mapping requires both, vertex and edge kernel, to be dirac kernels!");
		}
		
		int n = graphs.size();
		ArrayList<SparseFeatureVector<Triple<V, Integer, V>>> result = new ArrayList<SparseFeatureVector<Triple<V,Integer,V>>>(n);

		EdgeVectorizer<V, Integer> ev = new EdgeVectorizer<V, Integer>();
		for (LGraph<V, ?> lg : graphs) {
			LGraph<V, Integer> fwtLG = fwt.transform(lg);
			result.add(ev.transform(fwtLG));
		}
		
		return result;
	}
	
	/**
	 * Computes the Floyd Warshall transformation of a graph, i.e. a
	 * complete graph with edge labels representing shortest path
	 * distances.
	 * 
	 * @author kriege
	 *
	 * @param <V> the preserved vertex label type
	 */
	public static class FloydWarshallTransformation<V> implements Transformation<LGraph<V,?>, LGraph<V, Integer>> {

		@Override
        public LGraph<V, Integer> transform(LGraph<V, ?> in) {
			return SPTools.getAPSPTransformation(in);
		}
		
	}
	
	@Override
    public String toString() {
		return "SPKernel+\n"+ek.toString();
	}

	@Override
    public String getID() {
		return "SP_"+vertexKernel.getID()+"_"+edgeKernel.getID();
	}
}
