package hsahn.comparison.kernel.graph;

import hsahn.comparison.kernel.Kernel;
import hsahn.graph.Graph.Vertex;
import hsahn.graph.LGraph;

import java.util.ArrayList;

/**
 * Computes a weighted product graph and provides its weighted
 * adjacency matrix.
 * A single instance of this class can be used to subsequently
 * compute multiple product graphs avoiding the repeated allocation
 * and GCing of variables.
 *    
 * @author kriege
 *
 * @param <V> vertex type of the input graphs
 * @param <E> edge type of the input graphs
 */
public abstract class WeightedProductGraph<V,E> {
	
	Kernel<? super E> eKernel;
	Kernel<? super V> vKernel;
	
	ArrayList<Vertex> compG1 = new ArrayList<Vertex>();
	ArrayList<Vertex> compG2 = new ArrayList<Vertex>();
	int n;
	int m;
	double[][] edgeWeights;
	double[] vertexWeights;

	
	WeightedProductGraph(Kernel<? super E> edgeKernel, Kernel<? super V> vertexKernel) {
		this.eKernel = edgeKernel;
		this.vKernel = vertexKernel;
	}
	
	public void setVertexKernel(Kernel<? super V> vertexKernel) {
		vKernel = vertexKernel;		
	}
	
	public Kernel<? super V> getVertexKernel() {
		return vKernel;
	}
	
	public void setEdgeKernel(Kernel<? super E> edgeKernel) {
		eKernel = edgeKernel;		
	}

	public Kernel<? super E> getEdgeKernel() {
		return eKernel;
	}
	
	public int getVertexCount() {
		return n;
	}
	
	public int getEdgeCount() {
		return m;
	}

	public double[] getVertexWeights() {
		return vertexWeights;
	}
	
	
	public double[][] getEdgeWeights() {
		return edgeWeights;
	}
	
	/**
	 * Computes the weighted product graph of the two given graphs. The 
	 * resulting graph can be retrieved via getVertexWeights(), getEdgeWeights(),
	 * getVertexCount() and getEdgeCount(). 
	 * 
	 * Note: Do not use getVertexWeighths().length or getEdgeWeights().length!
	 * These values may be larger than getVertexCount() and the additional
	 * elements may be filled with out-dated values.
	 * 
	 * @param lg1 first labeled graph
	 * @param lg2 second labeled graph
	 */
	public abstract void compute(LGraph<V,E> lg1, LGraph<V,E> lg2);
	
	
	protected double[] initVertexValue(int size, boolean fillZero) {
		if (vertexWeights != null && vertexWeights.length >= size) {
			if (fillZero) {
				for (int i=0; i<size; i++) {
					vertexWeights[i] = 0d;
				}
			}
			return vertexWeights;
		} else {
			return new double[size];
		}
	}

	protected double[][] initEdgeValue(int size, boolean fillZero) {
		if (edgeWeights != null && edgeWeights.length >= size) {
			if (fillZero) {
				for (int i=0; i<size; i++) {
					for (int j=0; j<size; j++) {
						edgeWeights[i][j] = 0d;
					}
				}
			}
			return edgeWeights;
		} else {
			return new double[size][size];
		}
	}
	
	/**
	 * Releases memory allocated by this product graph.
	 */
	public void clear() {
		vertexWeights = null;
		edgeWeights = null;
		compG1.clear();
		compG2.clear();
	}
	
	
//	public static String matrixToString(double[][] m) {
//		NumberFormat numberFormat =	NumberFormat.getNumberInstance(Locale.getDefault());
//		{
//			numberFormat.setMaximumFractionDigits(9);
//		}
//			
//		StringBuffer sb = new StringBuffer();
//		for (int i=0; i<m.length; i++) {
//			for (int j=0; j<m.length; j++) {
//				sb.append("\t"+m[i][j]);
////				sb.append("\t"+(m[i][j]>0?"1":"0"));
//			}
//			sb.append("\n");
//		}
//		return sb.toString();
//	}
//
//	public static int nonZeroCount(double[][] m) {
//		int c=0;
//		for (double d[] : m) {
//			for (double d2 : d) {
//				if (d2!=0) c++;
//			}
//		}
//		return c/2;
//	}

}
