package hsahn.comparison.kernel.graph;

import hsahn.algorithm.graph.isomorphism.labelrefinement.VertexLabelConverter;
import hsahn.algorithm.graph.isomorphism.labelrefinement.WeisfeilerLehmanRefiner;
import hsahn.algorithm.graph.vectorization.EdgeVectorizer;
import hsahn.comparison.kernel.Kernel;
import hsahn.comparison.kernel.basic.DiracKernel;
import hsahn.comparison.kernel.graph.ShortestPathKernel.FloydWarshallTransformation;
import hsahn.concepts.TransformationTools;
import hsahn.datastructure.SparseFeatureVector;
import hsahn.datastructure.Triple;
import hsahn.graph.LGraph;

import java.util.ArrayList;
import java.util.Collection;

/**
 * Implementation of the Weisfeiler-Lehman Shortest Path Kernel (Shervashidze,
 * 2011) supporting 1:1 kernel evaluation as well as explicit mapping into
 * feature space for fast n:n computation.
 * 
 * This kernel is equivalent to the following kernel <code>wlspk</code>: <code>
 * 		DiracKernel dk = new DiracKernel();
 * 		ShortestPathKernel<Integer> spk = new ShortestPathKernel<Integer>(dk, dk);
 * 		WeisfeilerLehmanKernel<Integer, Integer, SparseFeatureVector<Triple<Integer,Integer,Integer>>> wlspk =
 * 			new WeisfeilerLehmanKernel<Integer, Integer, SparseFeatureVector<Triple<Integer,Integer,Integer>>>(n, spk);
 * </code> However, this implementation is more efficient since the
 * Floyd-Warshall Transformation is only computed once per evaluation and not
 * for each graph in the Weisfeiler-Lehman graph sequence. This is equivalent
 * since shortest path do not change when applying a vertex label refinement.
 * 
 * Note: This class is similar in spirit to the NSPD Kernel (Costa, 2011).
 * 
 * @author kriege
 */
public class WeisfeilerLehmanShortestPathKernel<V, E> extends
        WeisfeilerLehmanKernel<V, E, Triple<Integer, Integer, Integer>> {

    FloydWarshallTransformation<Integer> fwt;
    // TODO distinguish edgeKernel and edgeLabelKernel;
    // this should be consistent in several class and var names
    Kernel<? super Integer> edgeKernel;
    EdgeKernel<Integer, Integer> ek;

    /**
     * Creates a new instance of the Weisfeiler-Lehman subtree kernel.
     * 
     * @param sequenceSize
     *            the number of graphs in the computed WL sequence
     * @param edgeKernel
     *            kernel used to compare shortest path distances
     */
    public WeisfeilerLehmanShortestPathKernel(int sequenceSize, Kernel<? super Integer> edgeKernel) {
        this(sequenceSize, edgeKernel, false);
    }

    /**
     * Creates a new instance of the Weisfeiler-Lehman subtree kernel.
     * 
     * @param sequenceSize
     *            the number of graphs in the computed WL sequence
     * @param edgeKernel
     *            kernel used to compare shortest path distances
     * @param stableLabelConverter
     *            iff true {@link #computeExplicitMapping(Collection)} can be
     *            called multiple times without loosing the label mapping. i.e.
     *            the generated explicit mappings of different runs can be
     *            compared to each other. Be aware, that this can lead to many
     *            unused attributes in the feature vectors if the same kernel
     *            object is used for muiltiple datasets that should not be
     *            compared to each other.
     */
    public WeisfeilerLehmanShortestPathKernel(int sequenceSize, Kernel<? super Integer> edgeKernel,
            boolean stableLabelConverter) {
        super(sequenceSize, null, stableLabelConverter);
        this.edgeKernel = edgeKernel;
        this.fwt = new FloydWarshallTransformation<Integer>();
        // TODO put this into base kernel, remove ek
        this.ek = new EdgeKernel<Integer, Integer>(edgeKernel, new DiracKernel());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public double[] computeComposition(LGraph<V, E> g1, LGraph<V, E> g2) {
        double[] d = new double[sequenceSize];

        // do initial label refinement
        VertexLabelConverter<E> vlc = new VertexLabelConverter<E>();
        LGraph<Integer, E> slg1 = vlc.refineGraph(g1);
        LGraph<Integer, E> slg2 = vlc.refineGraph(g2);
        vlc.clearLabelMap();

        // compute the FWT once for all iterations
        LGraph<Integer, Integer> fwtg1 = fwt.transform(slg1);
        LGraph<Integer, Integer> fwtg2 = fwt.transform(slg2);

        // first iteration
        d[0] = ek.compute(fwtg1, fwtg2);

        // additional iterations
        WeisfeilerLehmanRefiner<E> wlr = new WeisfeilerLehmanRefiner<E>();
        for (int i = 1; i < sequenceSize; i++) {
            slg1 = vlc.refineGraph(wlr.refineGraph(slg1));
            slg2 = vlc.refineGraph(wlr.refineGraph(slg2));
            vlc.clearLabelMap();
            // change the vertex labels of the fw transformed graph to
            // the vertex labels refined according to the *original* graph
            fwtg1.setVertexLabel(slg1.getVertexLabel());
            fwtg2.setVertexLabel(slg2.getVertexLabel());
            d[i] = ek.compute(fwtg1, fwtg2);
        }

        return d;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public ArrayList<SparseFeatureVector<Triple<Integer, Integer, Integer>>> computeExplicitMapping(
            Collection<? extends LGraph<V, E>> graphs) {
        if (!(edgeKernel instanceof DiracKernel)) {
            throw new IllegalStateException("Explicit mapping requires the edge kernel to be dirac kernel!");
        }

        if (stableLabelConverter) {
            vlc.setIteration(0);
        } else {
            vlc = new VertexLabelConverter<>();
        }

        int n = graphs.size();

        // assign integer label
        // using the same VertexLabelConverter assures that new labels (from
        // refinement) will
        // be assigned higher integer values, such that labels of different
        // iterations can be
        // distinguished
        ArrayList<LGraph<Integer, E>> lgs = TransformationTools.transformAll(vlc, graphs);
        if (!stableLabelConverter) {
            vlc.clearLabelMap();
        }

        // compute FWT
        ArrayList<LGraph<Integer, Integer>> fwtLgs = TransformationTools.transformAll(fwt, lgs);

        // create initial feature vectors
        EdgeVectorizer<Integer, Integer> ev = new EdgeVectorizer<Integer, Integer>();
        ArrayList<SparseFeatureVector<Triple<Integer, Integer, Integer>>> result = TransformationTools.transformAll(ev,
                fwtLgs);

        WeisfeilerLehmanRefiner<E> wlr = new WeisfeilerLehmanRefiner<E>();
        for (int i = 1; i < sequenceSize; i++) {
            // refinement
            ArrayList<LGraph<String, E>> refLgs = TransformationTools.transformAll(wlr, lgs);
            // compression
            if (stableLabelConverter) {
                vlc.setIteration(i);
            }
            lgs = TransformationTools.transformAll(vlc, refLgs);
            if (!stableLabelConverter) {
                vlc.clearLabelMap();
            }

            for (int j = 0; j < n; j++) {
                // set label to fwt graphs
                LGraph<Integer, Integer> fwtLg = fwtLgs.get(j);
                fwtLg.setVertexLabel(lgs.get(j).getVertexLabel());
                result.get(j).add(ev.transform(fwtLg));
            }
        }

        return result;
    }

    @Override
    public String toString() {
        return "Weisfeiler-Lehman shortest path kernel:\n" + super.toString();
    }

    @Override
    public String getID() {
        return "WLSP_" + sequenceSize + "_" + edgeKernel.getID();
    }
}
