package hsahn.mcs;

import hsahn.datastructure.Pair;
import hsahn.graph.Graph;
import hsahn.graph.Graph.Edge;
import hsahn.graph.Graph.Vertex;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;

public class CKCliqueFinder {

	private Graph graph;
	private int k;
	private int states;
	private long kCliqueCount;
	private CliqueHander ch;
	private ArrayList<Edge>[] adj;
	
	public CKCliqueFinder(Graph graph, int k) {
		this.graph = graph;
		this.k = k;
		this.adj = new ArrayList[graph.getVertexCount()];
		for (int i=0; i< graph.getVertexCount(); i++) {
			Vertex v = graph.getVertex(i);
			adj[i] = new ArrayList<Edge>(v.getDegree());
			for (Edge e : v.edges())
				adj[i].add(e);
			Collections.sort(adj[i], new EdgeComparator(v));
		}

	}
	
	public long computeKCliques() {
		states = 0;
		kCliqueCount = 0;
		
		ArrayList<Vertex> cand = new ArrayList<Vertex>(graph.getVertexCount());
		for (Vertex v : graph.vertices()) {
			cand.add(v);
		}

		maxClique(0, cand, new ArrayList<Vertex>());
		
		return kCliqueCount;
	}
	
		private void maxClique(int size, ArrayList<Vertex> candP, ArrayList<Vertex> candD) {
		states++;

		while (!candP.isEmpty()/* && (size + candP.size() + candD.size() >= k)*/) {
			
			Vertex v = popCandidate(candP);
			
			//kCliqueCount++;
			kCliqueCount += size+1;
			if (size+1 == k) {
				continue;
			}
			
			Pair<ArrayList<Vertex>, ArrayList<Vertex>> candPair;
			if (size == 0) {
				candPair = newIntersectionSetInit(candP, candD, v);
			} else {
				candPair = newIntersectionSet(candP, candD, v);
			}
			ArrayList<Vertex> newCandP = candPair.getFirst();
			ArrayList<Vertex> newCandD = candPair.getSecond();
			maxClique(size+1, newCandP, newCandD);
		}
	}
	
	private Vertex popCandidate(ArrayList<Vertex> cand) {
		return cand.remove(cand.size()-1);
	}
	
	private Pair<ArrayList<Vertex>,ArrayList<Vertex>> newIntersectionSet(ArrayList<Vertex> setP, ArrayList<Vertex> setD, Vertex x) {
		ArrayList<Vertex> resultP = new ArrayList<Vertex>();
		ArrayList<Vertex> resultD = new ArrayList<Vertex>();

		Iterator<Vertex> itP = setP.iterator();
		Iterator<Vertex> itD = setD.iterator();
		Vertex v = null;
		Vertex vP = itP.hasNext() ? itP.next() : null;
		Vertex vD = itD.hasNext() ? itD.next() : null;
		for (Edge e : adj[x.getIndex()]) {
			v = e.getOppositeVertex(x);
			
			while (itP.hasNext() && vP.getIndex() < v.getIndex()) {
				vP = itP.next();
			}
			if (v == vP) {
				resultP.add(v);
				continue;
			}
			while (itD.hasNext() && vD.getIndex() < v.getIndex()) {
				vD = itD.next();
			}
			if (v == vD) {
				// has c-edge
				if (isCEdge(graph, e))
					resultP.add(v);
				else
					resultD.add(v);
			}
		}
		
		return new Pair<ArrayList<Vertex>, ArrayList<Vertex>>(resultP, resultD);
	}


	
	private Pair<ArrayList<Vertex>,ArrayList<Vertex>> newIntersectionSetInit(ArrayList<Vertex> setP, ArrayList<Vertex> setD, Vertex x) {

		ArrayList<Vertex> resultP = new ArrayList<Vertex>();
		ArrayList<Vertex> resultD = new ArrayList<Vertex>();
		Vertex v = null;
		Iterator<Vertex> itP = setP.iterator();
		Vertex vP = itP.hasNext() ? itP.next() : null;
		for (Edge e : adj[x.getIndex()]) {
			v = e.getOppositeVertex(x);
			while (itP.hasNext() && vP.getIndex() < v.getIndex()) {
				vP = itP.next();
			}
			if (v == vP) {
				if (isCEdge(graph, e))
					resultP.add(e.getOppositeVertex(x));
				else
					resultD.add(e.getOppositeVertex(x));
			}
		}
		
		return new Pair<ArrayList<Vertex>, ArrayList<Vertex>>(resultP, resultD);
	}
	
	private boolean isCEdge(Graph g, Edge e) {
		if (g instanceof AssociationGraph) {
			AssociationGraph a = (AssociationGraph)g;
			return a.isCEdge(e);
		} else {
			return false;
		}
	}
	
	public int getStates() {
		return states;
	}
	
	public static interface CliqueHander {
		public void handleClique(ArrayList<Vertex> clique);
	}
	
	public static class PrintCliqueHandler implements CliqueHander {
		@Override
        public void handleClique(ArrayList<Vertex> clique) {
			
		}
	}
	
	public static class EdgeComparator implements Comparator<Edge> {
		Vertex v;
		public EdgeComparator(Vertex v) {
			this.v = v;
		}
		@Override
        public int compare(Edge o1, Edge o2) {
			int iU = o1.getOppositeVertex(v).getIndex();
			int iV = o2.getOppositeVertex(v).getIndex();
			if (iU > iV) {
				return 1;
			} else if (iU < iV) {
				return -1;
			} else {
				return 0;
			}
		}
	}
}
