package hsahn.mcs;

import hsahn.graph.Graph;
import hsahn.graph.Graph.Vertex;

import java.util.BitSet;
import java.util.HashSet;
import java.util.Set;

/**
 * Note: Clique models use an internal state and should not be reused!
 */
public interface CliqueModel {
	
	/**
	 * Reports a state in the search space that is associated with a clique.
	 * Allows the model to decide if the clique should be reported, i.e. to
	 * distinct maximal/maximum cliques
	 */
	public void processClique(Graph g, BitSet R, BitSet P);
	
	/**
	 * Allows to implement pruning techniques.
	 */
	public boolean pruneState(BitSet R, BitSet P);
	
	/**
	 * Returns the cliques found.
	 */
	public Set<Set<Vertex>> getCliques();
	
	/**
	 * Finds and stores all maximum cliques, i.e. cliques that
	 * have a maximum number of vertices.
	 * 
	 * Note: Does not work with symmetry breaking!
	 */
	public class AllMaximum implements CliqueModel {
		
		private Set<Set<Vertex>> maximumCliques = new HashSet<Set<Vertex>>();
		private int maximumCliqueSize = 0;
		
		@Override
        public void processClique(Graph g, BitSet R, BitSet P) {
			if (P.isEmpty() && R.cardinality() >= maximumCliqueSize) {
				HashSet<Vertex> clique = new HashSet<Vertex>();
				for (int i = R.nextSetBit(0); i >= 0; i = R.nextSetBit(i+1)) {
					clique.add(g.getVertex(i));
				}
				if (clique.size()>maximumCliqueSize) {
					maximumCliqueSize = clique.size();
					maximumCliques.clear();
				}
				if (clique.size()==maximumCliqueSize) {
					maximumCliques.add(clique);
				}
			}			
		}
		
		@Override
        public boolean pruneState(BitSet R, BitSet P) {
			if (R.cardinality() + P.cardinality() < maximumCliqueSize) {
				return true;
			}
//			if (maxDegree + 1 <= maxSize) {
//				//System.out.println("pruned");
//				return true;
//			}
			return false;
		}
		
		@Override
        public Set<Set<Vertex>> getCliques() {
			return maximumCliques;
		}
	}

	/**
	 * Finds and stores a single maximum cliques (chosen arbitrary).
	 */
	public class SingleMaximum implements CliqueModel {
		
		private Set<Set<Vertex>> maximumClique = new HashSet<Set<Vertex>>();
		private int maximumCliqueSize = 0;
		
		@Override
        public void processClique(Graph g, BitSet R, BitSet P) {
			if (P.isEmpty() && R.cardinality() > maximumCliqueSize) {
				HashSet<Vertex> clique = new HashSet<Vertex>();
				for (int i = R.nextSetBit(0); i >= 0; i = R.nextSetBit(i+1)) {
					clique.add(g.getVertex(i));
				}
				if (clique.size()>maximumCliqueSize) {
					maximumCliqueSize = clique.size();
					maximumClique.clear();
					maximumClique.add(clique);
				}
			}			
		}

		@Override
        public boolean pruneState(BitSet R, BitSet P) {
			if (R.cardinality() + P.cardinality() <= maximumCliqueSize) {
				return true;
			}
//			if (maxDegree + 1 <= maxSize) {
//				//System.out.println("pruned");
//				return true;
//			}
			return false;
		}
		
		@Override
        public Set<Set<Vertex>> getCliques() {
			return maximumClique;
		}
	}
	
}
