package index.features.extractor;

import graph.MoleculeGraph;

import index.features.storage.FeatureStorage;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.openscience.cdk.CDKConstants;
import org.openscience.cdk.interfaces.IAtom;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IBond;

/**
 * A copy of the CDK fingerprinter, modified for comparability.
 */
public class CDKPathExtractor extends FeatureExtractor<String> {
	
	private int searchDepth;
	
	public CDKPathExtractor(MoleculeGraph g, FeatureStorage<? super String, ?> featureStorage, int searchDepth) {
		super(g, featureStorage);
		this.searchDepth = searchDepth;
	}

	
	public void extractFeatures() {
		Map paths = findPathes(((MoleculeGraph)graph).getAtomContainer(), searchDepth);
		for (Iterator e = paths.values().iterator(); e.hasNext(); ) {
			featureStorage.processFeature((String)e.next());
		}
	}
	
	protected Map findPathes(IAtomContainer ac, int searchDepth)
	{
		Map paths = new HashMap();
		List currentPath = new ArrayList();
		for (int f = 0; f < ac.getAtomCount(); f++)
		{
			currentPath.clear();
			currentPath.add(ac.getAtom(f));
			checkAndStore(currentPath, paths);
			depthFirstSearch(ac, ac.getAtom(f), paths, currentPath, 0, searchDepth);
		}
		return paths;
	}
	
	private void depthFirstSearch(IAtomContainer ac, IAtom root, Map paths, List currentPath, int currentDepth, int searchDepth)
	{
		List bonds = ac.getConnectedBondsList(root);

		/*
		 *  try
		 *  {
		 *  logger.debug("Currently at atom no. " + (ac.getAtomNumber(root)  + 1) + " with symbol "  + root.getSymbol());
		 *  }
		 *  catch(Exception exc){}
		 */
		org.openscience.cdk.interfaces.IAtom nextAtom = null;

		/*
		 *  try
		 *  {
		 *  logger.debug("Currently at atom no. " + (ac.getAtomNumber(root)  + 1) + " with symbol "  + root.getSymbol());
		 *  }
		 *  catch(Exception exc){}
		 */
		//Atom tempAtom = null;
		List newPath = null;
		//String symbol = null;
		String bondSymbol = null;
		currentDepth++;
		//logger.info("New incremented searchDepth " + currentDepth);
		//logger.info("Current Path is: " + currentPath);
		for (int f = 0; f < bonds.size(); f++)
		{
			IBond bond = (IBond)bonds.get(f);
			nextAtom = bond.getConnectedAtom(root);

			/*
			 *  try
			 *  {
			 *  logger.debug("Found connected atom no. " + (ac.getAtomNumber(nextAtom) + 1) + " with symbol "  + nextAtom.getSymbol() + "...");
			 *  }
			 *  catch(Exception exc){}
			 */
			if (!currentPath.contains(nextAtom))
			{
				newPath = new ArrayList(currentPath);
				bondSymbol = this.getBondSymbol(bond);
				newPath.add(bondSymbol);
				//logger.debug("Bond has symbol " + bondSymbol);
				newPath.add(nextAtom);
				checkAndStore(newPath, paths);

				if (currentDepth < searchDepth)
				{
					depthFirstSearch(ac, nextAtom, paths, newPath, currentDepth, searchDepth);
					//logger.debug("DepthFirstSearch Fallback to searchDepth " + currentDepth);
				}
			} else
			{
				//logger.debug("... already visited!");
			}
		}
	}

	private void checkAndStore(List newPath, Map paths)
	{
		String newPathString = "";
		for (int f = 0; f < newPath.size(); f++)
		{
			if ((newPath.get(f)) instanceof IAtom)
			{
				newPathString += convertSymbol(((IAtom) newPath.get(f)).getSymbol());
			} else
			{
				newPathString += (String) newPath.get(f);
			}
		}
		//logger.debug("Checking for existence of Path " +  newPathString);
		String storePath = new String(newPathString);
		String reversePath = new StringBuffer(storePath).reverse().toString();
		/*
		 *  Pathes can be found twice (search from one or the other side)
		 *  so they will occur in reversed order. We only handle the
		 *  lexicographically smaller path (This is an arbitrary choice)
		 */
		if (reversePath.compareTo(newPathString) < 0)
		{
			/*
			 *  reversePath is smaller than newPath
			 *  so we keep reversePath
			 */
			storePath = reversePath;
		}
		if (!paths.containsKey(storePath))
		{
			paths.put(storePath, storePath);
			//System.out.println(storePath);
		}
	}

	private String convertSymbol(String symbol)
	{
		String returnSymbol = symbol;
		if (symbol.equals("Cl"))
		{
			symbol = "X";
		} else if (symbol.equals("Si"))
		{
			symbol = "Y";
		} else if (symbol.equals("Br"))
		{
			symbol = "Z";
		}
		return returnSymbol;
	}


	/**
	 *  Gets the bondSymbol attribute of the Fingerprinter class
	 *
	 *@param  bond  Description of the Parameter
	 *@return       The bondSymbol value
	 */
	protected String getBondSymbol(org.openscience.cdk.interfaces.IBond bond)
	{
		String bondSymbol = "";
		if (bond.getFlag(CDKConstants.ISAROMATIC))
		{
			bondSymbol = ":";
		} else if (bond.getOrder() == IBond.Order.SINGLE)
		{
			bondSymbol = "-";
		} else if (bond.getOrder() == IBond.Order.DOUBLE)
		{
			bondSymbol = "=";
		} else if (bond.getOrder() == IBond.Order.TRIPLE)
		{
			bondSymbol = "#";
		}
		return bondSymbol;
	}


}
