package match;

import graph.DefaultNode;
import graph.Graph;
import graph.Node;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import match.MatchDelegate.CollectAll;
import match.pattern.SearchPattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * PartialMapping based on the mapping used for the VF2 algorithm, but extends the approach by symmetry breaking conditions
 * presented by Grochow et. al. in 2007. In case of multiple automorphisms of the given pattern the algorithm adds conditions
 * to "isFeasibleCandidate()" function that allow only automorphisms that map nodes to its identity and therefore reduce 
 * the number of isomorphisms over the same host subgraph (to 1). 
 * @author Markus Kloss
 *
 * @param <NL> The node label
 * @param <EL> The edge label
 */
public class PartialMappingAuto<NL, EL> extends PartialMappingTerminal {
    
    private static final Logger logger = LoggerFactory.getLogger(PartialMappingAuto.class);
    
    protected Map<Integer, LinkedList<AutoCondition<NL, EL>>> SymConditions;
    protected Map<Integer, LinkedList<Node<NL, EL>>> EquivalenceClasses;
    public LinkedList<LinkedList<NodePair<NL, EL>>> Automorphisms;

    /** Constructor
     * @param searchPattern The given search pattern {@link Graph}
     * @param hostGraph The given host {@link Graph}
     */
    public PartialMappingAuto(SearchPattern searchPattern, Graph hostGraph) {
        super(searchPattern, hostGraph);
        
        this.SymConditions = new HashMap<Integer, LinkedList<AutoCondition<NL, EL>>>();
        this.EquivalenceClasses = new HashMap<Integer, LinkedList<Node<NL, EL>>>();
        this.Automorphisms = new LinkedList<LinkedList<NodePair<NL, EL>>>();
        
        // Generate the data that are necessary for the symmetry breaking conditions:
        this.calculateAutomorphismsOfPatternGraph();
        this.EquivalenceClasses = this.calculateEquivalenceClasses(this.Automorphisms);
        this.generateConditions();
    }

    @Override
    public boolean isFeasibleCandidate(Node patternNode, Node hostNode)
    {
        if(super.isFeasibleCandidate(patternNode, hostNode))
        {
            LinkedList<Node<NL, EL>> correspEquivClass = null;
            for(LinkedList<Node<NL, EL>> equivClass : this.EquivalenceClasses.values())
            {
                if(equivClass.contains(patternNode))
                {
                    correspEquivClass = equivClass;
                }
            }
            if(this.SymConditions.containsKey(correspEquivClass.getFirst().getIndex()))
            {
                for(AutoCondition<NL, EL> cond : this.SymConditions.get(correspEquivClass.getFirst().getIndex()))
                {
                    //TODO: Entsprechende Condition-/Automorphismen-Berechnungsmethoden in diese Klasse verschieben,
                    
                    if(cond.firstNode.equals(patternNode))
                    {
                        if(this.mapPH[cond.getSecondNode().getIndex()] != -1)
                        {
                            if(hostNode.getIndex() >= this.mapPH[cond.getSecondNode().getIndex()])
                            {
                                return false;
                            }
                        }
                    }
                    if(cond.secondNode.equals(patternNode))
                    {
                        if(this.mapPH[cond.getFirstNode().getIndex()] != -1)
                        {
                            if(this.mapPH[cond.getFirstNode().getIndex()] >= hostNode.getIndex())
                            {
                                return false;
                            }
                        }
                    }
                }
            }
            return true;
        }
        
        return false;
    }
    
    /**
     * Calculates the automorphisms of the pattern {@link Graph}
     */
    protected void calculateAutomorphismsOfPatternGraph()
    {
        Matcher<NL, EL> matcher = new MatcherVF2(this.pattern, this.pattern);
        
        CollectAll<NL, EL> md = new CollectAll<NL, EL>();        
        matcher.setMatchDelegate(md);        
        matcher.match();
        
        this.Automorphisms = md.getMatches();
    }
    
    protected Map<Integer, LinkedList<Node<NL, EL>>> calculateEquivalenceClasses(LinkedList<LinkedList<NodePair<NL, EL>>> automorphisms)
    {
        Map<Integer, LinkedList<Node<NL, EL>>> equivalenceClasses = new HashMap<Integer, LinkedList<Node<NL, EL>>>();
        
        // One equiv class for each pattern node
        for(DefaultNode<NL, EL> node : (List<DefaultNode<NL, EL>>)this.pattern.nodes())
        {
            equivalenceClasses.put(node.getIndex(), new LinkedList<>());
            equivalenceClasses.get(node.getIndex()).add(node);
        }
        
        // Filling equivalence classes with similar nodes
        for(LinkedList<NodePair<NL, EL>> autom : automorphisms)
        {
            for(NodePair<NL, EL> pair : autom)
            {
                if(!pair.patternNode.equals(pair.hostNode) && !equivalenceClasses.get(pair.hostNode.getIndex()).contains(pair.patternNode))
                {
                    equivalenceClasses.get(pair.hostNode.getIndex()).add(pair.patternNode);
                }
            }
        }
        
        // Remove doubled classes
        LinkedList<Integer> classesToRemove = new LinkedList<Integer>();
        for(LinkedList<Node<NL, EL>> equivClass : equivalenceClasses.values())
        {
            if(equivClass != null)
            {
                for(Node<NL, EL> node : equivClass)
                {
                    if(equivalenceClasses.get(node.getIndex()) != null)
                    {
                        if(!equivalenceClasses.get(node.getIndex()).equals(equivClass))
                        {
                            equivalenceClasses.put(node.getIndex(), null);
                            classesToRemove.add(node.getIndex());
                        }
                    }
                }
            }
        }
        
        // Clean up unnecessary class indexes
        for(Integer i : classesToRemove)
        {
            equivalenceClasses.remove(i);
        }
        
        return equivalenceClasses;
    }

    /**
     * Generates the symmetry breaking conditions (algorithm of Grochow et. al. of 2007)
     */
    protected void generateConditions()
    {
        Map<Integer, LinkedList<Node<NL, EL>>> equivalenceClasses = this.calculateEquivalenceClasses(this.Automorphisms);

        // Generate set of equivalence class representing nodes:
        LinkedList<Node<NL, EL>> reprNodes = new LinkedList<Node<NL, EL>>();
        for(LinkedList<Node<NL, EL>> equivClass : this.calculateEquivalenceClasses(this.Automorphisms).values())
        {
            reprNodes.add(equivClass.getFirst());
        }
        
        // Generating conditions for each equivalence class:
        for(Node<NL, EL> reprNode : reprNodes)
        {
            if(equivalenceClasses.get(reprNode.getIndex()).size() > 1)
            {
                LinkedList<AutoCondition<NL, EL>> conditions = new LinkedList<AutoCondition<NL, EL>>();
                
                Node<NL, EL> currReprNode = reprNode;
                LinkedList<Node<NL, EL>> currEquivClass = equivalenceClasses.get(reprNode.getIndex());
                LinkedList<LinkedList<NodePair<NL, EL>>> currAutoms = this.Automorphisms;
                            
                while(currAutoms.size() > 1 && currEquivClass.size() > 1)
                {
                    // Generating a new condition to prevent permutation of the current representation node:
                    Node<NL, EL> minNode = currReprNode;
                    for(Node<NL, EL> node : currEquivClass)
                    {
                        if(minNode.equals(currReprNode))
                        {
                            minNode = node;
                        }
                        if(minNode.getIndex() > node.getIndex() && !node.equals(currReprNode))
                        {
                            minNode = node;
                        }
                    }
                    if(!minNode.equals(currReprNode))
                    {
                        AutoCondition<NL, EL> cond = new AutoCondition<NL, EL>(currReprNode, minNode);
                        conditions.add(cond);
                    }             
                    
                    // Updating currAutoms, currCompClass and currReprNode:
                    LinkedList<LinkedList<NodePair<NL, EL>>> newCurrAutoms = new LinkedList<LinkedList<NodePair<NL, EL>>>();
                    for(LinkedList<NodePair<NL, EL>> autom : currAutoms)
                    {
                        for(NodePair<NL, EL> pair : autom)
                        {
                            if(pair.hostNode.equals(currReprNode) && pair.patternNode.equals(currReprNode))
                            {
                                newCurrAutoms.add(autom);
                            }
                        }
                    }
                    currAutoms = newCurrAutoms;
                    
                    // The new complexity classes should be a subset of compClass
                    LinkedList<Node<NL, EL>> maxSubClass = null;
                    for(LinkedList<Node<NL, EL>> newCompClass : this.calculateEquivalenceClasses(currAutoms).values())
                    {
                        if(currEquivClass.containsAll(newCompClass))
                        {
                            if(maxSubClass == null || maxSubClass.size() < newCompClass.size())
                            {
                                maxSubClass = newCompClass;
                            }
                        }
                    }
                    
                    currEquivClass = maxSubClass;
                    currReprNode = maxSubClass.getFirst();
                }
                
                this.SymConditions.put(reprNode.getIndex(), conditions);
            }
        }
    }
}
