package struclust.comparison;

import graph.Graph;
import struclust.Cluster;
import struclust.graph.GraphContainer;

import com.google.common.base.Preconditions;

/**
 * Calculated the Distance between two {@link Cluster}s based on the accumulated
 * {@link GraphGraphDistance}s of the pairwise representatives.
 * 
 * @author Till Schäfer
 * 
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class AccRepCCD<NL, EL, G extends Graph<NL, EL>> implements ClusterClusterDistance<NL, EL, G> {

    private GraphGraphDistance<NL, EL, G> ggDist;
    private AccFunction acc;

    /**
     * Constructor
     * 
     * @param ggDist
     *            the {@link GraphGraphDistance} to compare the representatives
     * @param acc
     *            the accumulation function to accumulate the single
     *            representative comparisons
     */
    public AccRepCCD(GraphGraphDistance<NL, EL, G> ggDist, AccFunction acc) {
        Preconditions.checkNotNull(ggDist);
        Preconditions.checkNotNull(acc);
        
        this.ggDist = ggDist;
        this.acc = acc;
    }

    @Override
    public double calc(Cluster<NL, EL, G> cluster1, Cluster<NL, EL, G> cluster2) {
        Preconditions.checkNotNull(cluster1);
        Preconditions.checkNotNull(cluster2);
        int repSize1 = cluster1.getRepresentatives().size();
        int repSize2 = cluster2.getRepresentatives().size();
        Preconditions.checkArgument(repSize1 != 0);
        Preconditions.checkArgument(repSize2 != 0);

        double retVal;

        /*
         * one big switch with duplicate code is faster than a switch in each
         * iteration
         */
        switch (acc) {
        case MIN:
            retVal = Double.POSITIVE_INFINITY;
            for (GraphContainer<NL, EL, G> rep1 : cluster1.getRepresentatives()) {
                for (GraphContainer<NL, EL, G> rep2 : cluster2.getRepresentatives()) {
                    retVal = Math.min(retVal, ggDist.calc(rep1, rep2));
                }
            }
            break;
        case MAX:
            retVal = Double.NEGATIVE_INFINITY;
            for (GraphContainer<NL, EL, G> rep1 : cluster1.getRepresentatives()) {
                for (GraphContainer<NL, EL, G> rep2 : cluster2.getRepresentatives()) {
                    retVal = Math.max(retVal, ggDist.calc(rep1, rep2));
                }
            }
            break;
        case AVG:
            retVal = 0;
            for (GraphContainer<NL, EL, G> rep1 : cluster1.getRepresentatives()) {
                for (GraphContainer<NL, EL, G> rep2 : cluster2.getRepresentatives()) {
                    retVal += ggDist.calc(rep1, rep2) / (repSize1 * repSize2);
                }
            }
            break;
        case SUM:
            retVal = 0;
            for (GraphContainer<NL, EL, G> rep1 : cluster1.getRepresentatives()) {
                for (GraphContainer<NL, EL, G> rep2 : cluster2.getRepresentatives()) {
                    retVal += ggDist.calc(rep1, rep2);
                }
            }
            break;
        default:
            throw new UnsupportedOperationException("This accumulation function is not supported");
        }

        return retVal;
    }

    @Override
    public String getDescription() {
        return "Accumulated Representative CC-Distance (" + ggDist.getDescription() + ", " + acc.name() + ")";
    }

    @Override
    public boolean supportsEmptyRepresentatives() {
        return false;
    }

    @Override
    public boolean supportsEmptyClusters() {
        return true;
    }

}
