package struclust.comparison;

import graph.Graph;
import hsahn.comparison.distance.KernelDistance;
import hsahn.comparison.kernel.Kernel;
import hsahn.graph.GraphTools;
import hsahn.graph.LGraph;
import struclust.graph.GraphContainer;

/**
 * Wrapper class to use {@link KernelDistance} in struclust. Computes the
 * Euclidean distance in the reproducing kernel hilbert space.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class KernelGGD<NL, EL, G extends Graph<NL, EL>> implements GraphGraphDistance<NL, EL, G> {

    private KernelDistance<LGraph<NL, EL>> dist;
    private Kernel<LGraph<NL, EL>> kernel;

    /**
     * Constructor
     * 
     * @param kernel
     *            the used kernel for distance calculation
     */
    public KernelGGD(Kernel<LGraph<NL, EL>> kernel) {
        this.kernel = kernel;
        dist = new KernelDistance<>(kernel);
    }

    @Override
    public double calc(GraphContainer<NL, EL, G> graph1, GraphContainer<NL, EL, G> graph2) {
        /*
         * PERF: This can be done faster by precomputing the explicit mapping
         * with ExplicitMappingKernel#computeExplicitMapping() in advance and
         * apply these mappings to the InnerProductKernel afterwards. This would
         * also save the Graph conversion.
         * 
         * => This will need a redesign of precomputed values (i.e.
         * fingerprints, explicit mappings, etc). modularization!
         */
        LGraph<NL, EL> lg1 = GraphTools.convert(graph1);
        LGraph<NL, EL> lg2 = GraphTools.convert(graph2);
        return dist.compute(lg1, lg2);
    }

    @Override
    public double maxDist() {
        /*
         * XXX: this is actually possible, but must be interpreted different for
         * different kernels, therefore we will assume, there is no such thing
         * as a maximum distance.
         * 
         * Possible solution: The kernel itself (kernel.compute(g1, g2)) returns
         * 0 if there is no "equality" at all. This can be used to return
         * Double.positive_infinitiv in this place. However, it depends in the
         * kernel about how likely this situation is.
         * 
         * For FLRW kernel this might be quite usable, as it is possible to
         * specify the path length and this corresponds to the common subgraph
         * size. I.e. a path is only found if there exist a common subgraph with
         * that size.
         * 
         * For WS-SP oder WS-ST this might be unusable on the other hand, as a
         * single common node would mean that maximum distance is not reached.
         */
        return Double.NaN;
    }

    @Override
    public String getDescription() {
        return "KernelGGD: " + kernel.getID();
    }

}
