package struclust.comparison;

import java.util.Date;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import graph.DefaultGraph;
import graph.Graph;
import graph.LabelEdge;
import struclust.graph.GraphContainer;
import struclust.graph.Graphs;
import struclust.mcs.McsSizeDecisionAlgorithm;

/**
 * Implements a maximum common subgraph distance which simply decides if there
 * is a mcs with a minimum size. The distance only has binary values 0 and 1:
 * 
 * <pre>
 * thresh < |mcs| / max{|graph1|, |graph2|}
 * </pre>
 *
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class McsMinSizeGGD<NL, EL, G extends Graph<NL, EL>> implements GraphGraphDistance<NL, EL, G> {
    private static final Logger logger = LoggerFactory.getLogger(McsEnumWskrGGD.class);

    /**
     * true iff node induced, false iff edgeInduced
     */
    private boolean nodeInduced;
    /**
     * nodeIduced = true -> null
     */
    private McsSizeDecisionAlgorithm<LabelEdge<NL, EL>, NL, DefaultGraph<LabelEdge<NL, EL>, NL>> edgeAlgo;
    /**
     * nodeIduced = false -> null
     */
    private McsSizeDecisionAlgorithm<NL, EL, G> nodeAlgo;

    private double thresh;

    /**
     * Constructor
     * 
     * @param nodeInduced
     *            if true the node induced connected subgraphs will be computed,
     *            otherwise the edge induced connected subgraphs are computed
     * @param nodeAlgo
     *            the algorithm for connected induced maximum common subgraph
     *            computation. if nodeInduced is false this may be null
     * @param edgeAlgo
     *            the algorithm for connected induced maximum common subgraph
     *            computation. if nodeInduced is true this may be null
     * @param thresh
     *            the threshold below which the distance returns 1. values must
     *            be in the interval [0,1].
     */
    public McsMinSizeGGD(boolean nodeInduced, McsSizeDecisionAlgorithm<NL, EL, G> nodeAlgo,
            McsSizeDecisionAlgorithm<LabelEdge<NL, EL>, NL, DefaultGraph<LabelEdge<NL, EL>, NL>> edgeAlgo, double thresh) {
        this.nodeInduced = nodeInduced;
        this.nodeAlgo = nodeAlgo;
        this.edgeAlgo = edgeAlgo;
        this.thresh = thresh;
    }

    @Override
    public double calc(GraphContainer<NL, EL, G> graph1, GraphContainer<NL, EL, G> graph2) {
        Date before = new Date();
        boolean existMcsWithMinSize;
        int count1 = nodeInduced ? graph1.getNodeCount() : graph1.getEdgeCount();
        int count2 = nodeInduced ? graph2.getNodeCount() : graph2.getEdgeCount();
        int absThresh = (int) (thresh * Math.max(count1, count2) + 0.5);

        if (nodeInduced) {
            existMcsWithMinSize = nodeAlgo.calc(graph1.getGraph(), graph2.getGraph(), absThresh);
        } else {
            DefaultGraph<LabelEdge<NL, EL>, NL> lineGraph1 = Graphs.lineGraph(graph1.getGraph());
            DefaultGraph<LabelEdge<NL, EL>, NL> lineGraph2 = Graphs.lineGraph(graph2.getGraph());

            existMcsWithMinSize = edgeAlgo.calc(lineGraph1, lineGraph2, absThresh);
        }

        Date after = new Date();
        logger.debug("AbsThresh: {} ({}), count1: {}, count2: {}, time to compute: {}", absThresh,
                nodeInduced ? "nodeInduced" : "edgeInduced", count1, count2, after.getTime() - before.getTime());
        return existMcsWithMinSize ? 0 : 1;
    }

    @Override
    public double maxDist() {
        return 1;
    }

    @Override
    public String getDescription() {
        return "McsMinSizeGGD (" + (nodeInduced ? "node induced" : "edge induced") + ", algo: "
                + (nodeInduced ? nodeAlgo.getDescription() : edgeAlgo.getDescription()) + ", thresh: " + thresh + ")";
    }

}
