package struclust.comparison;

import graph.DefaultGraph;
import graph.Graph;
import graph.LabelEdge;

import java.util.Date;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.graph.GraphContainer;
import struclust.graph.Graphs;
import struclust.mcs.McsSizeAlgorithm;

/**
 * Implements the maximum common subgraph distance from [1]. That is:<br>
 * 
 * <pre>
 * 1 - |mcs| / (|graph1| + |graph2| - |mcs|)
 * </pre>
 * 
 * <br>
 * 
 * <b>[1]</b> W. D. Wallis, P. Shoubridge, M. Kraetzl, and D. Ray. Graph
 * distances using graph union. Pattern Recognition Letters, 22(6/7):701–704,
 * 2001.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class McsWskrGGD<NL, EL, G extends Graph<NL, EL>> implements GraphGraphDistance<NL, EL, G> {
    private static final Logger logger = LoggerFactory.getLogger(McsEnumWskrGGD.class);

    /**
     * true iff node induced, false iff edgeInduced
     */
    private boolean nodeInduced;
    /**
     * nodeIduced = true -> null
     */
    private McsSizeAlgorithm<LabelEdge<NL, EL>, NL, DefaultGraph<LabelEdge<NL, EL>, NL>> edgeAlgo;
    /**
     * nodeIduced = false -> null
     */
    private McsSizeAlgorithm<NL, EL, G> nodeAlgo;

    /**
     * Constructor
     * 
     * @param nodeInduced
     *            if true the node induced connected subgraphs will be computed,
     *            otherwise the edge induced connected subgraphs are computed
     * @param nodeAlgo
     *            the algorithm for connected induced maximum common subgraph
     *            computation. if nodeInduced is false this may be null
     * @param edgeAlgo
     *            the algorithm for connected induced maximum common subgraph
     *            computation. if nodeInduced is true this may be null
     */
    public McsWskrGGD(boolean nodeInduced, McsSizeAlgorithm<NL, EL, G> nodeAlgo,
            McsSizeAlgorithm<LabelEdge<NL, EL>, NL, DefaultGraph<LabelEdge<NL, EL>, NL>> edgeAlgo) {
        this.nodeInduced = nodeInduced;
        this.nodeAlgo = nodeAlgo;
        this.edgeAlgo = edgeAlgo;
    }

    @Override
    public double calc(GraphContainer<NL, EL, G> graph1, GraphContainer<NL, EL, G> graph2) {
        Date before = new Date();
        int mcs;
        if (nodeInduced) {
            mcs = nodeAlgo.calc(graph1.getGraph(), graph2.getGraph());
        } else {
            DefaultGraph<LabelEdge<NL, EL>, NL> lineGraph1 = Graphs.lineGraph(graph1.getGraph());
            DefaultGraph<LabelEdge<NL, EL>, NL> lineGraph2 = Graphs.lineGraph(graph2.getGraph());

            mcs = edgeAlgo.calc(lineGraph1, lineGraph2);
        }

        int count1 = nodeInduced ? graph1.getNodeCount() : graph1.getEdgeCount();
        int count2 = nodeInduced ? graph2.getNodeCount() : graph2.getEdgeCount();

        Date after = new Date();
        logger.debug("MCS Size: {} ({}), count1: {}, count2: {}, time to compute: {}", mcs, nodeInduced ? "nodeInduced"
                : "edgeInduced", count1, count2, after.getTime() - before.getTime());
        return 1d - ((double) mcs / (count1 + count2 - mcs));
    }

    @Override
    public double maxDist() {
        return 1;
    }

    @Override
    public String getDescription() {
        return "McsWskrGGD (" + (nodeInduced ? "node induced" : "edge induced") + ", algo:"
                + (nodeInduced ? nodeAlgo.getDescription() : edgeAlgo.getDescription()) + ")";
    }

}
