package struclust.competingalgorithms.commonsio;

import graph.Graph;
import graph.GraphSize;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.math3.ml.clustering.DBSCANClusterer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.Cluster;
import struclust.ClusterFactory;
import struclust.StructuralClustering;
import struclust.eval.EvalResult;
import struclust.graph.GraphContainer;

import com.google.common.base.Preconditions;

/**
 * Wraps the {@link DBSCANClusterer} inside a {@link StructuralClustering}.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class WrappedDbscan<NL, EL, G extends Graph<NL, EL>> implements StructuralClustering<NL, EL, G> {
    private static final Logger logger = LoggerFactory.getLogger(WrappedDbscan.class);
    
    private DBSCANClusterer<WrappedGraphClusterable> dbscan;
    private List<GraphContainer<NL, EL, G>> graphs;
    private List<WrappedGraphClusterable> wrappedGraphs;
    private DbscanConf<NL, EL, G> conf;

    /**
     * Constructor
     * 
     * @param conf
     *            the algorithm configuration
     * @param graphs
     *            the graphs to cluster
     */
    public WrappedDbscan(DbscanConf<NL, EL, G> conf, List<GraphContainer<NL, EL, G>> graphs) {
        Preconditions.checkNotNull(conf);

        this.conf = conf;
        this.graphs = graphs;
        this.wrappedGraphs = WrappedGraphClusterable.wrap(graphs);
        this.dbscan = new DBSCANClusterer<>(conf.eps, conf.minPts, new WrappedGGD<NL, EL, G>(graphs, conf.dist));
    }

    @Override
    public void addAsMetadata(EvalResult result) {
        result.addMetadata("clusterer", this.getClass().getName());
        result.addMetadata(conf);
    }

    @Override
    public List<Cluster<NL, EL, G>> calc() {
        List<org.apache.commons.math3.ml.clustering.Cluster<WrappedGraphClusterable>> clustering = dbscan
                .cluster(wrappedGraphs);
        logger.info("number of clusters: {}", clustering.size());
        List<Cluster<NL, EL, G>> retVal = new ArrayList<>(clustering.size());
        Set<GraphContainer<NL, EL, G>> noiseGraphs = new HashSet<>(graphs);
        ClusterFactory<NL, EL, G> clusterFactory = new ClusterFactory<>(GraphSize.edges);
        
        for (org.apache.commons.math3.ml.clustering.Cluster<WrappedGraphClusterable> wrappedCluster : clustering) {
            List<WrappedGraphClusterable> wrappedMembers = wrappedCluster.getPoints();
            List<GraphContainer<NL, EL, G>> unwrappedMembers = new ArrayList<>(wrappedMembers.size());
            for (WrappedGraphClusterable wrappedMember : wrappedMembers) {
                GraphContainer<NL, EL, G> graph = graphs.get(wrappedMember.getId());
                noiseGraphs.remove(graph);
                unwrappedMembers.add(graph);
            }
            Cluster<NL, EL, G> cluster = clusterFactory.newCluster(unwrappedMembers);
            retVal.add(cluster);
        }

        retVal.add(clusterFactory.newCluster(noiseGraphs));
        logger.info("noise cluster size: {}", noiseGraphs.size());
        
        return retVal;
    }

    @Override
    public DbscanConf<NL, EL, G> getConf() {
        return conf;
    }

    @Override
    public void calcFingerprints() {}

    @Override
    public void updateReps(List<Cluster<NL, EL, G>> clusters) {}

}
