package struclust.competingalgorithms.commonsio;

import graph.Graph;

import java.util.List;

import org.apache.commons.math3.ml.distance.DistanceMeasure;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.comparison.GraphGraphDistance;
import struclust.graph.GraphContainer;

/**
 * Wraps a {@link GraphGraphDistance} inside a {@link DistanceMeasure}.
 * 
 * Note: This is only applicable to special prepared datasets which use the
 * {@link WrappedGraphClusterable}.
 * 
 * 
 * @author Till Schäfer
 * 
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 *
 */
public class WrappedGGD<NL, EL, G extends Graph<NL, EL>> implements DistanceMeasure {
    private static final long serialVersionUID = 3305881630448274445L;
    private static final Logger logger = LoggerFactory.getLogger(WrappedGGD.class);

    private List<GraphContainer<NL, EL, G>> id2Graph;
    private GraphGraphDistance<NL, EL, G> dist;

    /**
     * Constructor
     * 
     * @param id2Graph
     *            a mapping from a graph id to a {@link Graph} object. For best
     *            performance, this {@link List} should have a complexity of
     *            O(1) for {@link List#get(int)}.
     * @param dist
     *            the {@link GraphGraphDistance} to use
     */
    public WrappedGGD(List<GraphContainer<NL, EL, G>> id2Graph, GraphGraphDistance<NL, EL, G> dist) {
        this.id2Graph = id2Graph;
        this.dist = dist;
    }

    /**
     * Getter for the wrapped {@link GraphGraphDistance}
     * 
     * @return the wrapped {@link GraphGraphDistance}
     */
    public GraphGraphDistance<NL, EL, G> getWrappedGraphGraphDistance() {
        return dist;
    }

    @Override
    public double compute(double[] a, double[] b) {
        assert a.length == 1 && b.length == 1;
        GraphContainer<NL, EL, G> g1 = id2Graph.get((int) a[0]);
        GraphContainer<NL, EL, G> g2 = id2Graph.get((int) b[0]);
        assert g1 != null;
        assert g2 != null;

        double value = dist.calc(g1, g2);
        logger.debug("distance value: {}", value);
        return value;
    }

}
