package struclust.competingalgorithms.opensubspace;

import graph.Graph;
import graph.GraphSize;
import hsahn.datastructure.FeatureVector;
import i9.data.core.Attribute;
import i9.data.core.Instance;
import i9.subspace.proclus.Proclus;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map.Entry;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.Cluster;
import struclust.ClusterFactory;
import struclust.StructuralClustering;
import struclust.eval.EvalResult;
import struclust.graph.GraphContainer;
//import weka.core.Attribute;
//import weka.core.FastVector;
//import weka.core.Instance;
//import weka.core.Instances;
//import weka.core.SparseInstance;
import weka.core.FastVector;
import weka.core.Instances;

/**
 * Wraps the {@link Proclus} inside a {@link StructuralClustering}.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class WrappedProclus<NL, EL, G extends Graph<NL, EL>> implements StructuralClustering<NL, EL, G> {
    private static final Logger logger = LoggerFactory.getLogger(WrappedProclus.class);

    private static final double MISSING_VAL = 0;
    private ProclusConf<NL, EL, G> conf;
    private List<GraphContainer<NL, EL, G>> graphs;

    /**
     * Constructor
     * 
     * @param conf
     *            the configuration of the clustering algorithm
     * @param graphs
     *            the graphs to cluster
     */
    public WrappedProclus(ProclusConf<NL, EL, G> conf, List<GraphContainer<NL, EL, G>> graphs) {
        this.conf = conf;
        this.graphs = graphs;
    }

    @Override
    public void addAsMetadata(EvalResult result) {
        result.addMetadata("clusterer", this.getClass().getName());
        result.addMetadata(conf);
    }

    @Override
    public List<Cluster<NL, EL, G>> calc() {
        HashMap<GraphContainer<NL, EL, G>, FeatureVector<? extends Object>> gc2fv = new HashMap<>();
        HashMap<Object, Attribute> feature2Attr = new HashMap<>();
        HashSet<Object> features = new HashSet<>();

        for (GraphContainer<NL, EL, G> gc : graphs) {
            FeatureVector<? extends Object> fv = conf.toFeatureVectorFunction.apply(gc.getGraph());
            gc2fv.put(gc, fv);

            for (Object feature : fv.positiveFeatures()) {
                features.add(feature);
            }
        }

        FastVector attributes = new FastVector();
        for (Object feature : features) {
            Attribute attr = new Attribute(feature.toString());
            feature2Attr.put(feature, attr);
            attributes.addElement(attr);
        }
        logger.debug("{} explicit Attributes found", attributes.size());

        Instances instances = new Instances("Graph Instances", attributes, graphs.size());
        for (GraphContainer<NL, EL, G> gc : graphs) {
            Instance instance = new i9.data.core.Instance(attributes.size());
            for (Entry<? extends Object, Long> fvEntry : gc2fv.get(gc).positiveEntries()) {
                int indexOfAttr = feature2Attr.get(fvEntry.getKey()).index();
                Long valueOfAttr = fvEntry.getValue();
                instance.setValue(indexOfAttr, valueOfAttr);
            }
            double[] values = instance.getArray();
            for (int i = 0; i < values.length; i++) {
                if (Double.isNaN(values[i])) {
                    values[i] = MISSING_VAL;
                }
            }

            instances.add(instance);
        }

        Proclus proclus = new Proclus(instances, conf.numClusters, conf.avgDimension);
        try {
            proclus.runClustering();
        } catch (Exception e) {
            logger.error("Proclus clustering failed", e);
            throw new RuntimeException("Proclus clustering failed", e);
        }
        ArrayList<i9.subspace.base.Cluster> clustering = proclus.getClustering();

        List<Cluster<NL, EL, G>> retVal = new ArrayList<>(clustering.size());
        ClusterFactory<NL, EL, G> clusterFactory = new ClusterFactory<>(GraphSize.edges);
        for (i9.subspace.base.Cluster c : clustering) {
            ArrayList<GraphContainer<NL, EL, G>> elements = new ArrayList<>(c.m_objects.size());
            for (Integer objId : c.m_objects) {
                elements.add(graphs.get(objId));
            }
            retVal.add(clusterFactory.newCluster(elements));
        }

        return retVal;
    }

    @Override
    public ProclusConf<NL, EL, G> getConf() {
        return conf;
    }

    @Override
    public void calcFingerprints() {}

    @Override
    public void updateReps(List<Cluster<NL, EL, G>> clusters) {}

}
