package struclust.competingalgorithms.trickl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import com.trickl.cluster.KernelKMeans;

import cern.colt.matrix.DoubleMatrix2D;
import cern.colt.matrix.impl.DenseDoubleMatrix2D;
import graph.Graph;
import graph.GraphSize;
import hsahn.datastructure.FeatureVector;
import hsahn.graph.GraphTools;
import hsahn.graph.LGraph;
import struclust.Cluster;
import struclust.ClusterFactory;
import struclust.StructuralClustering;
import struclust.eval.EvalResult;
import struclust.graph.GraphContainer;

/**
 * A wrapper for the kernel k-means algorithm from the trickl clustering
 * library. This version of {@link WrappedKernelKMeans} makes use of an explicit
 * feature mapping, which make the kernel matrix computation faster in the most
 * cases.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 * @param <T>
 *            the {@link FeatureVector} type
 */
public class WrappedExplicitKernelKMeans<NL, EL, G extends Graph<NL, EL>, T>
        implements StructuralClustering<NL, EL, G> {

    private ExplicitKernelKMeansConf<NL, EL, G, T> conf;
    private List<GraphContainer<NL, EL, G>> graphs;

    /**
     * Constructor
     * 
     * @param conf
     *            the configuration of the clustering algorithm
     * @param graphs
     *            the graphs to cluster
     */
    public WrappedExplicitKernelKMeans(ExplicitKernelKMeansConf<NL, EL, G, T> conf,
            List<GraphContainer<NL, EL, G>> graphs) {
        this.conf = conf;
        this.graphs = graphs;
    }

    @Override
    public void addAsMetadata(EvalResult result) {
        result.addMetadata("clusterer", this.getClass().getName());
        result.addMetadata(conf);
    }

    @Override
    public List<Cluster<NL, EL, G>> calc() {
        ArrayList<Cluster<NL, EL, G>> retVal = new ArrayList<>(conf.numCluster);
        ClusterFactory<NL, EL, G> clusterFactory = new ClusterFactory<>(GraphSize.edges);
        DoubleMatrix2D kernel = new DenseDoubleMatrix2D(graphs.size(), graphs.size());

        Collection<LGraph<NL, EL>> lGraphs = new ArrayList<>(graphs.size());

        for (GraphContainer<NL, EL, G> graph : graphs) {
            lGraphs.add(GraphTools.convert(graph));
        }

        ArrayList<? extends FeatureVector<T>> explicitMapping = conf.kernel.computeExplicitMapping(lGraphs);
        lGraphs = null;

        for (int i = 0; i < graphs.size(); i++) {
            for (int j = i; j < graphs.size(); j++) {
                long kernelVal = explicitMapping.get(i).innerProduct(explicitMapping.get(j));
                kernel.setQuick(i, j, kernelVal);
                kernel.setQuick(j, i, kernelVal);
            }
        }

        KernelKMeans kkm = new KernelKMeans();
        kkm.cluster(kernel, conf.numCluster);
        DoubleMatrix2D partition = kkm.getPartition();

        for (int k = 0; k < conf.numCluster; k++) {
            Cluster<NL, EL, G> cluster = clusterFactory.emptyCluster();
            for (int i = 0; i < graphs.size(); i++) {
                boolean isMember = partition.getQuick(i, k) == 1.0;
                if (isMember) {
                    cluster.add(graphs.get(i));
                }
            }
            if (!cluster.isEmpty()) {
                retVal.add(cluster);
            }
        }

        return retVal;
    }

    @Override
    public ExplicitKernelKMeansConf<NL, EL, G, T> getConf() {
        return conf;
    }

    @Override
    public void calcFingerprints() {}

    @Override
    public void updateReps(List<Cluster<NL, EL, G>> clusters) {}

}
