package struclust.competingalgorithms.trickl;

import java.util.ArrayList;
import java.util.List;

import com.trickl.cluster.KernelKMeans;

import cern.colt.matrix.DoubleMatrix2D;
import cern.colt.matrix.impl.DenseDoubleMatrix2D;
import graph.Graph;
import graph.GraphSize;
import hsahn.graph.GraphTools;
import hsahn.graph.LGraph;
import struclust.Cluster;
import struclust.ClusterFactory;
import struclust.StructuralClustering;
import struclust.eval.EvalResult;
import struclust.graph.GraphContainer;

/**
 * A wrapper for the kernel k-means algorithm from the trickl clustering library
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class WrappedKernelKMeans<NL, EL, G extends Graph<NL, EL>> implements StructuralClustering<NL, EL, G> {

    private KernelKMeansConf<NL, EL, G> conf;
    private List<GraphContainer<NL, EL, G>> graphs;

    /**
     * Constructor
     * 
     * @param conf
     *            the configuration of the clustering algorithm
     * @param graphs
     *            the graphs to cluster
     */
    public WrappedKernelKMeans(KernelKMeansConf<NL, EL, G> conf, List<GraphContainer<NL, EL, G>> graphs) {
        this.conf = conf;
        this.graphs = graphs;
    }

    @Override
    public void addAsMetadata(EvalResult result) {
        result.addMetadata("clusterer", this.getClass().getName());
        result.addMetadata(conf);
    }

    @Override
    public List<Cluster<NL, EL, G>> calc() {
        ArrayList<Cluster<NL, EL, G>> retVal = new ArrayList<>(conf.numCluster);
        ClusterFactory<NL, EL, G> clusterFactory = new ClusterFactory<>(GraphSize.edges);
        DoubleMatrix2D kernel = new DenseDoubleMatrix2D(graphs.size(), graphs.size());

        for (int i = 0; i < graphs.size(); i++) {
            LGraph<NL, EL> lg1 = GraphTools.convert(graphs.get(i));
            for (int j = i; j < graphs.size(); j++) {
                LGraph<NL, EL> lg2 = GraphTools.convert(graphs.get(j));
                double kernelVal = conf.kernel.compute(lg1, lg2);
                kernel.setQuick(i, j, kernelVal);
                kernel.setQuick(j, i, kernelVal);
            }
        }

        KernelKMeans kkm = new KernelKMeans();
        kkm.cluster(kernel, conf.numCluster);
        DoubleMatrix2D partition = kkm.getPartition();

        for (int k = 0; k < conf.numCluster; k++) {
            Cluster<NL, EL, G> cluster = clusterFactory.emptyCluster();
            for (int i = 0; i < graphs.size(); i++) {
                boolean isMember = partition.getQuick(i, k) == 1.0;
                if (isMember) {
                    cluster.add(graphs.get(i));
                }
            }
            if (!cluster.isEmpty()) {
                retVal.add(cluster);
            }
        }

        return retVal;
    }

    @Override
    public KernelKMeansConf<NL, EL, G> getConf() {
        return conf;
    }

    @Override
    public void calcFingerprints() {}

    @Override
    public void updateReps(List<Cluster<NL, EL, G>> clusters) {}

}
