package struclust.convergence;

import java.util.Collection;
import java.util.LinkedList;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.Cluster;

/**
 * Assumes that the algorithm is convergent if the average changes over the last
 * avgCount calls of
 * {@link ConvergenceCriteria#convergent(int, int, Collection)} are no longer
 * decreasing.
 * 
 * Special case: until avgCount change values are collected,
 * {@link ConvergenceCriteria#convergent(int, int, Collection)} will always
 * return false.
 * 
 * @author Till Schäfer
 *
 */
public class ChangesDecreasingCC implements ConvergenceCriteria {
    private static final Logger logger = LoggerFactory.getLogger(ChangesDecreasingCC.class);

    private int avgCount;
    private LinkedList<Integer> values = new LinkedList<>();

    /**
     * Constructor
     * 
     * @param avgCount
     *            the number of values which are considered for average
     *            calculation.
     */
    public ChangesDecreasingCC(int avgCount) {
        this.avgCount = avgCount;
    }

    @Override
    public boolean convergent(int changes, int totalElements, Collection<? extends Cluster<?, ?, ?>> clustering) {
        if (values.size() < avgCount) {
            values.add(changes);
            return false;
        }

        assert values.size() == avgCount;

        double prevAvg = values.stream().mapToInt(x -> x.intValue()).average().getAsDouble();
        values.add(changes);
        values.remove();
        double currentAvg = values.stream().mapToInt(x -> x.intValue()).average().getAsDouble();

        logger.debug("previous Average: {}, current Average {}", prevAvg, currentAvg);

        return prevAvg <= currentAvg;
    }

    @Override
    public void reset() {
        values = new LinkedList<>();
        logger.debug("reseting");
    }
    
    @Override
    public String getDescription() {
        return "ChangesDecreasingCC over " + avgCount + " values)";
    }
}
