package struclust.convergence;

import java.util.Collection;
import java.util.LinkedList;
import java.util.function.ToDoubleFunction;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.Cluster;

import com.google.common.base.Preconditions;

/**
 * {@link ConvergenceCriteria} that takes a takes a arbitrary lamda function as
 * measurement. It is possible to specify a minimum increase or decrease
 * threshold (percentage wise). If this threshold is not reached
 * {@link #convergent(int, int, Collection)} returns true.
 * 
 * It is possible to specify a range over which the change will be measured.
 * I.e. the current value is compared with the value |range| iterations before.
 * E.g. one can specify, that the clustering is convergent if the value does not
 * exceed reaches the threshold over a specified number of iterations.
 * 
 * Furthermore, one can specify an average width. That is the current value and
 * the compared value (current - range) are averaged over the last |width|
 * values.
 * 
 * @author Till Schäfer
 *
 */
public class GenericRangeAvgCC implements ConvergenceCriteria {
    private static final Logger logger = LoggerFactory.getLogger(GenericRangeAvgCC.class);

    private ToDoubleFunction<Collection<? extends Cluster<?, ?, ?>>> measure;
    private int range;
    private int avgWidth;
    private double increase;
    private LinkedList<Double> values = new LinkedList<>();

    /**
     * Constructor
     * 
     * @param measure
     *            a function that maps the clustering to some double value
     * @param increase
     *            the increase (or decrease for negative values) threshold. this
     *            must be given as relative fraction, e.g. a value of 1 means
     *            100%. Must be different from 0.
     * @param range
     *            the range over which the values are compared, i.e. range =
     *            CurrentValueItaration - ComparisonValueItartion. Therefore, a
     *            value of 1 is the minimum value, meaning that the current and
     *            the last value are compared. If avgWidth > 1 is given, the
     *            range describes the difference between the end of the first
     *            averaging window and the start of the second one.
     * @param avgWidth
     *            the width of the average window for the current and the
     *            comparison value. Must be larger than 0.
     */
    public GenericRangeAvgCC(ToDoubleFunction<Collection<? extends Cluster<?, ?, ?>>> measure, double increase,
            int range, int avgWidth) {
        Preconditions.checkArgument(increase != 0);
        Preconditions.checkArgument(range >= 1);
        Preconditions.checkArgument(avgWidth >= 1);

        this.measure = measure;
        this.increase = increase;
        this.range = range;
        this.avgWidth = avgWidth;
    }

    @Override
    public String getDescription() {
        return "GenericRangeAvgCC(measure: " + measure.toString() + ", increase: " + increase + ", range: " + range
                + ", avgWidth: " + avgWidth + ")";
    }

    @Override
    public boolean convergent(int changes, int totalElements, Collection<? extends Cluster<?, ?, ?>> clustering) {
        double value = measure.applyAsDouble(clustering);
        logger.debug("Adding value {}", value);
        values.add(value);

        int numValues = range + 2 * avgWidth;
        boolean notEnoughValues = values.size() < numValues;
        if (notEnoughValues) {
            logger.debug("not enough values, convergent: false");
            return false;
        } else {
            double currentAvg = values.subList(numValues - avgWidth, numValues).stream().mapToDouble(d -> d).average()
                    .orElseThrow(() -> new IllegalStateException());
            double comparisonAvg = values.subList(0, avgWidth).stream().mapToDouble(d -> d).average()
                    .orElseThrow(() -> new IllegalStateException());
            double percDiff = (currentAvg - comparisonAvg) / comparisonAvg;

            values.removeFirst();

            boolean retVal = (increase > 0) ? percDiff < increase : percDiff > increase;
            logger.debug("currentAvg: {}, comparisonAvg: {}, percDiff: {}, convergent: {}", currentAvg, comparisonAvg,
                    percDiff, retVal);
            return retVal;
        }
    }

    @Override
    public void reset() {
        values.clear();
    }

}
