package struclust.convergence;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;

import struclust.Cluster;

/**
 * Considers a clustering to be convergent iff there exist at most x clusters
 * with a RepToMemberSize below a given threshold.
 * 
 * @author Till Schäfer
 */
public class RepToMemberSizeCC implements ConvergenceCriteria {

    private int numClusters;
    private double threshold;

    /**
     * Constructor
     * 
     * @param numClusters
     *            the number of clusters which are alowed to be below the
     *            threshold
     * @param threshold
     *            the actual RepToMemberSize threshold
     */
    public RepToMemberSizeCC(int numClusters, double threshold) {
        this.numClusters = numClusters;
        this.threshold = threshold;

    }

    @Override
    public boolean convergent(int changes, int totalElements, Collection<? extends Cluster<?, ?, ?>> clustering) {
        LinkedList<? extends Cluster<?, ?, ?>> sortedClusters = new LinkedList<>(clustering);
        Collections.sort(sortedClusters, new Comparator<Cluster<?, ?, ?>>() {
            @Override
            public int compare(Cluster<?, ?, ?> c1, Cluster<?, ?, ?> c2) {
                return Double.compare(c1.getRepToMemberSize(), c2.getRepToMemberSize());
            }
        });

        return sortedClusters.get(Math.min(numClusters, sortedClusters.size())).getRepToMemberSize() >= threshold;
    }

    @Override
    public void reset() {
        // does nothing
    }

    @Override
    public String getDescription() {
        return "RepToMemberSizeCC (numClusters: " + numClusters + " , threshold: " + threshold + ")";
    }

}
