package struclust.convergence;

import java.util.Collection;
import java.util.LinkedList;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.Cluster;

/**
 * Assumes that the algorithm is convergent if the current change value is the
 * worst in the last count calls of
 * {@link ConvergenceCriteria#convergent(int, int, Collection)}.
 * 
 * Special case: until count change values are collected,
 * {@link ConvergenceCriteria#convergent(int, int, Collection)} will always
 * return false.
 * 
 * @author Till Schäfer
 *
 */
public class WorstChangesInLastXCC implements ConvergenceCriteria {
    private static final Logger logger = LoggerFactory.getLogger(WorstChangesInLastXCC.class);

    private int count;
    private LinkedList<Integer> values = new LinkedList<Integer>();

    /**
     * Constructor
     * 
     * @param count
     *            the number of values which are considered
     */
    public WorstChangesInLastXCC(int count) {
        this.count = count;
    }

    @Override
    public boolean convergent(int changes, int totalElements, Collection<? extends Cluster<?, ?, ?>> clustering) {
        if (values.size() < count) {
            values.add(changes);
            return false;
        }

        assert values.size() == count;

        int prevMax = values.stream().mapToInt(x -> x.intValue()).max().getAsInt();
        values.add(changes);
        values.remove();

        logger.debug("previous max: {}, current value {}", prevMax, changes);

        return changes >= prevMax;
    }

    @Override
    public void reset() {
        values = new LinkedList<>();
        logger.debug("reseting");
    }

    @Override
    public String getDescription() {
        return "WorstChangesInLastXCC over " + count + " values)";
    }
}
