package struclust.granularity;

import graph.Graph;

import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ForkJoinPool;

import org.apache.commons.lang3.tuple.Pair;
import org.slf4j.Logger;

import struclust.Cluster;
import struclust.SharedMemorySCC;
import struclust.comparison.ClusterClusterDistance;
import struclust.datastructures.MutableInt;
import struclust.util.Concurrent;

/**
 * Selects two clusters for merge if their {@link ClusterClusterDistance} is
 * less than minSep. If merge conflicts occur (i.e. there exist more than one
 * possible merge for a single cluster), the lowest separation cluster pair is
 * selected.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class MinSepMC<NL, EL, G extends Graph<NL, EL>> implements MergeCriteria<NL, EL, G> {
    private static final Logger logger = org.slf4j.LoggerFactory.getLogger(MinSepMC.class);
    private ClusterClusterDistance<NL, EL, G> ccDist;
    private double distThreshold;
    private ForkJoinPool forkJoinPool;

    /**
     * Constructor
     * 
     * @param ccDist
     *            the distance for separation calculation
     * @param distThreshold
     *            the minimum separation value for a merge. note that this value
     *            is actually a distance threshold. a lower value is therefore
     *            less restrictive!
     * @param parallelism
     *            the parallelism (see {@link SharedMemorySCC#parallelism}
     */
    public MinSepMC(ClusterClusterDistance<NL, EL, G> ccDist, double distThreshold, MutableInt parallelism) {
        this.ccDist = ccDist;
        this.distThreshold = distThreshold;
        this.forkJoinPool = new ForkJoinPool(Concurrent.getEffectiveParallelism(parallelism.get()));
    }

    @Override
    public String getDescription() {
        return "MinSepMC (minSep = " + distThreshold + ", separation criteria: " + ccDist.getDescription() + ")";
    }

    @Override
    public LinkedList<Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>>> toMerge(List<Cluster<NL, EL, G>> clusters) {
        LinkedList<Pair<Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>>, Double>> candidates = new LinkedList<>();

        /*
         * PERF: This does not parallelize well for outliers (especially for MCS
         * distance measures)
         * 
         * This is a workaround for Java stream api to not support a ThreadPool
         * specification
         * 
         * PARALLEL stream: calculate merge pairs
         */
        try {
            forkJoinPool.submit(
                    () -> clusters.stream().parallel().forEach(c1 -> generateCandidates(clusters, candidates, c1)))
                    .get();
        } catch (InterruptedException e) {
            logger.warn("Interrupted: ", e);
            throw new RuntimeException(e);
        } catch (ExecutionException e) {
            logger.error("Error in merge pair calculation: ", e);
            throw new RuntimeException(e);
        }
        logger.debug("generated {} merge pair candidates", candidates.size());

        Collections.sort(candidates, (o1, o2) -> Double.compare(o1.getRight(), o2.getRight()));

        HashSet<Cluster<NL, EL, G>> blocked = new HashSet<>();
        LinkedList<Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>>> retVal = new LinkedList<>();
        for (Pair<Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>>, Double> pair : candidates) {
            Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>> realPair = pair.getLeft();
            if (!blocked.contains(realPair.getLeft()) && !blocked.contains(realPair.getRight())) {
                retVal.add(realPair);
                blocked.add(realPair.getLeft());
                blocked.add(realPair.getRight());
            }
        }

        logger.debug("selected {} cluster pairs for merge", retVal.size());

        return retVal;
    }

    private void generateCandidates(List<Cluster<NL, EL, G>> clusters,
            LinkedList<Pair<Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>>, Double>> candidates, Cluster<NL, EL, G> c1) {
        for (Cluster<NL, EL, G> c2 : clusters) {
            if (c1 == c2) {
                break;
            }
            boolean runIfEmptyReps = ccDist.supportsEmptyRepresentatives()
                    || (!c1.getRepresentatives().isEmpty() && !c2.getRepresentatives().isEmpty());
            boolean runIfEmptyClusters = ccDist.supportsEmptyClusters() || (!c1.isEmpty() && !c2.isEmpty());
            if (runIfEmptyReps && runIfEmptyClusters) {
                double dist = ccDist.calc(c1, c2);

                if (dist <= distThreshold) {
                    synchronized (candidates) {
                        candidates.add(Pair.of(Pair.of(c1, c2), dist));
                    }
                }
            }
        }
    }
}
