package struclust.granularity;

import graph.Graph;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import struclust.Cluster;
import struclust.ClusterFactory;
import struclust.SharedMemorySCC;
import struclust.StructuralClustering;
import struclust.graph.GraphContainer;
import struclust.preclustering.PreClustering;

/**
 * Uses the {@link PreClustering} module to re-cluster all graphs from the split
 * clusters (i.e. all members from the split clusters are merged).
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class RePreClusterSS<NL, EL, G extends Graph<NL, EL>> implements SplitStrategy<NL, EL, G> {
    private int qFactor;
    private PreClustering<NL, EL, G> preClust;

    /**
     * Constructor. The {@link PreClustering} from the {@link SharedMemorySCC} is
     * used.
     * 
     * @param conf
     *            the clustering configuration
     * @param qFactor
     *            the desired factor for the quantity of the resulting clusters
     *            relative to the size of toSplit.
     */
    public RePreClusterSS(SharedMemorySCC<NL, EL, G> conf, int qFactor) {
        this.preClust = conf.preClusterer;
        this.qFactor = qFactor;
    }

    /**
     * Constructor
     * 
     * @param preClust
     *            the configured {@link PreClustering} module.
     * @param qFactor
     *            the desired factor for the quantity of the resulting clusters
     *            relative to the size of toSplit.
     */
    public RePreClusterSS(PreClustering<NL, EL, G> preClust, int qFactor) {
        this.preClust = preClust;
        this.qFactor = qFactor;
    }

    @Override
    public List<Cluster<NL, EL, G>> split(List<Cluster<NL, EL, G>> toSplit, List<Cluster<NL, EL, G>> allClusters,
            ClusterFactory<NL, EL, G> clusterFactory, StructuralClustering<NL, EL, G> parentClusterer) {
        ArrayList<GraphContainer<NL, EL, G>> splitGraphs = new ArrayList<>();
        for (Cluster<NL, EL, G> cluster : toSplit) {
            splitGraphs.addAll(cluster);
        }

        /*
         * apply pre-clustering to splitGraphs
         */
        int numClusters = Math.min(toSplit.size() * qFactor, splitGraphs.size());
        List<? extends List<GraphContainer<NL, EL, G>>> preClustering = preClust.calc(splitGraphs, numClusters);
        ArrayList<Cluster<NL, EL, G>> newClusters = new ArrayList<>(preClustering.size());
        for (List<GraphContainer<NL, EL, G>> preCluster : preClustering) {
            Cluster<NL, EL, G> cluster = clusterFactory.newCluster(preCluster);
            newClusters.add(cluster);
        }

        /*
         * remove empty clusters
         */
        for (Iterator<Cluster<NL, EL, G>> iterator = newClusters.iterator(); iterator.hasNext();) {
            Cluster<NL, EL, G> cluster = iterator.next();
            if (cluster.isEmpty()) {
                iterator.remove();
            }
        }

        parentClusterer.updateReps(newClusters);

        ArrayList<Cluster<NL, EL, G>> retVal = new ArrayList<>(allClusters);
        retVal.removeAll(toSplit);
        retVal.addAll(newClusters);
        return retVal;
    }

    @Override
    public String getDescription() {
        return "RePreClusterSS (qFactor = " + qFactor + ", preClust:" + preClust.getDescription() + ")";
    }

}
