package struclust.granularity;

import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import graph.Graph;
import graph.GraphSize;
import struclust.Cluster;
import struclust.util.Ratings;

/**
 * Chooses a cluster for splitting iff
 * 
 * <pre>
 * {@link Ratings#relRepToMemberSize(Cluster, List)} < minScore
 * </pre>
 * 
 * @author Till Schäfer
 */
public class RelRepToMemberSizeSC implements SplitCriteria {
    private static Logger logger = LoggerFactory.getLogger(RelRepToMemberSizeSC.class);

    private double minScore;
    private int maxRepSize;
    private GraphSize gSize;

    /**
     * Constructor
     * 
     * @param minScore
     *            the minimimal score to not split a cluster
     */
    public RelRepToMemberSizeSC(double minScore) {
        this(minScore, Integer.MAX_VALUE, GraphSize.edges);
    }

    /**
     * Constructor
     * 
     * @param minScore
     *            the minimimal score to not split a cluster
     * @param maxRepSize
     *            if representative size is limited, clusters with
     *            representatives of this size should not be splitted
     * @param gSize
     *            the way the graph size is measured
     */
    public RelRepToMemberSizeSC(double minScore, int maxRepSize, GraphSize gSize) {
        this.minScore = minScore;
        this.maxRepSize = maxRepSize;
        this.gSize = gSize;
    }

    @Override
    public <NL, EL, G extends Graph<NL, EL>> List<Cluster<NL, EL, G>> toSplit(List<Cluster<NL, EL, G>> clusters) {
        LinkedList<Cluster<NL, EL, G>> retVal = new LinkedList<>();

        /*
         * Valid are clusters which are not empty and have no at least one
         * representative. Furthermore, if maxRepSize is specified, clusters
         * which reach the maxRepSize should not considered for splitting.
         */
        List<Cluster<NL, EL, G>> validClusters = clusters.stream()
                .filter(x -> x.getRepresentatives().size() != 0 && x.size() != 0
                        && x.getRepresentatives().stream().mapToInt(g -> g.size(gSize)).min().orElse(0) <= maxRepSize)
                .collect(Collectors.toList());

        if (validClusters.isEmpty()) {
            return retVal; // empty
        }

        for (Cluster<NL, EL, G> cluster : clusters) {
            if (cluster.getRepresentatives().isEmpty()
                    || Ratings.relRepToMemberSize(cluster, validClusters) < minScore) {
                retVal.add(cluster);
            }
        }

        if (logger.isDebugEnabled()) {
            logger.debug("selected {} clusters for splitting with {} graphs", retVal.size(),
                    retVal.stream().mapToInt(x -> x.size()).sum());
        }

        return retVal;
    }

    @Override
    public String getDescription() {
        return "RelRepToMemberSizeSC (minScore = " + minScore + ")";
    }

}
