package struclust.granularity;

import graph.Graph;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import org.apache.commons.lang3.tuple.Pair;
import org.slf4j.Logger;

import struclust.Cluster;
import struclust.comparison.GraphGraphDistance;
import struclust.datastructures.MutableInt;
import struclust.graph.GraphContainer;
import struclust.util.Concurrent;

import com.google.common.util.concurrent.ThreadFactoryBuilder;

/**
 * Selects two clusters for merge if there exist at least n representative pairs
 * with: <br>
 * 
 * <pre>
 * dist(rep1, rep2) &lt;= distThreshold
 * </pre>
 * 
 * <br>
 * 
 * This does exactly the same as {@link RepThresholdCountMC}, but with a very
 * fine grained parallelism. The finer parallelism introduces some extra
 * overhead, but lowers the problem of a single cluster comparisons dominating
 * the overall runtime. This problem was especially observed with MCS distances.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class RepThresholdCountFineParallelMC<NL, EL, G extends Graph<NL, EL>> implements MergeCriteria<NL, EL, G> {
    private static final Logger logger = org.slf4j.LoggerFactory.getLogger(RepThresholdCountFineParallelMC.class);

    private GraphGraphDistance<NL, EL, G> ggDist;
    private double distThreshold;
    private int n;
    private ExecutorService executor;

    /**
     * Constructor
     * 
     * @param parallelism
     *            the parallelism (how many threads are generated for each
     *            thread pool)
     * @param ggDist
     *            the distance for separation calculation
     * @param distThreshold
     *            the minimum separation value for a merge. note that this value
     *            is actually a distance threshold. a lower value is therefore
     *            less restrictive!
     * @param n
     *            the number
     */
    public RepThresholdCountFineParallelMC(MutableInt parallelism, GraphGraphDistance<NL, EL, G> ggDist,
            double distThreshold, int n) {
        this.ggDist = ggDist;
        this.distThreshold = distThreshold;
        this.n = n;
        executor = Executors.newFixedThreadPool(Concurrent.getEffectiveParallelism(parallelism.get()),
                new ThreadFactoryBuilder().setNameFormat("MergeCritera-%d").setDaemon(true)
                .build());
    }

    @Override
    public String getDescription() {
        return "RepThresholdCountFineParallelMC (minSep: " + distThreshold + ", separation criteria: "
                + ggDist.getDescription() + ", distThreshold: " + distThreshold + ", n: " + n + ")";
    }

    @Override
    public LinkedList<Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>>> toMerge(List<Cluster<NL, EL, G>> clusters) {
        LinkedList<Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>>> candidates = new LinkedList<>();

        generateCandidates(clusters, candidates);

        logger.debug("generated {} merge pair candidates", candidates.size());

        HashSet<Cluster<NL, EL, G>> blocked = new HashSet<>();
        LinkedList<Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>>> retVal = new LinkedList<>();
        for (Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>> pair : candidates) {
            Cluster<NL, EL, G> c1 = pair.getLeft();
            Cluster<NL, EL, G> c2 = pair.getRight();
            if (!blocked.contains(c1) && !blocked.contains(c2)) {
                retVal.add(pair);
                blocked.add(c1);
                blocked.add(c2);
            }
        }

        logger.debug("selected {} cluster pairs for merge", retVal.size());

        return retVal;
    }

    private void generateCandidates(List<Cluster<NL, EL, G>> clusters,
            LinkedList<Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>>> candidates) {

        List<Future<Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>>>> futures = new ArrayList<>(clusters.size()
                * clusters.size());
        // PARALLEL: calculate merge pairs
        for (Cluster<NL, EL, G> c1 : clusters) {
            for (Cluster<NL, EL, G> c2 : clusters) {
                futures.add(executor.submit(new SingleCandidateJob(c1, c2)));
            }
        }

        try {
            for (Future<Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>>> future : futures) {
                Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>> result = future.get();
                if (result != null) {
                    candidates.add(result);
                }
            }
        } catch (InterruptedException | ExecutionException e) {
            logger.warn("failed to calculate merge pair candidates");
            throw new RuntimeException(e);
        }
    }

    /**
     * 
     * @author Till Schäfer
     *
     */
    private class SingleCandidateJob implements Callable<Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>>> {
        private Cluster<NL, EL, G> c1;
        private Cluster<NL, EL, G> c2;

        public SingleCandidateJob(Cluster<NL, EL, G> c1, Cluster<NL, EL, G> c2) {
            this.c1 = c1;
            this.c2 = c2;
        }

        @Override
        public Pair<Cluster<NL, EL, G>, Cluster<NL, EL, G>> call() throws Exception {
            if (c1.getId() > c2.getId()) {

                int count = 0;
                for (GraphContainer<NL, EL, G> rep1 : c1.getRepresentatives()) {
                    for (GraphContainer<NL, EL, G> rep2 : c2.getRepresentatives()) {
                        double dist = ggDist.calc(rep1, rep2);

                        if (dist <= distThreshold) {
                            count++;
                        }
                        if (count >= n) {
                            return Pair.of(c1, c2);
                        }
                    }
                }
            }
            return null;
        }

    }
}
