package struclust.granularity;

import graph.Graph;

import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.Cluster;

/**
 * Always splits a specified number of clusters with worst RepToMembersSize. In
 * addition a maximum RepToMembersSize can be specified. Clusters with a
 * RepToMembersSize greater or equal to this threshold will never be split up.
 * 
 * @author Till Schäfer
 */
public class WorstRepToMemberSizeSC implements SplitCriteria {
    private static final Logger logger = LoggerFactory.getLogger(WorstRepToMemberSizeSC.class);
    private int num;
    private double maxRepToMembersSize;

    /**
     * Constructor
     * 
     * @param num
     *            the number of clusters to split
     * @param maxRepToMembersSize
     *            the maximum RepToMembersSize. i.e. clusters with a
     *            RepToMembersSize greater or equal to this threshold will never
     *            be split up.
     */
    public WorstRepToMemberSizeSC(int num, double maxRepToMembersSize) {
        this.num = num;
        this.maxRepToMembersSize = maxRepToMembersSize;

    }

    @Override
    public <NL, EL, G extends Graph<NL, EL>> List<Cluster<NL, EL, G>> toSplit(List<Cluster<NL, EL, G>> clusters) {
        LinkedList<Cluster<NL, EL, G>> sortedClusters = new LinkedList<>(clusters);

        Collections.sort(sortedClusters, new Comparator<Cluster<NL, EL, G>>() {
            @Override
            public int compare(Cluster<NL, EL, G> c1, Cluster<NL, EL, G> c2) {
                return Double.compare(c1.getRepToMemberSize(), c2.getRepToMemberSize());
            }
        });

        List<Cluster<NL, EL, G>> retVal = sortedClusters.subList(0, Math.min(sortedClusters.size(), num));

        for (Iterator<Cluster<NL, EL, G>> it = retVal.iterator(); it.hasNext();) {
            if (it.next().getRepToMemberSize() >= maxRepToMembersSize) {
                it.remove();
            }
        }

        if (logger.isDebugEnabled()) {
            logger.debug(
                    "selected {} clusters for splitting with {} graphs. Splitted Clusters RepToMembersSizes: max {}, min {}.",
                    retVal.size(), retVal.stream().mapToInt(x -> x.size()).sum(),
                    retVal.isEmpty() ? "NA" : retVal.get(retVal.size() - 1).getRepToMemberSize(),
                    retVal.isEmpty() ? "NA" : retVal.get(0).getRepToMemberSize());
        }
        return retVal;
    }

    @Override
    public String getDescription() {
        return "WorstRelRepToMemberSizeSC (num: " + num + ", maxRepToMembersSize: " + maxRepToMembersSize + ")";
    }

}
