package struclust.graph.generators;

import graph.DefaultGraph;
import graph.DefaultNode;

import java.util.ArrayList;
import java.util.Random;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ThreadLocalRandom;
import java.util.stream.IntStream;

import struclust.datastructures.MutableInt;
import struclust.util.Statistical;
import cern.jet.random.AbstractDistribution;

import com.google.common.base.Preconditions;

/**
 * Creates a random connected {@link DefaultGraph} with {@link Integer} labels.
 * The number of labels, the label distribution and the graphs size distribution
 * can be specified.
 * 
 * @author Till Schäfer
 * @author Nils Kriege
 */
public class ConnectedRGG implements RandomGraphGenerator<Integer, Integer, DefaultGraph<Integer, Integer>> {
    protected double cycleProbability;

    protected AbstractDistribution labelDistribution;
    protected AbstractDistribution sizeDistribution;

    protected double[] edgeLabelWeights;
    protected double[] vertexLabelWeights;

    protected int vertexLabelCount;
    protected int edgeLabelCount;

    protected MutableInt parallelism;

    /**
     * Constructor
     * 
     * @param vertexLabelCount
     *            the number of vertex labels. must be larger than 0.
     * @param edgeLabelCount
     *            the number of edge labels. must be larger than 0.
     * @param labelDistribution
     *            the distribution function for the labels
     * @param sizeDistribution
     *            the distribution function for the size (in terms of edges)
     * @param cycleProbability
     *            the probability for each edge inserted into a seed with more
     *            than two edges to create a cycle. must be in the interval
     *            [0,1].
     */
    public ConnectedRGG(int vertexLabelCount, int edgeLabelCount, AbstractDistribution labelDistribution,
            AbstractDistribution sizeDistribution, double cycleProbability) {
        Preconditions.checkArgument(vertexLabelCount > 0);
        Preconditions.checkArgument(edgeLabelCount > 0);
        Preconditions.checkNotNull(labelDistribution);
        Preconditions.checkNotNull(sizeDistribution);
        Preconditions.checkArgument(cycleProbability >= 0 && cycleProbability <= 1);

        this.edgeLabelCount = edgeLabelCount;
        this.vertexLabelCount = vertexLabelCount;
        this.labelDistribution = labelDistribution;
        this.sizeDistribution = sizeDistribution;
        this.cycleProbability = cycleProbability;
        this.vertexLabelWeights = Statistical.discreteWeightCumulativeDistribution(vertexLabelCount, labelDistribution);
        this.edgeLabelWeights = Statistical.discreteWeightCumulativeDistribution(edgeLabelCount, labelDistribution);
    }

    @Override
    public ArrayList<DefaultGraph<Integer, Integer>> createGraphs(int num) {
        ConcurrentLinkedQueue<DefaultGraph<Integer, Integer>> graphs = new ConcurrentLinkedQueue<>();
        
         // PARALLEL stream: create single graphs
        IntStream.range(0, num).parallel().forEach(i -> graphs.add(createSingleConnectedGraph()));

        ArrayList<DefaultGraph<Integer, Integer>> retVal = new ArrayList<>();
        graphs.stream().forEach(g -> retVal.add(g));

        assert retVal.stream().filter(g -> g == null).count() == 0;
        return retVal;
    }

    /**
     * Creates a seed graph with the given size (in terms of edges). Vertex and
     * edge labels are randomly added, the number of nodes depends on
     * <b>seedCycleProbability</b>.
     * 
     * @param size
     * @return
     */
    protected DefaultGraph<Integer, Integer> createSingleConnectedGraph() {
        DefaultGraph<Integer, Integer> g = new DefaultGraph<Integer, Integer>();
        int size;
        synchronized (sizeDistribution) {
            size = sizeDistribution.nextInt();
        }
        return extendSingleConnectedGraph(size, g);
    }

    /**
     * extends a single connected {@link DefaultGraph}
     * 
     * @param numExtensions
     *            the size of the graph
     * @param g
     *            the graph to extend
     * @return
     */
    protected DefaultGraph<Integer, Integer> extendSingleConnectedGraph(int numExtensions,
            DefaultGraph<Integer, Integer> g) {
        // RAND: random graph generation
        ThreadLocalRandom rand = ThreadLocalRandom.current();

        if (g.getNodeCount() == 0) {
            g.addNode(getRandomVertexLabel(rand));
        }

        for (int i = 0; i < numExtensions; i++) {
            DefaultNode<Integer, Integer> u = g.getNode(rand.nextInt(g.getNodeCount()));
            DefaultNode<Integer, Integer> v = null;

            if (rand.nextDouble() < cycleProbability) {
                ArrayList<DefaultNode<Integer, Integer>> selectableNodes = new ArrayList<>(g.getNodeCount());
                for (DefaultNode<Integer, Integer> n : g.nodes()) {
                    if (n != u && !g.hasEdge(u, n)) {
                        selectableNodes.add(n);
                    }
                }
                if (!selectableNodes.isEmpty()) {
                    v = selectableNodes.get(rand.nextInt(selectableNodes.size()));
                }
            }
            if (v == null) {
                v = g.addNode(getRandomVertexLabel(rand));
            }
            g.addEdge(u, v, getRandomEdgeLabel(rand));

        }
        return g;
    }

    protected int getRandomVertexLabel(Random rand) {
        return Statistical.drawFromDiscreteDistribution(vertexLabelWeights, rand) + 2;
    }

    protected int getRandomEdgeLabel(Random rand) {
        // assure vertex and edge labels are drawn from different sets
        return Statistical.drawFromDiscreteDistribution(edgeLabelWeights, rand) + 2 + vertexLabelCount;
    }

    @Override
    public String getDescription() {
        StringBuilder builder = new StringBuilder();
        builder.append("ConnectedRGG (");

        parameterString(builder);

        builder.append(")");
        return builder.toString();
    }

    protected void parameterString(StringBuilder builder) {
        builder.append("vertexLabelCount: ");
        builder.append(vertexLabelCount);
        builder.append(", ");

        builder.append("edgeLabelCount: ");
        builder.append(edgeLabelCount);
        builder.append(", ");

        builder.append("labelDistribution: ");
        builder.append(labelDistribution.toString());
        builder.append(", ");

        builder.append("sizeDistribution: ");
        builder.append(sizeDistribution);
        builder.append(", ");

        builder.append("cycleProbability: ");
        builder.append(cycleProbability);
    }
}
