package struclust.graph.generators;

import graph.DefaultGraph;
import graph.DefaultNode;
import graph.Graph;

import java.util.ArrayList;
import java.util.Random;

import struclust.util.Statistical;
import cern.jet.random.Exponential;
import cern.jet.random.Poisson;
import cern.jet.random.engine.MersenneTwister;

/**
 * An implementation of the graph generator described in
 * 
 * Frequent Subgraph Discovery, Michihiro Kuramochi and George Karypis, 1st IEEE
 * Conference on Data Mining, pp. 313-320, 2001
 * 
 * XXX: this is not thread-safe
 * 
 * @author Nils Kriege
 * @author Till Schäfer
 */
public class KuramochiKarypisRGG extends ConnectedRGG {

    int avgGraphSize;
    int avgFrqSubgraphSize;
    int frqSubgraphCount;
    int vertexLabelCount;
    int edgeLabelCount;

    // RAND: Random graph generation
    Random rand;
    ArrayList<DefaultGraph<Integer, Integer>> seedPool;

    double[] seedWeights;

    /**
     * Instantiates a new graph generator.
     * 
     * @param avgGraphSize
     *            the average size of the graphs (in terms of the number of
     *            edges)
     * @param avgFrqSubgraphSize
     *            the average size of potentially frequent subgraphs (in terms
     *            of the number of edges)
     * @param frqSubgraphCount
     *            the number of potentially frequent subgraphs
     * @param vertexLabelCount
     *            the number of vertex labels
     * @param edgeLabelCount
     *            the number of edge labels
     */
    public KuramochiKarypisRGG(int avgGraphSize, int avgFrqSubgraphSize, int frqSubgraphCount, int vertexLabelCount,
            int edgeLabelCount) {
        super(vertexLabelCount, edgeLabelCount, new Exponential(1.0, new MersenneTwister()), new Poisson(
                avgFrqSubgraphSize, new MersenneTwister()), 0.1);
        this.avgGraphSize = avgGraphSize;
        this.avgFrqSubgraphSize = avgFrqSubgraphSize;
        this.frqSubgraphCount = frqSubgraphCount;

        this.rand = new Random();

        createSeedPool();
    }

    @Override
    public ArrayList<DefaultGraph<Integer, Integer>> createGraphs(int num) {
        ArrayList<DefaultGraph<Integer, Integer>> graphs = new ArrayList<>(num);

        for (int i = 0; i < num; i++) {
            DefaultGraph<Integer, Integer> g = createGraph(Poisson.staticNextInt(avgGraphSize));
            graphs.add(g);
        }

        return graphs;
    }

    private DefaultGraph<Integer, Integer> createGraph(int size) {
        DefaultGraph<Integer, Integer> g = new DefaultGraph<Integer, Integer>();

        do {
            Graph<Integer, Integer> seed = getRandomSeed();
            int connectionEdgeCount = Poisson.staticNextInt(1.0);
            if (connectionEdgeCount == 0) {
                connectionEdgeCount = 1;
            }
            if (g.getEdgeCount() + connectionEdgeCount < size || rand.nextDouble() < 0.5d) {
                addSeedToGraph(g, seed, connectionEdgeCount);
            }
        } while (g.getEdgeCount() < size);

        return g;
    }

    private void addSeedToGraph(DefaultGraph<Integer, Integer> g, Graph<Integer, Integer> seed, int connectionEdgeCount) {
        int gNodeCount = g.getNodeCount();
        g.addGraph(seed);

        if (gNodeCount == 0)
            return;

        for (int i = 0; i < connectionEdgeCount; i++) {
            DefaultNode<Integer, Integer> u;
            DefaultNode<Integer, Integer> v;
            do {
                u = g.getNode(rand.nextInt(gNodeCount));
                v = g.getNode(rand.nextInt(seed.getNodeCount()) + gNodeCount);
            } while (g.hasEdge(u, v));
            g.addEdge(u, v, getRandomEdgeLabel(rand));
        }
    }

    /**
     * Generates a seed pool with <b>frqSubgraphCount</b> graphs, each of which
     * has a weight derived from an exponential distributed random number. The
     * size of each seed depends on random number obeying the poisson
     * distribution with mean <b>
     */
    private void createSeedPool() {
        this.seedWeights = Statistical.discreteWeightCumulativeDistribution(frqSubgraphCount, super.labelDistribution);
        seedPool = super.createGraphs(frqSubgraphCount);
    }

    /**
     * Returns one of the seeds in the pool. Each seed <b>s</b> is picked with
     * probability depending on its weight.
     */
    protected Graph<Integer, Integer> getRandomSeed() {
        return seedPool.get(Statistical.drawFromDiscreteDistribution(seedWeights, rand));
    }

    @Override
    public String getDescription() {
        StringBuilder builder = new StringBuilder();
        builder.append("KuramochiKarypisRGG (");

        super.parameterString(builder);
        builder.append(", ");

        builder.append("avgGraphSize: ");
        builder.append(avgGraphSize);
        builder.append(", ");

        builder.append("avgFrqSubgraphSize: ");
        builder.append(avgFrqSubgraphSize);
        builder.append(", ");

        builder.append("frqSubgraphCount: ");
        builder.append(frqSubgraphCount);
        builder.append(", ");

        builder.append(")");
        return builder.toString();
    }
}
