package struclust.graph.generators;

import graph.DefaultGraph;
import graph.DefaultNode;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.concurrent.ThreadLocalRandom;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import org.apache.commons.math3.util.Pair;

import struclust.util.Statistical;
import struclust.util.Streams;
import cern.jet.random.AbstractDistribution;

import com.google.common.base.Preconditions;
import com.google.common.collect.HashMultimap;

/**
 * Generates a set of graphs that exhibit a cluster structure. Each single graph
 * is constructed from multiple seed patterns. Some of them come from a cluster
 * specific seed pool and the other come from a shared seed pool. The seeds are
 * combined by adding random edges between them.
 * 
 * <br>
 * Detailed description of the construction process:
 * <ul>
 * <li>For each generated graph a cluster is randomly selected respecting the
 * cluster weights.</li>
 * <li>All cluster specific seed patterns are added to the graph.</li>
 * <li>More seeds patterns are added from the shared seed pool. The seeds are
 * selected randomly with uniform distribution. The number of seed patterns from
 * the shared pool are determined by the usedCommonSeedsDistribution</li>.
 * <li>The disconnected seed patterns are randomly connected by adding edges
 * between their nodes. The number of edges is determined by the
 * connectionCountDistribution.</li>
 * </ul>
 * 
 * @author Till Schäfer
 *
 */
public class SeedCombinationCGG extends ConnectedRGG implements
        ClusteredGraphGenerator<Integer, Integer, DefaultGraph<Integer, Integer>> {

    private ArrayList<ArrayList<DefaultGraph<Integer, Integer>>> clusterSeedPools;
    private ArrayList<DefaultGraph<Integer, Integer>> commonSeedPool;
    private AbstractDistribution usedCommonSeedsDistribution;
    private double[] clusterWeights;
    private AbstractDistribution connectionCountDistribution;
    private int numClusters;
    private AbstractDistribution clusterWeightDistribution;
    private int clusterSeedPoolSize;

    /**
     * Constructor
     * 
     * @param vertexLabelCount
     *            the number of vertex labels. must be larger than 0.
     * @param edgeLabelCount
     *            the number of edge labels. must be larger than 0.
     * @param labelDistribution
     *            the distribution function for the labels
     * @param seedSizeDistribution
     *            the distribution function for the size of the seeds (in terms
     *            of edges)
     * @param seedCycleProbability
     *            the probability for each edge inserted into a seed with more
     *            than two edges to create a cycle. must be in the interval
     *            [0,1].
     * @param numClusters
     *            the number of desired clusters (note: this is only an upper
     *            bound. the bound is tight with a high probability, if the
     *            number of graphs to create is large enough). must be larger
     *            than 0.
     * @param clusterWeightDistribution
     *            the random weight distribution for the cluster assignment. a
     *            graph is created for a cluster regarding to the cluster
     *            weights.
     * @param clusterSeedPoolSize
     *            the number of seeds per cluster
     * @param commonSeedPoolSize
     *            the number of seeds which are shared over all clusters. must
     *            be larger or equal to 0.
     * @param usedCommonSeedsDistribution
     *            distribution to determine the number of seeds used from the
     *            shared seed pool.
     * @param connectionCountDistribution
     *            distribution to determine the number of edged to connect each
     *            pair of seeds.
     */
    // RAND: this whole class is using random distributions all over the place
    public SeedCombinationCGG(int vertexLabelCount, int edgeLabelCount, AbstractDistribution labelDistribution,
            AbstractDistribution seedSizeDistribution, double seedCycleProbability, int numClusters,
            AbstractDistribution clusterWeightDistribution, int clusterSeedPoolSize, int commonSeedPoolSize,
            AbstractDistribution usedCommonSeedsDistribution, AbstractDistribution connectionCountDistribution) {
        super(vertexLabelCount, edgeLabelCount, labelDistribution, seedSizeDistribution, seedCycleProbability);

        Preconditions.checkArgument(numClusters > 0);
        Preconditions.checkNotNull(clusterWeightDistribution);
        Preconditions.checkArgument(clusterSeedPoolSize >= 0);
        Preconditions.checkArgument(commonSeedPoolSize != 0 || usedCommonSeedsDistribution == null);

        this.usedCommonSeedsDistribution = usedCommonSeedsDistribution;
        this.connectionCountDistribution = connectionCountDistribution;

        clusterSeedPools = new ArrayList<>(numClusters);
        for (int i = 0; i < numClusters; i++) {
            clusterSeedPools.add(super.createGraphs(clusterSeedPoolSize));
        }
        commonSeedPool = super.createGraphs(commonSeedPoolSize);

        this.clusterWeights = Statistical.discreteWeightCumulativeDistribution(numClusters, clusterWeightDistribution);

        this.numClusters = numClusters;
        this.clusterWeightDistribution = clusterWeightDistribution;
        this.clusterSeedPoolSize = clusterSeedPoolSize;
    }

    @Override
    public ArrayList<DefaultGraph<Integer, Integer>> createGraphs(int num) {
        ArrayList<DefaultGraph<Integer, Integer>> retVal = new ArrayList<>();
        createGraphsClustered(num).stream().forEach(c -> retVal.addAll(c));
        return retVal;
    }

    @Override
    public ArrayList<List<DefaultGraph<Integer, Integer>>> createGraphsClustered(int num) {
        HashMultimap<Integer, DefaultGraph<Integer, Integer>> clusterIdToGraphs = HashMultimap.create(
                clusterSeedPools.size(), 2 * num / clusterSeedPools.size() + 1);

        // PARALLEL stream: create single graphs
        IntStream.range(0, num).parallel().forEach(i -> createSingleGraph(clusterIdToGraphs));

        ArrayList<List<DefaultGraph<Integer, Integer>>> retVal = new ArrayList<>(clusterIdToGraphs.keySet().size());
        for (Integer key : clusterIdToGraphs.keySet()) {
            retVal.add(new LinkedList<>(clusterIdToGraphs.get(key)));
        }
        return retVal;
    }

    private void createSingleGraph(HashMultimap<Integer, DefaultGraph<Integer, Integer>> clusterIdToGraphs) {
        DefaultGraph<Integer, Integer> graph = new DefaultGraph<>();
        ArrayList<ArrayList<DefaultNode<Integer, Integer>>> seedNodes = new ArrayList<>();

        Random rand = ThreadLocalRandom.current();
        int cluster = Statistical.drawFromDiscreteDistribution(clusterWeights, rand);

        for (DefaultGraph<Integer, Integer> seed : clusterSeedPools.get(cluster)) {
            seedNodes.add(graph.addGraph(seed).getFirst());
        }

        if (commonSeedPool.size() > 0) {
            int numCommonSeeds;
            synchronized (usedCommonSeedsDistribution) {
                numCommonSeeds = usedCommonSeedsDistribution.nextInt();
            }
            for (int i = 0; i < numCommonSeeds; i++) {
                int nextSeedIndex = rand.nextInt(commonSeedPool.size());
                seedNodes.add(graph.addGraph(commonSeedPool.get(nextSeedIndex)).getFirst());
            }
        }

        Streams.orderedPairsStream(seedNodes).forEach(p -> addEdgesBetweenSeeds(p.getFirst(), p.getSecond(), graph));

        synchronized (clusterIdToGraphs) {
            clusterIdToGraphs.put(cluster, graph);
        }
    }

    private void addEdgesBetweenSeeds(ArrayList<DefaultNode<Integer, Integer>> first,
            ArrayList<DefaultNode<Integer, Integer>> second, DefaultGraph<Integer, Integer> graph) {
        int numConnections;
        synchronized (connectionCountDistribution) {
            numConnections = Math.min(connectionCountDistribution.nextInt(), first.size() * second.size());
        }

        ArrayList<Pair<DefaultNode<Integer, Integer>, DefaultNode<Integer, Integer>>> possibleNodePairs = Streams
                .crossProductStream(first, second).collect(Collectors.toCollection(ArrayList::new));

        List<Pair<DefaultNode<Integer, Integer>, DefaultNode<Integer, Integer>>> pairsToAdd = Statistical
                .getSampleWithFloyd(possibleNodePairs, numConnections);

        // RAND: get random edge label
        Random rand = ThreadLocalRandom.current();
        for (Pair<DefaultNode<Integer, Integer>, DefaultNode<Integer, Integer>> pair : pairsToAdd) {
            graph.addEdge(pair.getFirst(), pair.getSecond(), getRandomEdgeLabel(rand));
        }
    }

    @Override
    public String getDescription() {
        StringBuilder builder = new StringBuilder();
        builder.append("SeedCombinationCGG (");

        builder.append("Seeds(");
        super.parameterString(builder);
        builder.append("), ");

        builder.append("numClusters: ");
        builder.append(numClusters);
        builder.append(", ");

        builder.append("clusterWeightDistribution: ");
        builder.append(clusterWeightDistribution);
        builder.append(", ");

        builder.append("clusterSeedPoolSize: ");
        builder.append(clusterSeedPoolSize);
        builder.append(", ");

        builder.append("commonSeedPoolSize: ");
        builder.append(commonSeedPool.size());
        builder.append(", ");

        builder.append("usedCommonSeedsDistribution: ");
        builder.append(usedCommonSeedsDistribution);
        builder.append(", ");

        builder.append("connectionCountDistribution: ");
        builder.append(connectionCountDistribution);

        builder.append(")");
        return builder.toString();
    }

}
