package struclust.graph.generators;

import graph.DefaultGraph;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.ThreadLocalRandom;
import java.util.stream.IntStream;

import struclust.util.Statistical;
import cern.jet.random.AbstractDistribution;

import com.google.common.collect.HashMultimap;

/**
 * {@link RandomGraphGenerator} that creates an initial set of seed patterns and
 * created random extensions from them.
 * 
 * @author Till Schäfer
 *
 */
public class SeedExtensionCGG extends ConnectedRGG implements
        ClusteredGraphGenerator<Integer, Integer, DefaultGraph<Integer, Integer>> {

    private AbstractDistribution extensionSizeDistribution;
    private ArrayList<DefaultGraph<Integer, Integer>> seeds;

    double[] seedWeights;

    /**
     * Constructor
     * 
     * @param vertexLabelCount
     *            the number of vertex labels
     * @param edgeLabelCount
     *            the number of edge labels
     * @param labelDistribution
     *            the distribution of labels (used for random weighting of each
     *            label)
     * @param numSeeds
     *            the number of seed patterns
     * @param seedSizeDistribution
     *            the size distribution for the seeds
     * @param seedSelectionDistribution
     *            the random weight distribution for each seed pattern. a seed
     *            pattern is selected for extension in accordance to its weight.
     * @param extensionSizeDistribution
     *            the size of the extension (randomly selected from this
     *            distribution
     * @param cycleProbability
     *            the probability for each edge inserted into a seed with more
     *            than two edges to create a cycle
     */
    public SeedExtensionCGG(int vertexLabelCount, int edgeLabelCount, AbstractDistribution labelDistribution,
            int numSeeds, AbstractDistribution seedSizeDistribution, AbstractDistribution seedSelectionDistribution,
            AbstractDistribution extensionSizeDistribution, double cycleProbability) {
        super(vertexLabelCount, edgeLabelCount, labelDistribution, seedSizeDistribution, cycleProbability);

        this.extensionSizeDistribution = extensionSizeDistribution;

        seeds = super.createGraphs(numSeeds);
        this.seedWeights = Statistical.discreteWeightCumulativeDistribution(numSeeds, seedSelectionDistribution);
    }

    @Override
    public ArrayList<DefaultGraph<Integer, Integer>> createGraphs(int num) {
        ArrayList<DefaultGraph<Integer, Integer>> retVal = new ArrayList<>();
        createGraphsClustered(num).stream().forEach(c -> retVal.addAll(c));
        return retVal;
    }

    /**
     * Getter for the seed patterns
     * 
     * @return the seed patterns
     */
    public ArrayList<DefaultGraph<Integer, Integer>> getSeeds() {
        return seeds;
    }

    private void createSingleExtendedGraph(
            HashMultimap<DefaultGraph<Integer, Integer>, DefaultGraph<Integer, Integer>> extendedPatterns) {
        int randSeedIndex = Statistical.drawFromDiscreteDistribution(seedWeights, ThreadLocalRandom.current());
        DefaultGraph<Integer, Integer> seed = seeds.get(randSeedIndex);
        DefaultGraph<Integer, Integer> extendedGraph = seed.clone();
        extendSingleConnectedGraph(extensionSizeDistribution.nextInt(), extendedGraph);
        synchronized (extendedPatterns) {
            extendedPatterns.put(seed, extendedGraph);
        }
    }

    @Override
    public ArrayList<List<DefaultGraph<Integer, Integer>>> createGraphsClustered(int num) {

        HashMultimap<DefaultGraph<Integer, Integer>, DefaultGraph<Integer, Integer>> extendedPatterns = HashMultimap
                .create(seeds.size(), num / seeds.size() + 1);
        
        // PARALLEL stream: create single graphs
        IntStream.range(0, num).parallel().forEach(i -> createSingleExtendedGraph(extendedPatterns));
        
        ArrayList<List<DefaultGraph<Integer, Integer>>> retVal = new ArrayList<>(extendedPatterns.keySet().size());
        for (DefaultGraph<Integer, Integer> key : extendedPatterns.keySet()) {
            retVal.add(new LinkedList<>(extendedPatterns.get(key)));
        }
        return retVal;
    }

    @Override
    public String getDescription() {
        StringBuilder builder = new StringBuilder();
        builder.append("ConnectedRGG (");

        builder.append("Seeds(");
        super.parameterString(builder);
        builder.append("), ");

        builder.append("numSeeds: ");
        builder.append(seeds.size());
        builder.append(", ");

        builder.append("extensionSizeDistribution: ");
        builder.append(extensionSizeDistribution);

        builder.append(")");
        return builder.toString();
    }
}
