package struclust.mcs;

import graph.Edge;
import graph.Graph;
import graph.Node;

import java.util.ArrayList;
import java.util.List;

/**
 * Labeled Association Graph between two DefaultGraphs.
 * 
 * TODO: are labels really required here?
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * 
 */
public class MatrixAssociationGraph<NL, EL> {

    /**
     * indices of g1 for each node in the association graph
     */
    private ArrayList<Integer> nodes1 = new ArrayList<>();
    /**
     * indices of g2 for each node in the association graph
     */
    private ArrayList<Integer> nodes2 = new ArrayList<>();
    /**
     * <pre>
     *  0: no edge
     *  1: c-edge
     * -1: d-edge
     * </pre>
     */
    private int adj[][];

    /**
     * Constructor
     * 
     * @param g1
     *            the first {@link Graph}
     * @param g2
     *            the second {@link Graph}
     */
    public MatrixAssociationGraph(Graph<NL, EL> g1, Graph<NL, EL> g2) {
        for (Node<NL, EL> n1 : g1.nodes()) {
            for (Node<NL, EL> n2 : g2.nodes()) {
                if (n1.getLabel().equals(n2.getLabel())) {
                    nodes1.add(n1.getIndex());
                    nodes2.add(n2.getIndex());
                }
            }
        }

        adj = new int[nodes1.size()][nodes1.size()];

        for (int indexA = 0; indexA < nodes1.size(); indexA++) {
            for (int indexB = 0; indexB < nodes1.size(); indexB++) {
                // PERF: use symetry?
                // if (indexA == indexB) {
                // break;
                // }

                boolean selfMapping1 = nodes1.get(indexA) == nodes1.get(indexB);
                boolean selfMapping2 = nodes2.get(indexA) == nodes2.get(indexB);
                if (indexA != indexB && !selfMapping1 && !selfMapping2) {
                    Edge<NL, EL> g1Edge = g1.getEdge(g1.getNode(nodes1.get(indexA)), g1.getNode(nodes1.get(indexB)));
                    Edge<NL, EL> g2Edge = g2.getEdge(g2.getNode(nodes2.get(indexA)), g2.getNode(nodes2.get(indexB)));
                    if (g1Edge == null && g2Edge == null) {
                        adj[indexA][indexB] = -1;
                    } else if (g1Edge != null && g2Edge != null && g1Edge.getLabel().equals(g2Edge.getLabel())) {
                        adj[indexA][indexB] = 1;
                    }
                }
            }
        }
    }

    /**
     * @return the number of nodes
     */
    public int getVertexCount() {
        return nodes1.size();
    }

    /**
     * Return the internal stored matrix
     * 
     * meaning of entries:
     * 
     * <pre>
     *  0: no edge
     *  1: c-edge
     * -1: d-edge
     * </pre>
     * 
     * @return adjacency matrix
     */
    public int[][] getAdjacencies() {
        return adj;
    }

    /**
     * Return a copied matrix of adjacencies with no distinction between c and d
     * edges
     * 
     * meaning of entries:
     * 
     * <pre>
     *  0: no edge
     *  1: edge
     * </pre>
     * 
     * @return adjacency matrix
     */
    public int[][] getSimplifiedAdjacencies() {
        int n = getVertexCount();
        int[][] retVal = new int[n][n];

        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                retVal[i][j] = Math.abs(adj[i][j]);
            }
        }

        return retVal;
    }

    /**
     * Calculate the degrees of each vertex
     * 
     * @return the vertex degrees
     */
    public int[] getDegrees() {
        int n = getVertexCount();
        int[] degrees = new int[n];

        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                degrees[i] += Math.abs(adj[i][j]);
            }
        }

        return degrees;
    }

    /**
     * Returns the associated node indices in g1 for a List of indices of the
     * AssociationGraph
     * 
     * @param assIndices
     *            the association graph node indices
     * @return the node indices of g1
     */
    public ArrayList<Integer> getGraph1Indices(List<Integer> assIndices) {
        ArrayList<Integer> retVal = new ArrayList<>(assIndices.size());

        for (Integer assIndex : assIndices) {
            retVal.add(nodes1.get(assIndex));
        }

        return retVal;
    }

    /**
     * Returns the associated node indices in g2 for a List of indices of the
     * AssociationGraph
     * 
     * @param assIndices
     *            the association graph node indices
     * @return the node indices of g2
     */
    public ArrayList<Integer> getGraph2Indices(List<Integer> assIndices) {
        ArrayList<Integer> retVal = new ArrayList<>(assIndices.size());

        for (Integer assIndex : assIndices) {
            retVal.add(nodes2.get(assIndex));
        }

        return retVal;
    }

    @Override
    public String toString() {
        StringBuilder str = new StringBuilder();
        str.append("        ");
        for (int i = 0; i < nodes1.size(); i++) {
            str.append(String.format("%03d/%03d ", nodes1.get(i), nodes2.get(i)));
        }
        str.append(System.getProperty("line.separator"));
        for (int i1 = 0; i1 < nodes1.size(); i1++) {
            str.append(String.format("%03d/%03d ", nodes1.get(i1), nodes2.get(i1)));
            for (int i2 = 0; i2 < nodes2.size(); i2++) {
                str.append(String.format("%7s ", adj[i1][i2]));
            }
            str.append(System.getProperty("line.separator"));
        }

        return str.toString();
    }
}
