package struclust.mcs;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;

import graph.Graph;
import struclust.datastructures.CountingInt;

/**
 * Compute whether there exists a MCCIS of size minSize or not. This is the
 * decision variant of {@link MccisKriegeColor2MSA}.
 * 
 * @author Nils Kriege
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class MccisKriegeColor2MSDA<NL, EL, G extends Graph<NL, EL>> implements McsSizeDecisionAlgorithm<NL, EL, G> {
    /**
     * Performance counter for calls of extedCandFast
     */
    public static int calls = 0;

    @Override
    public boolean calc(G graph1, G graph2, int minSize) {
        MatrixAssociationGraph<NL, EL> mag = new MatrixAssociationGraph<>(graph1, graph2);

        CountingInt maxSize = new CountingInt(Math.max(minSize, 0));

        int n = mag.getVertexCount();
        int[] degrees = mag.getDegrees();
        int[][] adj = mag.getAdjacencies();

        LinkedList<Integer> p = new LinkedList<>();
        for (int i = 0; i < n; i++)
            p.add(i);

        int[] colorClass = new int[p.size()];
        ArrayList<Integer> classCount = new ArrayList<>();
        greedycoloring(p, adj, degrees, colorClass, classCount);
        int colors = classCount.size();

        int[] d = new int[n];
        int i = 0;
        for (int v : p) {
            d[i++] = v;
        }

        boolean existMcsWithMinSize = extendCandFastInit(maxSize, new LinkedList<Integer>(), d, adj, degrees,
                colorClass, classCount, colors, minSize);
        return existMcsWithMinSize;
    }

    /**
     * s
     * 
     * <pre>
     *   +-----------------------------+----------------+-----+
     * d |                             | d-candidates   |     |
     *   +-----------------------------+----------------+-----+
     *                              lBound           uBound
     * </pre>
     */
    private boolean extendCandFast(CountingInt maxSize, LinkedList<Integer> clique, LinkedList<Integer> p, int[] d,
            int lBound, int uBound, int[][] adj, int[] degrees, int[] colorClass, ArrayList<Integer> classCount,
            int colors, int minSize) {
        calls++;

        Collections.sort(p, (Integer i1, Integer i2) -> Integer.compare(classCount.get(colorClass[i1]),
                classCount.get(colorClass[i2])));

        while (!p.isEmpty()) {
            if (clique.size() + colors < minSize) {
                return false;
            }

            int i = p.remove(p.size() - 1);
            clique.add(i);
            colors = removeFromColoring(colorClass, classCount, colors, i);
            int[] edgeType = adj[i];

            if (clique.size() >= minSize) {
                return true;
            }

            ArrayList<Integer> newClassCount = new ArrayList<>(classCount);
            int newColors = colors;

            // prepare candidate set for recursive call
            LinkedList<Integer> newP = new LinkedList<>();
            for (Integer v : p) {
                if (edgeType[v] != 0)
                    newP.add(v);
                else
                    newColors = removeFromColoring(colorClass, newClassCount, newColors, v);
            }

            int newUBound = uBound;
            int newLBound = lBound;
            if (lBound <= uBound) {
                while (edgeType[d[newUBound]] == 0 && newUBound > lBound) {
                    newColors = removeFromColoring(colorClass, newClassCount, newColors, d[newUBound]);
                    newUBound--;
                }
                while (newLBound <= newUBound && edgeType[d[newLBound]] == 0) {
                    newColors = removeFromColoring(colorClass, newClassCount, newColors, d[newLBound]);
                    newLBound++;
                }
                int nm = newLBound - 1;
                while (++nm <= newUBound) {
                    int v = d[nm];
                    if (edgeType[v] < 0)
                        continue;
                    else if (edgeType[v] > 0)
                        newP.add(v);
                    else
                        newColors = removeFromColoring(colorClass, newClassCount, newColors, v);
                    // swap if edgeType >= 0
                    swap(d, newLBound, nm);
                    newLBound++;
                }
            }

            assert newP.size() + dSize(newLBound, newUBound) >= newColors;
            if (!newP.isEmpty()) {
                boolean existMcsWithMinSize = extendCandFast(maxSize, clique, newP, d, newLBound, newUBound, adj,
                        degrees, colorClass, newClassCount, newColors, minSize);
                if (existMcsWithMinSize) {
                    return true;
                }
            }
            clique.removeLast();
        }

        // reachable only if p is empty
        return false;
    }

    private int removeFromColoring(int[] colorClass, ArrayList<Integer> classCount, int colors, int vertice) {
        int cClassOfVertice = colorClass[vertice];
        Integer numSameClass = classCount.get(cClassOfVertice);
        assert numSameClass > 0;
        if (numSameClass == 1) {
            classCount.set(cClassOfVertice, 0);
            colors--;
        } else {
            classCount.set(cClassOfVertice, numSameClass - 1);
        }
        return colors;
    }

    /**
     * <pre>
     *   +-----------------------------+----------------+-----+
     * d |                             | d-candidates   |     |
     *   +-----------------------------+----------------+-----+
     *                              lBound           uBound
     * </pre>
     * 
     * @param classCount
     * @param colorClass
     * @param colors
     * @param minSize
     */
    private boolean extendCandFastInit(CountingInt maxSize, LinkedList<Integer> clique, int[] d, int[][] adj,
            int[] degrees, int[] colorClass, ArrayList<Integer> classCount, int colors, int minSize) {
        calls++;

        final int lBound = 0;
        final int uBound = d.length - 1;

        if (lBound <= uBound) {
            maxSize.increment();
        }

        for (int it = lBound; it <= uBound; it++) {
            if (clique.size() + colors < minSize) {
                return false;
            }

            int i = d[it];
            colors = removeFromColoring(colorClass, classCount, colors, i);
            clique.add(i);

            if (clique.size() >= minSize) {
                return true;
            }

            // prepare candidate set for recursive call
            LinkedList<Integer> newP = new LinkedList<>();

            int[] edgeType = adj[i];
            ArrayList<Integer> newClassCount = new ArrayList<>(classCount);
            int newColors = colors;
            int newUBound = uBound;
            int newLBound = it + 1;
            while (newUBound >= newLBound && edgeType[d[newUBound]] == 0) {
                newColors = removeFromColoring(colorClass, newClassCount, newColors, d[newUBound]);
                newUBound--;
            }
            while (newLBound <= newUBound && edgeType[d[newLBound]] == 0) {
                newColors = removeFromColoring(colorClass, newClassCount, newColors, d[newLBound]);
                newLBound++;
            }
            int nm = newLBound - 1;
            while (++nm <= newUBound) {
                int v = d[nm];
                if (edgeType[v] < 0)
                    continue;
                if (edgeType[v] > 0)
                    newP.add(v);
                else
                    newColors = removeFromColoring(colorClass, newClassCount, newColors, v);
                swap(d, newLBound, nm);
                newLBound++;
            }

            assert newP.size() + dSize(newLBound, newUBound) >= newColors;
            if (!newP.isEmpty()) {
                boolean existMcsWithMinSize = extendCandFast(maxSize, clique, newP, d, newLBound, newUBound, adj,
                        degrees, colorClass, newClassCount, newColors, minSize);
                if (existMcsWithMinSize) {
                    return true;
                }
            }
            clique.removeLast();
        }

        // reachable only if mag is empty
        return minSize == 0;
    }

    private void greedycoloring(LinkedList<Integer> P, int[][] adj, int[] degrees, int[] colorclass,
            List<Integer> classcount) {
        int n = P.size();

        minWidthOrder(P, degrees, adj);

        @SuppressWarnings("unchecked")
        List<Integer>[] colourClass = new LinkedList[n];
        for (int i = 0; i < n; i++)
            colourClass[i] = new LinkedList<>();

        for (Integer v : P) {
            int k = 0;
            while (conflicts(v, colourClass[k], adj))
                k++;
            colourClass[k].add(v);
            colorclass[v] = k;
        }

        for (List<Integer> cClass : colourClass) {
            if (cClass.size() == 0) {
                break;
            }
            classcount.add(cClass.size());
        }
    }

    private void minWidthOrder(LinkedList<Integer> vertices, int[] degrees, int[][] adj) {
        int[] deg = new int[degrees.length];
        System.arraycopy(degrees, 0, deg, 0, degrees.length);
        Stack<Integer> S = new Stack<>();
        while (!vertices.isEmpty()) {
            Integer v = vertices.getFirst();
            for (Integer u : vertices)
                if (deg[u] < deg[v])
                    v = u;
            S.push(v);
            vertices.remove(v);
            for (Integer u : vertices)
                if (adj[v][u] != 0)
                    deg[u]--;
        }
        while (!S.isEmpty())
            vertices.addFirst(S.pop());
    }

    private boolean conflicts(int v, List<Integer> colourClass, int[][] adj) {
        for (int i = 0; i < colourClass.size(); i++) {
            int w = colourClass.get(i);
            if (adj[v][w] != 1)
                return true;
        }
        return false;
    }

    private int dSize(int lBound, int uBound) {
        return Math.max(uBound - lBound + 1, 0);
    }

    private void swap(int[] d, int newLBound, int nm) {
        int tmp;
        tmp = d[newLBound];
        d[newLBound] = d[nm];
        d[nm] = tmp;
    }

    @Override
    public String getDescription() {
        return "MccisKriegeColor2MSDA";
    }

    @Override
    public boolean isConnected() {
        return true;
    }

    @Override
    public boolean isInduced() {
        return true;
    }

}
