package struclust.mcs;

import java.util.LinkedList;
import java.util.List;
import java.util.Stack;

import graph.Graph;
import struclust.datastructures.CountingInt;

/**
 * Compute the maximum common connected induced subgraph size for two graphs.
 * 
 * Variant of {@link MccisKriegeMSA}, that uses a greedy coloring for stricter
 * pruning.
 * 
 * @author Nils Kriege
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class MccisKriegeColorMSA<NL, EL, G extends Graph<NL, EL>> implements McsSizeAlgorithm<NL, EL, G> {
    public static int calls = 0;

    @Override
    public int calc(G graph1, G graph2) {
        MatrixAssociationGraph<NL, EL> mag = new MatrixAssociationGraph<>(graph1, graph2);

        CountingInt maxSize = new CountingInt(0);

        int n = mag.getVertexCount();
        int[] degrees = mag.getDegrees();
        int[][] adj = mag.getAdjacencies();

        LinkedList<Integer> p = new LinkedList<>();
        for (int i = 0; i < n; i++)
            p.add(i);
        minWidthOrder(p, degrees, adj);
        int[] d = new int[n];
        int i = 0;
        for (int v : p) {
            d[i++] = v;
        }

        extendCandFastInit(maxSize, new LinkedList<Integer>(), d, adj, degrees);

        return maxSize.get();
    }

    /**
     * s
     * 
     * <pre>
     *   +-----------------------------+----------------+-----+
     * d |                             | d-candidates   |     |
     *   +-----------------------------+----------------+-----+
     *                              lBound           uBound
     * </pre>
     */
    private void extendCandFast(CountingInt maxSize, LinkedList<Integer> clique, LinkedList<Integer> p, int[] d,
            int lBound, int uBound, int[][] adj, int[] degrees) {
        calls++;
        minWidthOrder(p, degrees, adj);

        while (!p.isEmpty()) {
            int i = p.remove(p.size() - 1);

            int[] edgeType = adj[i];

            clique.add(i);

            if (clique.size() > maxSize.get()) {
                maxSize.increment();
            }

            // prepare candidate set for recursive call
            LinkedList<Integer> newP = new LinkedList<>();
            for (Integer v : p) {
                if (edgeType[v] != 0)
                    newP.add(v);
            }

            int newUBound = uBound;
            int newLBound = lBound;
            if (lBound <= uBound) {
                while (edgeType[d[newUBound]] == 0 && --newUBound > lBound) {}
                while (edgeType[d[newLBound]] == 0 && ++newLBound < newUBound) {}
                int nm = newLBound - 1;
                while (++nm <= newUBound) {
                    int v = d[nm];
                    if (edgeType[v] < 0)
                        continue;
                    if (edgeType[v] > 0)
                        newP.add(v);
                    // swap if edgeType >= 0
                    swap(d, newLBound, nm);
                    newLBound++;
                }
            }
            if (clique.size() + greedycolors(newP, d, adj, newLBound, newUBound, degrees) > maxSize.get()) {
                extendCandFast(maxSize, clique, newP, d, newLBound, newUBound, adj, degrees);
            }
            clique.removeLast();
        }
    }

    /**
     * <pre>
     *   +-----------------------------+----------------+-----+
     * d |                             | d-candidates   |     |
     *   +-----------------------------+----------------+-----+
     *                              lBound           uBound
     * </pre>
     */
    private void extendCandFastInit(CountingInt maxSize, LinkedList<Integer> clique, int[] d, int[][] adj, int[] degrees) {
        calls++;

        final int lBound = 0;
        final int uBound = d.length - 1;
        
        if (lBound <= uBound) {
            maxSize.increment();
        }
        
        for (int it = lBound; it <= uBound; it++) {
            int i = d[it];

            clique.add(i);

            // prepare candidate set for recursive call
            LinkedList<Integer> newP = new LinkedList<>();

            int newUBound = uBound;
            int[] edgeType = adj[i];
            while (edgeType[d[newUBound]] == 0 && --newUBound > it) {}
            int newLBound = it;
            while (++newLBound <= newUBound && edgeType[d[newLBound]] == 0) {}
            int nm = newLBound - 1;
            while (++nm <= newUBound) {
                int v = d[nm];
                if (edgeType[v] < 0)
                    continue;
                if (edgeType[v] > 0)
                    newP.add(v);
                swap(d, newLBound, nm);
                newLBound++;
            }

            if (clique.size() + greedycolors(newP, d, adj, newLBound, newUBound, degrees) > maxSize.get()) {
                extendCandFast(maxSize, clique, newP, d, newLBound, newUBound, adj, degrees);
            }
            clique.removeLast();
        }
    }

    private int dSize(int lBound, int uBound) {
        return Math.max(uBound - lBound + 1, 0);
    }

    private int greedycolors(LinkedList<Integer> P, int[] d, int[][] adj, int lBound, int uBound, int[] degrees) {
        int dSize = dSize(lBound, uBound);
        int n = P.size() + dSize;

         minWidthOrder(P, degrees, adj);

        @SuppressWarnings("unchecked")
        List<Integer>[] colourClass = new LinkedList[n];
        for (int i = 0; i < n; i++)
            colourClass[i] = new LinkedList<>();

        int colours = 0;
        for (int i = lBound; i <= uBound; i++) {
            int v = d[i];
            int k = 0;
            while (conflicts(v, colourClass[k], adj))
                k++;
            colourClass[k].add(v);
            colours = Math.max(colours, k + 1);
        }

        for (Integer v : P) {
            int k = 0;
            while (conflicts(v, colourClass[k], adj))
                k++;
            colourClass[k].add(v);
            colours = Math.max(colours, k + 1);
        }
        return colours;
    }

    private void minWidthOrder(LinkedList<Integer> vertices, int[] degrees, int[][] adj) {
        int[] deg = new int[degrees.length];
        System.arraycopy(degrees, 0, deg, 0, degrees.length);
        Stack<Integer> S = new Stack<>();
        while (!vertices.isEmpty()) {
            Integer v = vertices.getFirst();
            for (Integer u : vertices)
                if (deg[u] < deg[v])
                    v = u;
            S.push(v);
            vertices.remove(v);
            for (Integer u : vertices)
                if (adj[v][u] != 0)
                    deg[u]--;
        }
        while (!S.isEmpty())
            vertices.addFirst(S.pop());
    }

    private boolean conflicts(int v, List<Integer> colourClass, int[][] adj) {
        for (int i = 0; i < colourClass.size(); i++) {
            int w = colourClass.get(i);
            if (adj[v][w] != 1)
                return true;
        }
        return false;
    }

    private void swap(int[] d, int newLBound, int nm) {
        int tmp;
        tmp = d[newLBound];
        d[newLBound] = d[nm];
        d[nm] = tmp;
    }

    @Override
    public String getDescription() {
        return "MccisKriegeColorMSA";
    }

    @Override
    public boolean isConnected() {
        return true;
    }

    @Override
    public boolean isInduced() {
        return true;
    }

}
