package struclust.mcs;

import graph.Graph;
import graph.GraphFactory;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import struclust.datastructures.CountingInt;
import struclust.graph.Graphs;

/**
 * Compute the maximum common connected induced subgraphs for two graphs
 * 
 * @author Nils Kriege
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class MccisKriegeMEA<NL, EL, G extends Graph<NL, EL>> implements McsEnumAlgorithm<NL, EL, G> {

    private GraphFactory<NL, EL, G> gf;

    /**
     * Constructor
     * 
     * @param gf
     *            a GraphFactory for G
     */
    public MccisKriegeMEA(GraphFactory<NL, EL, G> gf) {
        this.gf = gf;
    }

    @Override
    public List<G> calc(G graph1, G graph2) {
        return calc(graph1, graph2, 0);
    }

    /**
     * Calculate the size (in terms of nodes + edges) of the maximum common
     * subgraph (MCS).
     * 
     * @param graph1
     *            the first {@link Graph}
     * @param graph2
     *            the second {@link Graph}
     * @param minSize
     *            performance optimization: if one is interested in mcs of a
     *            minimum size only, one can specify a percentage value here.
     *            the parameter can be used for further pruning. the pruning
     *            becomes especially effective if minsize is a large value. the
     *            minimal size of a mcs (the algorithm will return an empty set
     *            of mcs if the minsize is not reached)
     * @return the size of the MCS
     */
    public List<G> calc(G graph1, G graph2, int minSize) {
        MatrixAssociationGraph<NL, EL> assG = new MatrixAssociationGraph<>(graph1, graph2);

        CountingInt maxSize = new CountingInt(Math.max(minSize, 1));
        LinkedList<LinkedList<Integer>> maxCliques = new LinkedList<>();

        int n = assG.getVertexCount();
        int[] p = new int[n];
        for (int i = 0; i < n; i++)
            p[i] = i;
        // initialization
        extendCandFastInit(maxSize, maxCliques, new LinkedList<Integer>(), p, assG);

        LinkedList<G> retVal = new LinkedList<>();
        for (LinkedList<Integer> indices : maxCliques) {
            retVal.add(Graphs.subgraph(graph1, assG.getGraph1Indices(indices), gf));

            assert indices.size() == maxCliques.get(0).size(); // same size
        }

        Graphs.distinctGraphs(retVal);
        return retVal;
    }

    /**
     * s
     * 
     * <pre>
     *   +-----------------------------+----------------+-----+
     * d |                             | d-candidates   |     |
     *   +-----------------------------+----------------+-----+
     *                              lBound           uBound
     * </pre>
     */
    private void extendCandFast(CountingInt maxSize, LinkedList<LinkedList<Integer>> maxCliques,
            LinkedList<Integer> clique, ArrayList<Integer> p, int[] d, int lBound, int uBound,
            MatrixAssociationGraph<NL, EL> assG) {

        while (!p.isEmpty()) {
            int i = p.remove(p.size() - 1);

            int[] edgeType = assG.getAdjacencies()[i];

            clique.add(i);

            if (clique.size() > maxSize.get()) {
                maxSize.increment();
                maxCliques.clear();
                maxCliques.add(new LinkedList<>(clique));
            } else if (clique.size() == maxSize.get()) {
                maxCliques.add(new LinkedList<>(clique));
            }

            // prepare candidate set for recursive call
            ArrayList<Integer> newP = new ArrayList<Integer>();
            for (Integer v : p) {
                if (edgeType[v] != 0)
                    newP.add(v);
            }

            int newUBound = uBound;
            int newLBound = lBound;
            if (lBound <= uBound) {
                while (edgeType[d[newUBound]] == 0 && --newUBound > lBound) {}
                while (edgeType[d[newLBound]] == 0 && ++newLBound < newUBound) {}
                int nm = newLBound - 1;
                while (++nm <= newUBound) {
                    if (edgeType[d[nm]] < 0)
                        continue;
                    if (edgeType[d[nm]] > 0)
                        newP.add(d[nm]);
                    // swap
                    swap(d, newLBound, nm);
                    newLBound++;
                }
            }

            int dSize = Math.max(newUBound - newLBound + 1, 0);
            if (clique.size() + newP.size() + dSize >= maxSize.get()) {
                extendCandFast(maxSize, maxCliques, clique, newP, d, newLBound, newUBound, assG);
            }
            clique.removeLast();
        }
    }

    /**
     * <pre>
     *   +-----------------------------+----------------+-----+
     * d |                             | d-candidates   |     |
     *   +-----------------------------+----------------+-----+
     *                              lBound           uBound
     * </pre>
     */
    private void extendCandFastInit(CountingInt maxSize, LinkedList<LinkedList<Integer>> maxCliques,
            LinkedList<Integer> clique, int[] d, MatrixAssociationGraph<NL, EL> assG) {

        final int lBound = 0;
        final int uBound = d.length - 1;

        for (int it = lBound; it <= uBound; it++) {
            int i = d[it];

            clique.add(i);

            if (maxSize.get() == 1) {
                maxCliques.add(new LinkedList<>(clique));
            }

            // prepare candidate set for recursive call
            ArrayList<Integer> p = new ArrayList<Integer>();

            int newUBound = uBound;
            int[] edgeType = assG.getAdjacencies()[i];
            while (edgeType[d[newUBound]] == 0 && --newUBound > it) {}
            int newLBound = it;
            while (++newLBound <= newUBound && edgeType[d[newLBound]] == 0) {}
            int nm = newLBound - 1;
            while (++nm <= newUBound) {
                if (edgeType[d[nm]] < 0)
                    continue;
                if (edgeType[d[nm]] > 0)
                    p.add(d[nm]);
                swap(d, newLBound, nm);
                newLBound++;
            }

            int dSize = Math.max(newUBound - newLBound + 1, 0);
            if (clique.size() + p.size() + dSize >= maxSize.get()) {
                extendCandFast(maxSize, maxCliques, clique, p, d, newLBound, newUBound, assG);
            }
            clique.removeLast();
        }
    }

    private void swap(int[] d, int newLBound, int nm) {
        int tmp;
        tmp = d[newLBound];
        d[newLBound] = d[nm];
        d[nm] = tmp;
    }

    @Override
    public String getDescription() {
        return "MccisKriegeMEA";
    }

    @Override
    public boolean isConnected() {
        return true;
    }

    @Override
    public boolean isInduced() {
        return true;
    }

}
