package struclust.mcs;

import java.util.ArrayList;
import java.util.LinkedList;

import graph.Graph;
import struclust.datastructures.CountingInt;

/**
 * Compute the maximum common connected induced subgraph size for two graphs.
 * 
 * This is basically the MaximumSubgraphamtchingKernel, but for maxCliques.
 * 
 * @author Nils Kriege
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class MccisKriegeMSA<NL, EL, G extends Graph<NL, EL>> implements McsSizeAlgorithm<NL, EL, G> {
    public static int calls = 0;
    
    @Override
    public int calc(G graph1, G graph2) {
        MatrixAssociationGraph<NL, EL> mag = new MatrixAssociationGraph<>(graph1, graph2);

        CountingInt maxSize = new CountingInt(0);

        int n = mag.getVertexCount();
        int[] p = new int[n];
        for (int i = 0; i < n; i++)
            p[i] = i;
        // initialization
        extendCandFastInit(maxSize, new LinkedList<Integer>(), p, mag.getAdjacencies());

        return maxSize.get();
    }

    /**
     * s
     * 
     * <pre>
     *   +-----------------------------+----------------+-----+
     * d |                             | d-candidates   |     |
     *   +-----------------------------+----------------+-----+
     *                              lBound           uBound
     * </pre>
     */
    private void extendCandFast(CountingInt maxSize, LinkedList<Integer> clique, ArrayList<Integer> p, int[] d,
            int lBound, int uBound, int[][] adj) {
        calls++;
        while (!p.isEmpty()) {
            int i = p.remove(p.size() - 1);

            int[] edgeType = adj[i];

            clique.add(i);

            if (clique.size() > maxSize.get()) {
                maxSize.increment();
            }

            // prepare candidate set for recursive call
            ArrayList<Integer> newP = new ArrayList<>();
            for (Integer v : p) {
                if (edgeType[v] != 0)
                    newP.add(v);
            }

            int newUBound = uBound;
            int newLBound = lBound;
            if (lBound <= uBound) {
                while (edgeType[d[newUBound]] == 0 && --newUBound > lBound) {}
                while (edgeType[d[newLBound]] == 0 && ++newLBound < newUBound) {}
                int nm = newLBound - 1;
                while (++nm <= newUBound) {
                    int v = d[nm];
                    if (edgeType[v] < 0)
                        continue;
                    if (edgeType[v] > 0)
                        newP.add(v);
                    // swap if edgeType >= 0
                    swap(d, newLBound, nm);
                    newLBound++;
                }
            }
            int dSize = Math.max(newUBound - newLBound + 1, 0);
            if (clique.size() + newP.size() + dSize > maxSize.get()) {
                extendCandFast(maxSize, clique, newP, d, newLBound, newUBound, adj);
            }
            clique.removeLast();
        }
    }

    /**
     * <pre>
     *   +-----------------------------+----------------+-----+
     * d |                             | d-candidates   |     |
     *   +-----------------------------+----------------+-----+
     *                              lBound           uBound
     * </pre>
     */
    private void extendCandFastInit(CountingInt maxSize, LinkedList<Integer> clique, int[] d, int[][] adj) {
        calls++;
        final int lBound = 0;
        final int uBound = d.length - 1;
        
        if (lBound <= uBound) {
            maxSize.increment();
        }

        for (int it = lBound; it <= uBound; it++) {
            int i = d[it];

            clique.add(i);

            // prepare candidate set for recursive call
            ArrayList<Integer> p = new ArrayList<>();

            int newUBound = uBound;
            int[] edgeType = adj[i];
            while (edgeType[d[newUBound]] == 0 && --newUBound > it) {}
            int newLBound = it;
            while (++newLBound <= newUBound && edgeType[d[newLBound]] == 0) {}
            int nm = newLBound - 1;
            while (++nm <= newUBound) {
                int v = d[nm];
                if (edgeType[v] < 0)
                    continue;
                if (edgeType[v] > 0)
                    p.add(v);
                swap(d, newLBound, nm);
                newLBound++;
            }

            int dSize = Math.max(newUBound - newLBound + 1, 0);
            if (clique.size() + p.size() + dSize > maxSize.get()) {
                extendCandFast(maxSize, clique, p, d, newLBound, newUBound, adj);
            }
            clique.removeLast();
        }
    }

    private void swap(int[] d, int newLBound, int nm) {
        int tmp;
        tmp = d[newLBound];
        d[newLBound] = d[nm];
        d[nm] = tmp;
    }

    @Override
    public String getDescription() {
        return "MccisKriegeMSA";
    }

    @Override
    public boolean isConnected() {
        return true;
    }

    @Override
    public boolean isInduced() {
        return true;
    }

}
