package struclust.mcs;

import graph.Graph;
import graph.GraphFactory;
import graph.GraphSize;

import java.util.ArrayList;
import java.util.List;

import struclust.datastructures.MutableInt;
import struclust.graph.GraphContainer;
import struclust.mining.RandomGM;
import struclust.mining.RandomGM.RandomGMConf;
import struclust.mining.SimpleSC;
import struclust.util.Collections;

/**
 * Computes a heuristic maximum connected (not induced) subgraph.
 * 
 * Uses the {@link RandomGM} for heuristic MCS detection. It simply mines
 * numRandGraph many frequent maximal subgraphs with support 1 and chooses the
 * largest one as heuristic MCS.
 * 
 * @author Till Schäfer
 * 
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class MccsRandMineMSA<NL, EL, G extends Graph<NL, EL>> implements McsSizeAlgorithm<NL, EL, G> {

    private int numRandGraphs;
    private RandomGMConf<NL, EL, G> gmConf;
    private GraphSize s;

    /**
     * Constructor
     * 
     * @param parallelism
     *            the parallelism (how many threads are generated for each
     *            thread pool)
     * @param gFactory
     *            the {@link GraphFactory}
     * @param numRandGraphs
     *            the number of maximal frequent subgraphs that are randomly
     *            mined for each MCS computation
     * @param s
     *            way the graph size is measured
     */
    public MccsRandMineMSA(MutableInt parallelism, GraphFactory<NL, EL, G> gFactory, int numRandGraphs,
            GraphSize s) {
        gmConf = new RandomGMConf<>(gFactory, parallelism);
        gmConf.parallel = true;
        gmConf.supportCounting = new SimpleSC<>(parallelism);
        gmConf.useSupSizeProduct = false;
        gmConf.gSize = s;
        this.numRandGraphs = numRandGraphs;
        this.s = s;
    }

    @Override
    public int calc(G graph1, G graph2) {
        RandomGM<NL, EL, G> miner = new RandomGM<>(gmConf);
        List<GraphContainer<NL, EL, G>> graphs = Collections.asLinkedList(new GraphContainer<>(graph1),
                new GraphContainer<>(graph2));
        ArrayList<GraphContainer<NL, EL, G>> minedGraphs = miner.mineGraphs(graphs, numRandGraphs, 1.0);

        return minedGraphs.stream().mapToInt((graph) -> graph.size(s)).max().orElse(0);
    }

    @Override
    public String getDescription() {
        return "RandMineMcsMSA(numRandGraphs: " + numRandGraphs + ")";
    }

    @Override
    public boolean isConnected() {
        return true;
    }

    @Override
    public boolean isInduced() {
        return false;
    }
}
