package struclust.mcs;

import graph.Graph;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Stack;

import struclust.datastructures.MutableInt;

/**
 * Thread safe implementation of MCSa Algorithm from [1].
 * 
 * Computes a induced but not necessarily connected maximum common subgraph.
 * 
 * Attention: LICENCE is unclear
 * 
 * <br>
 * <br>
 * [1] Prosser, P. Exact Algorithms for Maximum Clique: a computational study
 * CoRR, 2012, abs/1207.4616
 * 
 * @author Till Schäfer
 * 
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class McisProsserMcsaMSA<NL, EL, G extends Graph<NL, EL>> implements McsSizeAlgorithm<NL, EL, G> {

    private int style = 3; // used sorting of vertices

    @Override
    public int calc(G graph1, G graph2) {
        MutableInt maxSize = new MutableInt(0);

        clique(new MatrixAssociationGraph<>(graph1, graph2), maxSize);

        return maxSize.get();
    }

    private void clique(MatrixAssociationGraph<NL, EL> mag, MutableInt maxSize) {
        int[][] adj = mag.getSimplifiedAdjacencies();
        int n = mag.getVertexCount();

        int[] degree = mag.getDegrees();

        @SuppressWarnings("unchecked")
        ArrayList<Integer>[] colourClass = new ArrayList[n];
        ArrayList<Integer> C = new ArrayList<Integer>(n);
        ArrayList<Integer> P = new ArrayList<Integer>(n);
        ArrayList<Integer> ColOrd = new ArrayList<Integer>(n);
        for (int i = 0; i < n; i++)
            colourClass[i] = new ArrayList<Integer>(n);
        orderVertices(ColOrd, adj, degree, n);
        expand(C, P, ColOrd, maxSize, adj, colourClass, degree);
    }

    private void expand(ArrayList<Integer> C, ArrayList<Integer> P, ArrayList<Integer> ColOrd, MutableInt maxSize,
            int[][] adj, ArrayList<Integer>[] colourClass, int[] degree) {
        int m = ColOrd.size();
        int[] colour = new int[m];
        numberSort(C, ColOrd, P, colour, adj, colourClass);
        for (int i = m - 1; i >= 0; i--) {
            int v = P.get(i);
            if (C.size() + colour[i] <= maxSize.get())
                return;
            C.add(v);
            ArrayList<Integer> newP = new ArrayList<Integer>(i);
            ArrayList<Integer> newColOrd = new ArrayList<Integer>(i);
            for (int j = 0; j <= i; j++) {
                int u = P.get(j);
                if (adj[v][u] == 1)
                    newP.add(u);
                int w = ColOrd.get(j);
                if (adj[v][w] == 1)
                    newColOrd.add(w);
            }
            if (newP.isEmpty() && C.size() > maxSize.get())
                maxSize.set(C.size());
            if (!newP.isEmpty())
                expand(C, newP, newColOrd, maxSize, adj, colourClass, degree);
            C.remove(C.size() - 1);
            P.remove(i);
            ColOrd.remove((Integer) v);
        }
    }

    private void numberSort(ArrayList<Integer> C, ArrayList<Integer> ColOrd, ArrayList<Integer> P, int[] colour,
            int[][] adj, ArrayList<Integer>[] colourClass) {
        int colours = 0;
        int m = ColOrd.size();
        for (int i = 0; i < m; i++)
            colourClass[i].clear();
        for (int i = 0; i < m; i++) {
            int v = ColOrd.get(i);
            int k = 0;
            while (conflicts(v, colourClass[k], adj))
                k++;
            colourClass[k].add(v);
            colours = Math.max(colours, k + 1);
        }
        P.clear();
        int i = 0;
        for (int k = 0; k < colours; k++)
            for (int j = 0; j < colourClass[k].size(); j++) {
                int v = (colourClass[k].get(j));
                P.add(v);
                colour[i++] = k + 1;
            }
    }

    private boolean conflicts(int v, ArrayList<Integer> colourClass, int[][] A) {
        for (int i = 0; i < colourClass.size(); i++) {
            int w = colourClass.get(i);
            if (A[v][w] == 1)
                return true;
        }
        return false;
    }

    private void orderVertices(ArrayList<Integer> ColOrd, int[][] adj, int[] degree, int n) {
        Vertex[] V = new Vertex[n];
        for (int i = 0; i < n; i++)
            V[i] = new Vertex(i, degree[i]);
        for (int i = 0; i < n; i++)
            for (int j = 0; j < n; j++)
                if (adj[i][j] == 1)
                    V[i].nebDeg = V[i].nebDeg + degree[j];
        if (style == 1)
            Arrays.sort(V);
        if (style == 2)
            minWidthOrder(V, adj, n);
        if (style == 3)
            Arrays.sort(V, new MCRComparator());
        for (Vertex v : V)
            ColOrd.add(v.index);
    }

    private void minWidthOrder(Vertex[] V, int[][] A, int n) {
        ArrayList<Vertex> L = new ArrayList<Vertex>(n);
        Stack<Vertex> S = new Stack<Vertex>();
        for (Vertex v : V)
            L.add(v);
        while (!L.isEmpty()) {
            Vertex v = L.get(0);
            for (Vertex u : L)
                if (u.degree < v.degree)
                    v = u;
            S.push(v);
            L.remove(v);
            for (Vertex u : L)
                if (A[u.index][v.index] == 1)
                    u.degree--;
        }
        int k = 0;
        while (!S.isEmpty())
            V[k++] = S.pop();
    }

    protected static class Vertex implements Comparable<Vertex> {

        int index, degree, colour, saturation, nebDeg;
        boolean[] domain;

        public Vertex(int index, int degree) {
            this.index = index;
            this.degree = degree;
            nebDeg = 0;
        }

        void init(int n) {
            domain = new boolean[n];
            Arrays.fill(domain, true);
            colour = -1;
            saturation = 0;
        }

        int colour() {
            for (colour = 0; !domain[colour]; colour++);
            domain[colour] = false;
            return colour;
        }

        // colour vertex with lowest possible colour
        void removeColour(int i) {
            if (domain[i]) {
                saturation++;
                domain[i] = false;
            }
        }

        @Override
        public int compareTo(Vertex v) {
            if (degree < v.degree || degree == v.degree && index > v.index)
                return 1;
            return -1;
        }
    }

    protected static class MCRComparator implements Comparator<Vertex> {

        @Override
        public int compare(Vertex o1, Vertex o2) {
            Vertex u = o1;
            Vertex v = o2;
            if (u.degree < v.degree || u.degree == v.degree && u.nebDeg < v.nebDeg || u.degree == v.degree
                    && u.nebDeg == v.nebDeg && u.index > v.index)
                return 1;
            return -1;
        }
        /*
         * to sort vertices by decreasing degree, tie breaking on neighbourhood
         * degree (nebDeg)
         */
    }

    @Override
    public boolean isConnected() {
        return false;
    }

    @Override
    public boolean isInduced() {
        return true;
    }

    @Override
    public String getDescription() {
        return "McisProsserMcsaMSA";
    }

}
