package struclust.mining;

import graph.Graph;

import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.Cluster;

import com.google.common.base.Preconditions;

/**
 * Calculates the minimal support value, by using the average representative
 * size in comparison to the average cluster members size
 * (\frac{avgRepSize}{factor x avgMemberSize}). It scales linear from a
 * specified minimum to a specified maximum minimal support value with this
 * fraction.
 * 
 * It is possible to specify a factor for the avgMemberSize, such that one can
 * reach the maximum minimal support for lower members sizes.
 * 
 * Special Case: if the {@link Cluster} does not contain any representative or
 * member, the minimum minimal support will be returned.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class FixedRepToMemberSizeMSS<NL, EL, G extends Graph<NL, EL>> implements MinSupStrategy<NL, EL, G> {

    private static final Logger logger = LoggerFactory.getLogger(FixedRepToMemberSizeMSS.class);

    private double min;
    private double max;
    private double factor;

    /**
     * Constructor
     * 
     * @param min
     *            the minimum support value
     * @param max
     *            the maximum support value
     * @param factor
     *            the factor to adjust the member size (e.g. 0.5 if you want to
     *            reach maximum support for a representative size that is equal
     *            to 0.5 times the member size)
     */
    public FixedRepToMemberSizeMSS(double min, double max, double factor) {
        Preconditions.checkArgument(min < max);
        Preconditions.checkArgument(factor > 0);

        this.min = min;
        this.max = max;
        this.factor = factor;
    }

    @Override
    public double calc(List<Cluster<NL, EL, G>> all, Cluster<NL, EL, G> current) {
        if (current.getRepresentatives().size() == 0 || current.size() == 0) {
            return min;
        }

        double avgMemberSize = current.getAvgMemberSize();
        double avgRepSize = current.getAvgRepSize();

        double scaling = avgRepSize / (factor * avgMemberSize);
        double range = max - min;
        double retVal = Math.min(scaling * range + min, max);

        logger.debug("MinSup: {}, Cluster {}, avgRepSize: {}, avgMemberSize {}", retVal, current.getId(), avgRepSize,
                avgMemberSize);
        return retVal;
    }

    @Override
    public String getDescription() {
        return "RelRepToMemberSizeMSS (min " + min + ", max " + max + ", factor " + factor + ")";
    }

}
