package struclust.mining;

import graph.Graph;
import index.features.FingerprintBuilder;

import java.util.Iterator;
import java.util.List;

import match.MatcherFast;
import match.pattern.SearchPattern;
import struclust.SharedMemorySCC;
import struclust.datastructures.MutableInt;
import struclust.graph.GraphContainer;

import com.google.common.base.Preconditions;

/**
 * Calculated the Support exactly with fingerprint prefiltering for
 * acceleration. Requires precalculated fingerprints for the graphs.
 * 
 * @author Till Schäfer
 * 
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class PrefilteredSC<NL, EL, G extends Graph<NL, EL>> extends AbstractSC<NL, EL, G> {

    /**
     * Constructor
     * 
     * @param fpBuilder
     *            the {@link FingerprintBuilder} to calculate fingerprints for
     *            the search patterns
     * @param parallelism
     *            the parallelism (see {@link SharedMemorySCC#parallelism}
     */
    public PrefilteredSC(FingerprintBuilder fpBuilder, MutableInt parallelism) {
        super(fpBuilder, parallelism);
    }

    @Override
    public boolean supported(List<GraphContainer<NL, EL, G>> graphs, G pattern, double minSup, int numHypothesis,
            boolean parallel) {
        Preconditions.checkArgument(minSup <= 1.0 && minSup >= 0.0);

        double supThreshold = minSup * graphs.size();
        List<GraphContainer<NL, EL, G>> candidates = calculateCandidates(graphs, pattern, null);

        // abort if we are already below the support
        if (candidates.size() < supThreshold) {
            return false;
        }

        int count = 0;
        int candidatesSize = candidates.size();
        SearchPattern<NL, EL> sp = new SearchPattern<>(pattern, false);
        Iterator<GraphContainer<NL, EL, G>> it = candidates.iterator();
        for (int i = 1; i <= candidates.size(); i++) {
            // PERF: parallelize
            if (new MatcherFast<NL, EL>(sp, it.next()).match()) {
                count++;
            }

            // we can abort counting if we are already supported
            if (count >= supThreshold) {
                return true;
            }

            // we can abort counting if we cannot reach the Threshold anymore
            int remainingChecks = candidatesSize - i;
            if (remainingChecks + count < supThreshold) {
                return false;
            }
        }

        assert !it.hasNext();
        return false;
    }

    @Override
    public double support(List<GraphContainer<NL, EL, G>> graphs, G pattern, boolean parallel) {
        assert graphs.size() > 0;

        List<GraphContainer<NL, EL, G>> candidates = calculateCandidates(graphs, pattern, null);
        int count = numSuportedGraphs(candidates, pattern, parallel);
        double retVal = (double) count / graphs.size();

        return retVal;
    }

    @Override
    public String getDescription() {
        return "PrefilteredSC";
    }
}
