package struclust.preclustering;

import graph.Graph;
import graph.GraphFactory;
import index.features.FingerprintBuilder;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.Future;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.SharedMemorySCC;
import struclust.comparison.GraphGraphDistance;
import struclust.graph.GraphContainer;
import struclust.graph.Graphs;
import struclust.mining.RandomGM;

import com.google.common.util.concurrent.ThreadFactoryBuilder;

/**
 * {@link PreClustering} that uses dissimilar frequent substructures for cluster
 * identification. This approach is similar to ORIGAMI [1] pattern mining, but
 * simply uses the maximum size clique. It roughly follows the steps below.
 * 
 * <ul>
 * <li>mine a user specified number of random frequent subgraphs</li>
 * <li>construct a similarity graph: An edge between two substructures exist,
 * iff their distance is smaller than a user specified value alpha
 * <li>find (inclusion) maximal cliques with a leader-style approach
 * <li>choose the clique that has maximum size
 * </ul>
 * <br>
 *
 * <b>Attention:</b> In {@link #calc(ArrayList, int)} numClusters is ignored.<br>
 * 
 * <b>[1]</b> Mohammad Al Hasan and Vineet Chaoji and Saeed Salem and Jérémy
 * Besson and Mohammed Javeed Zaki, M. A. H. ORIGAMI: Mining Representative
 * Orthogonal Graph Patterns Proceedings of the 7th IEEE International
 * Conference on Data Mining (ICDM 2007), October 28-31, 2007, Omaha, Nebraska,
 * USA, IEEE Computer Society, 2007, 153-162
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class DissimilarRepPC<NL, EL, G extends Graph<NL, EL>> implements PreClustering<NL, EL, G> {
    private static final Logger logger = LoggerFactory.getLogger(DissimilarRepPC.class);

    private int numFp;
    private double minSup;
    private int minClusterSize;
    private int numCliques;
    private double alpha;
    private GraphGraphDistance<NL, EL, G> cliqueDist;
    private FingerprintBuilder fpBuilder;
    private GraphGraphDistance<NL, EL, G> assignmentDist;
    private RandomGM<NL, EL, G> gm;

    private ExecutorService executor;
    private ForkJoinPool forkJoinPool;

    /**
     * Constructor
     * 
     * @param conf
     *            the clustering configuration
     * @param numFp
     *            the number of random frequent subgraphs to mine
     * @param minSup
     *            the minimal support value for random frequent subgraph
     *            generation. If this value is in the range of [0,1] it is
     *            interpreted as a percentage value. If it larger than 1 it is
     *            interpreted as an absolute value.
     * @param minClusterSize
     *            the minimal desired cluster size. This value affects the
     *            internally used minSup Value. The internally used minSup value
     *            is equal to max(minSup; minClustersSize/graphs.size). E.g. if
     *            you want a minimal cluster size of 10 and you have 500 graphs,
     *            the minSup must be larger or equal to 0.05 in order to mine
     *            substructures that are embedded in at least 10 graphs. <b>
     *            Attention:</b> This setting is only considered if the minSup
     *            value specified as percentage!
     * @param numCliques
     *            the number of random inclusion maximal cliques that are
     *            generated
     * @param alpha
     *            the distance threshold for similarity graph construction
     * @param cliqueDist
     *            the distance measure to create the similarity graph for clique
     *            construction.
     * @param assignmentDist
     *            the distance measure to assign the graphs to the minimal
     *            distant representative. If this value is null, cliqueDist is
     *            used.
     * @param gFactory
     *            Factory to create graphs
     * @param fpBuilder
     *            the {@link FingerprintBuilder} for fingerprint creation
     */
    public DissimilarRepPC(SharedMemorySCC<NL, EL, G> conf, int numFp, double minSup, int minClusterSize,
            int numCliques, double alpha, GraphGraphDistance<NL, EL, G> cliqueDist,
            GraphGraphDistance<NL, EL, G> assignmentDist, GraphFactory<NL, EL, G> gFactory, FingerprintBuilder fpBuilder) {
        this.numFp = numFp;
        this.minSup = minSup;
        this.minClusterSize = minClusterSize;
        this.numCliques = numCliques;
        this.alpha = alpha;
        this.cliqueDist = cliqueDist;
        this.assignmentDist = assignmentDist == null ? cliqueDist : assignmentDist;
        this.fpBuilder = fpBuilder;
        this.forkJoinPool = new ForkJoinPool(conf.getParallelism());
        this.executor = Executors.newFixedThreadPool(conf.getParallelism(),
                new ThreadFactoryBuilder().setNameFormat("DissimalarRepPC-Clique-%d").setDaemon(true).build());

        RandomGM.RandomGMConf<NL, EL, G> gmConf = new RandomGM.RandomGMConf<NL, EL, G>(gFactory, conf);
        gmConf.useSupSizeProduct = false;
        gmConf.parallel = true;
        gmConf.gSize = conf.gSize;
        gm = new RandomGM<>(gmConf);

    }

    @Override
    public String getDescription() {
        return "DissimilarSubStructurePC (numFP: " + numFp + ", minSup: " + minSup + ", minClusterSize: "
                + minClusterSize + ", numCliques: " + numCliques + ", alpha: " + alpha + ", dist: "
                + cliqueDist.getDescription() + ")";
    }

    @Override
    public List<? extends List<GraphContainer<NL, EL, G>>> calc(ArrayList<GraphContainer<NL, EL, G>> graphs) {
        return doCalc(graphs);
    }

    @Override
    public List<? extends List<GraphContainer<NL, EL, G>>> calc(ArrayList<GraphContainer<NL, EL, G>> graphs,
            int numClusters) {
        return doCalc(graphs);
    }

    private ArrayList<ArrayList<GraphContainer<NL, EL, G>>> doCalc(ArrayList<GraphContainer<NL, EL, G>> graphs) {
        if (graphs.isEmpty()) {
            logger.warn("calling calc with an empty set of graphs");
            return new ArrayList<>();
        }
        ArrayList<GraphContainer<NL, EL, G>> minedGraphs = mineFp(graphs);
        logger.debug("{} pattern mined", minedGraphs.size());
        // PERF: parallelize (after profiling if this is an issue)
        Graphs.distinctGraphs(minedGraphs);
        logger.debug("{} distinct patterns", minedGraphs.size());
        generateFingerprints(minedGraphs);
        ArrayList<ArrayList<GraphContainer<NL, EL, G>>> cliques = generateCliques(minedGraphs);
        logger.debug("generated {} cliques", cliques.size());
        assert cliques.size() > 0;
        ArrayList<GraphContainer<NL, EL, G>> bestClique = selectBestClique(graphs, cliques);
        logger.debug("best clique has size {}", bestClique.size());
        ArrayList<ArrayList<GraphContainer<NL, EL, G>>> clusters = createClusters(graphs, bestClique);
        return clusters;
    }

    private void generateFingerprints(List<GraphContainer<NL, EL, G>> graphs) {
        /*
         * PERF: This is only needed for Distances that rely on bit
         * fingerprints.
         */
        try {
            /*
             * This is a workaround for Java stream api to not support a
             * ThreadPool specification
             * 
             * PARALLEL stream: fingerprint calculation
             */
            forkJoinPool.submit(
                    () -> graphs.stream().parallel().filter(x -> !x.hasFp())
                            .forEach(x -> x.setFp(fpBuilder.getFingerprint(x)))).get();
        } catch (InterruptedException e) {
            logger.warn("Interrupted: ", e);
            throw new RuntimeException(e);
        } catch (ExecutionException e) {
            logger.error("Error in fingerprint calculation: ", e);
            throw new RuntimeException(e);
        }
    }

    private ArrayList<ArrayList<GraphContainer<NL, EL, G>>> createClusters(ArrayList<GraphContainer<NL, EL, G>> graphs,
            ArrayList<GraphContainer<NL, EL, G>> bestClique) {
        int numClusters = bestClique.size();
        ArrayList<ArrayList<GraphContainer<NL, EL, G>>> clusters = new ArrayList<>(numClusters);

        for (int i = 0; i < numClusters; i++) {
            clusters.add(new ArrayList<>());
        }

        try {
            /*
             * This is a workaround for Java stream api to not support a
             * ThreadPool specification
             * 
             * PARALLEL stream: cluster assignment
             */
            forkJoinPool
                    .submit(() -> graphs.stream().parallel()
                            .forEach(x -> assignToMinRep(numClusters, clusters, bestClique, x))).get();
        } catch (InterruptedException e) {
            logger.warn("Interrupted: ", e);
            throw new RuntimeException(e);
        } catch (ExecutionException e) {
            logger.error("Error in parallel assignment: ", e);
            throw new RuntimeException(e);
        }

        return clusters;
    }

    private void assignToMinRep(int numClusters, ArrayList<ArrayList<GraphContainer<NL, EL, G>>> clusters,
            ArrayList<GraphContainer<NL, EL, G>> bClique, GraphContainer<NL, EL, G> gc) {
        double minDist = Double.POSITIVE_INFINITY;
        int minRep = -1;
        for (int i = 0; i < numClusters; i++) {
            double distVal = assignmentDist.calc(bClique.get(i), gc);
            if (distVal < minDist) {
                minDist = distVal;
                minRep = i;
            }
        }

        ArrayList<GraphContainer<NL, EL, G>> minRepCluster = clusters.get(minRep);
        synchronized (minRepCluster) {
            minRepCluster.add(gc);
        }
    }

    private ArrayList<GraphContainer<NL, EL, G>> selectBestClique(ArrayList<GraphContainer<NL, EL, G>> graphs,
            ArrayList<ArrayList<GraphContainer<NL, EL, G>>> cliques) {

        return cliques.stream().max(Comparator.comparing(c -> c.size())).get();
    }

    private ArrayList<ArrayList<GraphContainer<NL, EL, G>>> generateCliques(List<GraphContainer<NL, EL, G>> minedGraphs) {

        List<Future<ArrayList<GraphContainer<NL, EL, G>>>> futures = new ArrayList<>(numCliques);
        ArrayList<ArrayList<GraphContainer<NL, EL, G>>> cliques = new ArrayList<>(numCliques);

        // PARALLEL: clique generation
        for (int i = 0; i < numCliques; i++) {
            futures.add(executor.submit(new SingleCliqueGenerator(minedGraphs)));
        }

        for (Future<ArrayList<GraphContainer<NL, EL, G>>> future : futures) {
            try {
                cliques.add(future.get());
            } catch (InterruptedException | ExecutionException e) {
                logger.error("Error in parallel clique generation", e);
                throw new RuntimeException(e);
            }
        }

        return cliques;
    }

    private ArrayList<GraphContainer<NL, EL, G>> generateSingleClique(List<GraphContainer<NL, EL, G>> minedGraphs) {
        Collections.shuffle(minedGraphs);
        ArrayList<GraphContainer<NL, EL, G>> currentClique = new ArrayList<>();
        for (GraphContainer<NL, EL, G> gc : minedGraphs) {
            if (currentClique.isEmpty()) {
                currentClique.add(gc);
            } else {
                try {
                    /*
                     * This is a workaround for Java stream api to not support a
                     * ThreadPool specification
                     * 
                     * PARALLEL stream: calculate minimal distance
                     */
                    forkJoinPool.submit(() -> add2CliqueIfValid(currentClique, gc)).get();
                } catch (InterruptedException e) {
                    logger.warn("Interrupted: ", e);
                    throw new RuntimeException(e);
                } catch (ExecutionException e) {
                    logger.error("Error in min dist calculation: ", e);
                    throw new RuntimeException(e);
                }
            }
        }
        return currentClique;
    }

    private void add2CliqueIfValid(List<GraphContainer<NL, EL, G>> currentClique, GraphContainer<NL, EL, G> gc) {
        // PARALLEL stream: calculate minimal distance
        double minDist = currentClique.stream().parallel().mapToDouble(x -> cliqueDist.calc(gc, x)).min().getAsDouble();
        if (minDist >= alpha) {
            currentClique.add(gc);
        }
    }

    private ArrayList<GraphContainer<NL, EL, G>> mineFp(ArrayList<GraphContainer<NL, EL, G>> graphs) {
        double effectiveMinSup = minSup > 1.0 ? minSup / graphs.size() : Math.max(minSup, (double) minClusterSize
                / graphs.size());

        ArrayList<GraphContainer<NL, EL, G>> minedGraphs;
        do {
            logger.debug("specified minSup: {}, effective minSup: {}", minSup, effectiveMinSup);
            minedGraphs = gm.mineGraphs(graphs, numFp, effectiveMinSup);
            effectiveMinSup /= 2;
        } while (minedGraphs.isEmpty());
        return minedGraphs;
    }

    private class SingleCliqueGenerator implements Callable<ArrayList<GraphContainer<NL, EL, G>>> {
        private List<GraphContainer<NL, EL, G>> minedGraphs;

        public SingleCliqueGenerator(List<GraphContainer<NL, EL, G>> minedGraphs) {
            this.minedGraphs = minedGraphs;
        }

        @Override
        public ArrayList<GraphContainer<NL, EL, G>> call() throws Exception {
            return generateSingleClique(minedGraphs);
        }

    }
}
