package struclust.preclustering;

import graph.Graph;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.graph.GraphContainer;

import com.google.common.base.Preconditions;

/**
 * Simply partitions the data set in (nearly) equally large parts. It is fake in
 * the way, that the data itself has no influence on the partitioning.
 * 
 * @author Till Schäfer
 * 
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class FakePC<NL, EL, G extends Graph<NL, EL>> implements PreClustering<NL, EL, G> {
    private static final Logger logger = LoggerFactory.getLogger(FakePC.class);

    private int quantity;

    /**
     * Constructor
     * 
     * @param quantity
     *            the number of clusters to create
     */
    public FakePC(int quantity) {
        this.quantity = quantity;
    }

    @Override
    public LinkedList<? extends List<GraphContainer<NL, EL, G>>> calc(ArrayList<GraphContainer<NL, EL, G>> graphs) {
        return calc(graphs, quantity);
    }

    @Override
    public String getDescription() {
        return "Fake PreClustering";
    }

    @Override
    public LinkedList<? extends List<GraphContainer<NL, EL, G>>> calc(ArrayList<GraphContainer<NL, EL, G>> graphs,
            int numClusters) {
        Preconditions.checkArgument(graphs.size() >= numClusters,
                "The quantitiy cannot be less than the number of elements to cluster");

        logger.trace("Starting Fake Pre Clustering");
        LinkedList<List<GraphContainer<NL, EL, G>>> retVal = new LinkedList<>();

        double stepWidth = (double) graphs.size() / numClusters;
        for (int i = 0; i < numClusters; i++) {
            int fromIndex = (int) (i * stepWidth);
            // correct possible rounding errors for the last iteration
            int toIndex = i == (numClusters - 1) ? graphs.size() - 1 : (int) ((i + 1) * stepWidth - 1);

            logger.debug("Cluster {} range: {}, {}", (i + 1), fromIndex, toIndex);

            retVal.add(graphs.subList(fromIndex, toIndex));
        }
        return retVal;
    }
}
