package struclust.repselection;

import graph.Graph;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.Cluster;
import struclust.SharedMemorySCC;
import struclust.comparison.CoverageCGD;
import struclust.graph.GraphContainer;
import struclust.graph.Graphs;

/**
 * Heuristically selects the representatives which maximize the average coverage
 * of the cluster. Since testing all (repCount choose totalReps) possible
 * subsets of representatives is too expensive, we only test a single new
 * representative at once. If this representative can increase the coverage we
 * instantly replace the worst old representative with the new one. Afterwards
 * we continue with the next new representative.
 * 
 * Special Case: if the number of representatives is smaller than conf.repCount,
 * the new representative is added without replacing an old one (if the coverage
 * is increased).
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class CoverageRSS<NL, EL, G extends Graph<NL, EL>> implements RepSelectionStrategy<NL, EL, G> {
    private static final Logger logger = LoggerFactory.getLogger(CoverageRSS.class);

    private CoverageCGD<NL, EL, G> dist;
    private SharedMemorySCC<NL, EL, G> conf;

    private boolean removeSubgraphs;

    /**
     * Constructor
     * 
     * @param conf
     *            the configuration
     * @param removeSubgraphs
     *            if true, all subgraphs are filtered out of the candidate set
     *            for representatives (i.e. only the supergraphs are considered
     *            as representatives)
     */
    public CoverageRSS(SharedMemorySCC<NL, EL, G> conf, boolean removeSubgraphs) {
        this.conf = conf;
        this.removeSubgraphs = removeSubgraphs;
        dist = new CoverageCGD<>(conf);
    }

    @Override
    public ArrayList<GraphContainer<NL, EL, G>> select(Cluster<NL, EL, G> cluster,
            ArrayList<GraphContainer<NL, EL, G>> newReps, List<Cluster<NL, EL, G>> clustering) {
        List<GraphContainer<NL, EL, G>> reps = new ArrayList<>(cluster.getRepresentatives());
        reps.addAll(newReps);

        if (removeSubgraphs) {
            Graphs.filterSubgraphs(reps);
        } else {
            Graphs.distinctGraphs(reps);
        }

        ArrayList<GraphContainer<NL, EL, G>> retVal = new ArrayList<>();

        for (GraphContainer<NL, EL, G> rep : reps) {
            Cluster<NL, EL, G> curCluster = new Cluster<>(Collections.emptyList(), new ArrayList<>(retVal), 0,
                    conf.gSize);
            double oldCoverage = retVal.size() == 0 ? 0 : 1 - cluster.stream()
                    .mapToDouble(x -> dist.calc(curCluster, x)).average().getAsDouble();

            if (retVal.size() < conf.repCount) {
                curCluster.getRepresentatives().add(rep);
                double coverage = 1 - cluster.stream().mapToDouble(x -> dist.calc(curCluster, x)).average()
                        .getAsDouble();

                assert coverage >= oldCoverage;
                if (coverage > oldCoverage) {
                    retVal.add(rep);
                }
            } else {
                double maxCoverage = Double.NEGATIVE_INFINITY;
                GraphContainer<NL, EL, G> replacedRep = null;

                for (GraphContainer<NL, EL, G> oldRep : retVal) {
                    curCluster.getRepresentatives().remove(oldRep);
                    curCluster.getRepresentatives().add(rep);
                    double coverage = 1 - cluster.stream().mapToDouble(x -> dist.calc(curCluster, x)).average()
                            .getAsDouble();

                    if (coverage > maxCoverage) {
                        maxCoverage = coverage;
                        replacedRep = oldRep;
                    }
                }

                if (oldCoverage < maxCoverage) {
                    retVal.remove(replacedRep);
                    retVal.add(rep);
                    logger.debug("replacing rep {} by rep {}", replacedRep, rep);
                }
            }

        }
        return retVal;
    }

    @Override
    public String getDescription() {
        return "BestCoverageRSS (removeSubgraphs: " + removeSubgraphs + ")";
    }

}
