package struclust.repselection;

import graph.Graph;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.stream.Collectors;

import org.apache.commons.lang3.tuple.Pair;

import struclust.Cluster;
import struclust.SharedMemorySCC;
import struclust.graph.GraphContainer;
import struclust.graph.Graphs;

/**
 * Selects the distinct representatives which have the highest score with
 * respect to the following formula:
 * 
 * <pre>
 * repEdgeCount * (innerSupport - otherSupport)
 * </pre>
 * 
 * where innerSupport is the support in cluster and otherSupport is the average
 * or maximum support in clustering.
 * 
 * @author Till Schäfer
 * 
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class DiscriminativeSupSizeRSS<NL, EL, G extends Graph<NL, EL>> implements RepSelectionStrategy<NL, EL, G> {

    private SharedMemorySCC<NL, EL, G> conf;
    private boolean useMax;
    private boolean removeSubgraphs;

    /**
     * Constructor
     * 
     * @param conf
     *            the configuration
     * @param useMax
     *            if true, otherSupport is calculated by the maximum support
     *            over all other clusters in clustering. if false, otherSupport
     *            is calculated by the average support over the complete data
     *            set.
     * @param removeSubgraphs
     *            if true, all subgraphs are filtered out of the candidate set
     *            for representatives (i.e. only the supergraphs are considered
     *            as representatives)
     */
    public DiscriminativeSupSizeRSS(SharedMemorySCC<NL, EL, G> conf, boolean useMax, boolean removeSubgraphs) {
        this.conf = conf;
        this.useMax = useMax;
        this.removeSubgraphs = removeSubgraphs;
    }

    @Override
    public ArrayList<GraphContainer<NL, EL, G>> select(Cluster<NL, EL, G> cluster,
            ArrayList<GraphContainer<NL, EL, G>> newReps, List<Cluster<NL, EL, G>> clustering) {

        List<GraphContainer<NL, EL, G>> reps = new ArrayList<>(newReps);
        reps.addAll(cluster.getRepresentatives());

        if (removeSubgraphs) {
            Graphs.filterSubgraphs(reps);
        } else {
            Graphs.distinctGraphs(reps);
        }

        // calculate the exact support for each representative
        final HashMap<GraphContainer<NL, EL, G>, Pair<Double, Double>> support = new HashMap<>(reps.size());
        for (GraphContainer<NL, EL, G> gc : reps) {
            Double innerSup = conf.supportCounting.support(cluster, gc.getGraph());
            Double otherSup;
            if (useMax) {
                otherSup = clustering.stream().filter(x -> x != cluster)
                        .mapToDouble(y -> conf.supportCounting.support(y, gc.getGraph())).max().orElse(0);
            } else {
                List<GraphContainer<NL, EL, G>> allGraphs = clustering.stream().flatMap(x -> x.stream())
                        .collect(Collectors.toList());
                otherSup = conf.supportCounting.support(allGraphs, gc.getGraph());
            }

            support.put(gc, Pair.of(innerSup, otherSup));
        }

        // sort the representatives by support x size
        Collections.sort(reps, new Comparator<GraphContainer<NL, EL, G>>() {
            @Override
            public int compare(GraphContainer<NL, EL, G> o1, GraphContainer<NL, EL, G> o2) {
                Pair<Double, Double> sup1 = support.get(o1);
                Pair<Double, Double> sup2 = support.get(o2);
                double score1 = o1.getEdgeCount() * (sup1.getLeft() - sup1.getRight());
                double score2 = o2.getEdgeCount() * (sup2.getLeft() - sup2.getRight());
                return -1 * Double.compare(score1, score2);
            }
        });

        return new ArrayList<>(reps.subList(0, Math.min(conf.repCount, reps.size())));
    }

    @Override
    public String getDescription() {
        return "DiscriminativeSupSizeRSS (useMax: " + useMax + ", removeSubgraphs: " + removeSubgraphs + ")";
    }

}
