package struclust.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Random;
import java.util.concurrent.ThreadLocalRandom;

import cern.jet.random.AbstractContinousDistribution;
import cern.jet.random.AbstractDistribution;

/**
 * Convenience methods for statistical operations.
 * 
 * @author Till Schäfer
 * @author Nils Kriege
 *
 */
public class Statistical {
    /**
     * Draws weights from a given {@link AbstractContinousDistribution} and
     * normalizes them to derive a discrete probability mass function.
     * Afterwards it is transformed to a cumulative distribution derived from
     * the density function.
     * 
     * @param size
     *            the number of values
     * @param distrib
     *            the distribution to draw the weights of
     * @return a cumulative distribution derived from the weights
     */
    public static double[] discreteWeightCumulativeDistribution(int size, AbstractDistribution distrib) {
        double[] weights = new double[size];
        double totalWeights = 0;
        for (int i = 0; i < size; i++) {
            double weight = distrib.nextDouble();
            weights[i] = weight;
            totalWeights += weight;
        }
        // normalize weights and cumulate
        double bound = 0;
        for (int i = 0; i < size; i++) {
            weights[i] /= totalWeights;
            bound = (weights[i] += bound);
        }

        // correct rounding errors
        assert Math.abs(weights[weights.length - 1] - 1) < 0.0001;
        weights[weights.length - 1] = 1;

        return weights;
    }

    /**
     * Draws a value from a discrete distribution of size length(values)
     * starting with element 0.
     * 
     * @param values
     *            the values of the discrete distribution
     * @param rand
     *            random number generator
     * @return the index of the element drawn.
     */
    public static int drawFromDiscreteDistribution(double[] values, Random rand) {
        // RAND: draw from discrete distribution
        double var = rand.nextDouble();
        int index = Arrays.binarySearch(values, var);
        if (index < 0) {
            index += 1;
            index *= -1;
        }

        assert index < values.length;
        return index;
    }

    /**
     * Random sampler that uses Floyd’s algorithm.
     * 
     * <br>
     * Has complexity O(|count|)
     * 
     * @param population
     *            the population to draw the sample from
     * @param sampleSize
     *            the size of the sample. if sampleSize is larger than
     *            population.size(), it will be automatically adjusted to
     *            population.size()
     * @return a random subset of population with size min{sampleSize,
     *         population.size()}
     */
    public static <T> ArrayList<T> getSampleWithFloyd(ArrayList<T> population, int sampleSize) {
        // RAND: draw random sample
        Random rand = ThreadLocalRandom.current();
        sampleSize = Math.min(sampleSize, population.size());

        ArrayList<T> retVal = new ArrayList<>(sampleSize);
        int n = population.size();
        for (int i = n - sampleSize; i < n; i++) {
            int pos = rand.nextInt(i + 1);
            T item = population.get(pos);
            if (retVal.contains(item))
                retVal.add(population.get(i));
            else
                retVal.add(item);
        }
        return retVal;
    }
}
