package struclust.util;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Spliterator;
import java.util.function.Consumer;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

import org.apache.commons.math3.util.Pair;

/**
 * Convenience methods for {@link Stream}s.
 * 
 * @author Till Schäfer
 *
 */
public class Streams {
    /**
     * Returns a Stream of all Pairs AxB (in accordance to
     * {@link CrossProductSpliterator})
     * 
     * @param a
     *            Collection A
     * @param b
     *            Collection B
     * @return a Stream of all {@link Pair}s AxB
     */
    public static <A, B> Stream<Pair<A, B>> crossProductStream(Collection<A> a, Collection<B> b) {
        return StreamSupport.stream(new CrossProductSpliterator<A, B>(a, b), false);
    }

    /**
     * Returns a Stream of all possible Ordered Pairs CxC from Collection C.
     * Ordered means that for a,b \in C either (a,b) or (b,a) is returned, but
     * never both.
     * 
     * <br>
     * <strong>Attention</strong>: Elements in C must be unique!
     * 
     * @param c
     *            Collection A
     * @return a Stream of all ordered {@link Pair}s CxC
     */
    public static <T> Stream<Pair<T, T>> orderedPairsStream(Collection<T> c) {
        return StreamSupport.stream(new OrderedPairsSpliterator<T>(c), false);
    }

    /**
     * Spliterator that returns all possible pairs of elements AxB.
     * 
     * @author Till Schäfer
     *
     * @param <A>
     *            type of elements from {@link Collection} A
     * @param <B>
     *            type of elements from {@link Collection} B
     */
    public static class CrossProductSpliterator<A, B> implements Spliterator<Pair<A, B>> {

        private Spliterator<A> splitA;
        private Collection<B> b;

        private A lastValA;
        private Iterator<B> itB;
        int countB = 0;
        private boolean isAValid;

        /**
         * Constructor
         * 
         * @param a
         *            Collection A
         * @param b
         *            Collection B
         */
        public CrossProductSpliterator(Collection<A> a, Collection<B> b) {
            init(a.spliterator(), b);
        }

        private CrossProductSpliterator(Spliterator<A> splitA, Collection<B> b) {
            init(splitA, b);
        }

        private void init(Spliterator<A> splitA, Collection<B> b) {
            this.splitA = splitA;
            isAValid = splitA.tryAdvance(val -> lastValA = val);
            this.b = b;
            itB = b.iterator();
        }

        @Override
        public boolean tryAdvance(Consumer<? super Pair<A, B>> action) {
            if (!isAValid) {
                return false;
            }

            if (itB.hasNext()) {
                action.accept(Pair.create(lastValA, itB.next()));
                countB++;
                return true;
            } else {
                isAValid = splitA.tryAdvance(val -> lastValA = val);
                itB = b.iterator();
                countB = 0;
                return tryAdvance(action);
            }
        }

        @Override
        public Spliterator<Pair<A, B>> trySplit() {
            Spliterator<A> splittedA = splitA.trySplit();
            if (splittedA != null) {
                return new CrossProductSpliterator<>(splittedA, b);
            } else {
                return null;
            }
        }

        @Override
        public long estimateSize() {
            return splitA.estimateSize() * b.size() + (b.size() - countB);
        }

        @Override
        public int characteristics() {
            return SIZED | SUBSIZED | NONNULL;
        }
    }

    /**
     * Returns all possible Ordered Pairs CxC from Collection C. Ordered means
     * that for a,b \in C either (a,b) or (b,a) is returned, but never both.
     * 
     * <br>
     * <strong>Attention</strong>: Elements in C must be unique!
     * 
     * @author Till Schäfer
     *
     * @param <T>
     *            type of elements from {@link Collection} C
     */
    public static class OrderedPairsSpliterator<T> implements Spliterator<Pair<T, T>> {

        private Spliterator<T> splitBase;
        private boolean isBaseValid;
        private T baseVal;
        private Collection<T> c;
        private Iterator<T> itRunner;
        int countRunner = 0;

        /**
         * Constructor
         * 
         * @param c
         *            {@link Collection} of elements
         */
        public OrderedPairsSpliterator(Collection<T> c) {
            assert (new HashSet<>(c)).size() == c.size();
            init(c, c.spliterator());
        }

        /**
         * Constructor
         * 
         * @param a
         *            Collection A
         * @param b
         *            Collection B
         */
        private OrderedPairsSpliterator(Collection<T> c, Spliterator<T> splitBase) {
            init(c, splitBase);
        }

        private void init(Collection<T> c, Spliterator<T> splitBase) {
            this.c = c;
            this.splitBase = splitBase;
            isBaseValid = splitBase.tryAdvance(val -> baseVal = val);
            itRunner = c.iterator();
        }

        @Override
        public boolean tryAdvance(Consumer<? super Pair<T, T>> action) {
            if (!isBaseValid) {
                return false;
            }

            if (itRunner.hasNext()) {
                T nextRunner = itRunner.next();
                if (nextRunner != baseVal) {
                    action.accept(Pair.create(baseVal, nextRunner));
                    countRunner++;
                    return true;
                } else {
                    return advanceBase(action);
                }
            } else {
                return advanceBase(action);
            }
        }

        protected boolean advanceBase(Consumer<? super Pair<T, T>> action) {
            isBaseValid = splitBase.tryAdvance(val -> baseVal = val);
            itRunner = c.iterator();
            countRunner = 0;
            return tryAdvance(action);
        }

        @Override
        public Spliterator<Pair<T, T>> trySplit() {
            Spliterator<T> splittedBase = splitBase.trySplit();
            if (splittedBase != null) {
                return new OrderedPairsSpliterator<>(c, splittedBase);
            } else {
                return null;
            }
        }

        @Override
        public long estimateSize() {
            return splitBase.estimateSize() * c.size() / 2;
        }

        @Override
        public int characteristics() {
            return NONNULL;
        }

    }
}
