package hsahn.comparison.kernel.graph;

import graph.DefaultGraph;
import graph.Node;
import hsahn.comparison.kernel.basic.DiracKernel;
import hsahn.datastructure.SparseFeatureVector;
import hsahn.datastructure.Triple;
import hsahn.graph.GraphTools;
import hsahn.graph.LGraph;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.junit.Assert;

/**
 * JUnit Test for {@link WeisfeilerLehmanShortestPathKernel}
 * 
 * @author Till Schäfer
 */
public class WeisfeilerLehmanShortestPathKernelTest {
    /**
     * Compares the output of a singleton call of
     * {@link WeisfeilerLehmanShortestPathKernel#computeExplicitMapping(java.util.Collection)}
     * with and without stableLabelConverter.
     */
    @org.junit.Test
    public void stableLabelConverterSingleTest() {
        DefaultGraph<Integer, Integer> ga = new DefaultGraph<>();

        Node<Integer, Integer> a1 = ga.addNode(2);
        Node<Integer, Integer> a2 = ga.addNode(2);
        Node<Integer, Integer> a3 = ga.addNode(1);
        Node<Integer, Integer> a4 = ga.addNode(1);
        Node<Integer, Integer> a5 = ga.addNode(2);
        Node<Integer, Integer> a6 = ga.addNode(1);
        Node<Integer, Integer> a7 = ga.addNode(1);
        Node<Integer, Integer> a8 = ga.addNode(2);
        Node<Integer, Integer> a9 = ga.addNode(5);

        ga.addEdge(a1, a3, 0);
        ga.addEdge(a2, a3, 0);
        ga.addEdge(a2, a4, 0);
        ga.addEdge(a3, a4, 0);
        ga.addEdge(a3, a6, 0);
        ga.addEdge(a3, a7, 0);
        ga.addEdge(a4, a5, 0);
        ga.addEdge(a4, a6, 0);
        ga.addEdge(a4, a7, 0);
        ga.addEdge(a6, a7, 0);
        ga.addEdge(a6, a8, 0);
        ga.addEdge(a7, a8, 0);
        ga.addEdge(a8, a9, 0);

        LGraph<Integer, Integer> lga = GraphTools.convert(ga);

        WeisfeilerLehmanShortestPathKernel<Integer, Integer> unstableKernel = new WeisfeilerLehmanShortestPathKernel<>(
                5, new DiracKernel(), false);
        WeisfeilerLehmanShortestPathKernel<Integer, Integer> stableKernel = new WeisfeilerLehmanShortestPathKernel<>(5,
                new DiracKernel(), true);

        ArrayList<SparseFeatureVector<Triple<Integer, Integer, Integer>>> mappingUnstable = unstableKernel
                .computeExplicitMapping(Collections.singleton(lga));
        ArrayList<SparseFeatureVector<Triple<Integer, Integer, Integer>>> mappingStableA = stableKernel
                .computeExplicitMapping(Collections.singleton(lga));

        Assert.assertEquals(mappingUnstable.get(0), mappingStableA.get(0));

    }

    /**
     * Compares the output of multiple singleton calls of
     * {@link WeisfeilerLehmanShortestPathKernel#computeExplicitMapping(java.util.Collection)}
     * with a single call over multiple graphs.
     */
    @org.junit.Test
    public void stableLabelConverterMultipleTest() {
        DefaultGraph<Integer, Integer> ga = new DefaultGraph<>();
        DefaultGraph<Integer, Integer> gb = new DefaultGraph<>();

        Node<Integer, Integer> a1 = ga.addNode(2);
        Node<Integer, Integer> a2 = ga.addNode(2);
        Node<Integer, Integer> a3 = ga.addNode(1);
        Node<Integer, Integer> a4 = ga.addNode(1);
        Node<Integer, Integer> a5 = ga.addNode(2);
        Node<Integer, Integer> a6 = ga.addNode(1);
        Node<Integer, Integer> a7 = ga.addNode(1);
        Node<Integer, Integer> a8 = ga.addNode(2);
        Node<Integer, Integer> a9 = ga.addNode(5);

        ga.addEdge(a1, a3, 0);
        ga.addEdge(a2, a3, 0);
        ga.addEdge(a2, a4, 0);
        ga.addEdge(a3, a4, 0);
        ga.addEdge(a3, a6, 0);
        ga.addEdge(a3, a7, 0);
        ga.addEdge(a4, a5, 0);
        ga.addEdge(a4, a6, 0);
        ga.addEdge(a4, a7, 0);
        ga.addEdge(a6, a7, 0);
        ga.addEdge(a6, a8, 0);
        ga.addEdge(a7, a8, 0);
        ga.addEdge(a8, a9, 0);

        Node<Integer, Integer> b1 = gb.addNode(2);
        Node<Integer, Integer> b2 = gb.addNode(1);
        Node<Integer, Integer> b3 = gb.addNode(1);
        Node<Integer, Integer> b4 = gb.addNode(1);
        Node<Integer, Integer> b5 = gb.addNode(1);
        Node<Integer, Integer> b6 = gb.addNode(1);
        Node<Integer, Integer> b7 = gb.addNode(1);
        Node<Integer, Integer> b8 = gb.addNode(1);
        Node<Integer, Integer> b9 = gb.addNode(2);
        Node<Integer, Integer> b10 = gb.addNode(5);
        Node<Integer, Integer> b11 = gb.addNode(5);
        Node<Integer, Integer> b12 = gb.addNode(2);

        gb.addEdge(b1, b2, 0);
        gb.addEdge(b1, b3, 0);
        gb.addEdge(b2, b3, 0);
        gb.addEdge(b2, b4, 0);
        gb.addEdge(b2, b5, 0);
        gb.addEdge(b3, b4, 0);
        gb.addEdge(b3, b5, 0);
        gb.addEdge(b3, b8, 0);
        gb.addEdge(b4, b5, 0);
        gb.addEdge(b4, b12, 0);
        gb.addEdge(b5, b8, 0);
        gb.addEdge(b5, b12, 0);
        gb.addEdge(b6, b7, 0);
        gb.addEdge(b6, b8, 0);
        gb.addEdge(b7, b8, 0);
        gb.addEdge(b7, b9, 0);
        gb.addEdge(b8, b9, 0);
        gb.addEdge(b8, b10, 0);
        gb.addEdge(b9, b10, 0);
        gb.addEdge(b10, b11, 0);

        List<LGraph<Integer, Integer>> graphs = new ArrayList<>(2);
        LGraph<Integer, Integer> lga = GraphTools.convert(ga);
        graphs.add(lga);
        LGraph<Integer, Integer> lgb = GraphTools.convert(gb);
        graphs.add(lgb);

        WeisfeilerLehmanShortestPathKernel<Integer, Integer> unstableKernel = new WeisfeilerLehmanShortestPathKernel<>(
                5, new DiracKernel(), false);
        WeisfeilerLehmanShortestPathKernel<Integer, Integer> stableKernel = new WeisfeilerLehmanShortestPathKernel<>(5,
                new DiracKernel(), true);

        ArrayList<SparseFeatureVector<Triple<Integer, Integer, Integer>>> mappingUnstable = unstableKernel
                .computeExplicitMapping(graphs);
        ArrayList<SparseFeatureVector<Triple<Integer, Integer, Integer>>> mappingStableA = stableKernel
                .computeExplicitMapping(Collections.singleton(lga));
        ArrayList<SparseFeatureVector<Triple<Integer, Integer, Integer>>> mappingStableB = stableKernel
                .computeExplicitMapping(Collections.singleton(lgb));

        SparseFeatureVector<Triple<Integer, Integer, Integer>> fvUnstableA = mappingUnstable.get(0);
        SparseFeatureVector<Triple<Integer, Integer, Integer>> fvUnstableB = mappingUnstable.get(1);
        SparseFeatureVector<Triple<Integer, Integer, Integer>> fvStableA = mappingStableA.get(0);
        SparseFeatureVector<Triple<Integer, Integer, Integer>> fvStableB = mappingStableB.get(0);
        Assert.assertEquals(fvUnstableA.innerProduct(fvUnstableB), fvStableA.innerProduct(fvStableB));

    }
}
