package struclust.mcs;

import static org.junit.Assert.assertEquals;
import graph.DefaultGraph;
import graph.Graph;
import graph.Node;

import java.util.Collections;

/**
 * JUnit Test class for {@link MatrixAssociationGraph}
 * 
 * @author Till Schäfer
 */
public class MatrixAssociationGraphTest {

    /**
     * Creates two complete Graphs -> association graph entries must be 1
     * (except for self mappings and dual mappings);
     */
    @org.junit.Test
    @SuppressWarnings("rawtypes")
    public void completeIsomorphicGraphTest() {
        Graph<String, String> g1 = new DefaultGraph<>();
        Graph<String, String> g2 = new DefaultGraph<>();
        Graph graphs[] = { g1, g2 };

        int size = 10;

        for (int i = 0; i < size; i++) {
            g1.addNode("A");
            g2.addNode("A");
        }

        for (Graph g : graphs) {
            @SuppressWarnings("unchecked")
            Graph<String, String> gg = g;
            for (Node<String, String> n1 : gg.nodes()) {
                for (Node<String, String> n2 : gg.nodes()) {
                    if (n1 == n2) {
                        break;
                    }
                    gg.addEdge(n1, n2, "a");
                }
            }
        }

        MatrixAssociationGraph<String, String> mag = new MatrixAssociationGraph<>(g1, g2);
        int[][] adjacencies = mag.getAdjacencies();

        assertEquals(size * size, adjacencies.length);
        assertEquals(size * size, adjacencies[0].length);

        for (int i = 0; i < adjacencies.length; i++) {
            int[] row = adjacencies[i];
            for (int j = 0; j < row.length; j++) {
                int val = row[j];
                boolean equalG1Index = mag.getGraph1Indices(Collections.singletonList(i)).get(0) == mag
                        .getGraph1Indices(Collections.singletonList(j)).get(0);
                boolean equalG2Index = mag.getGraph2Indices(Collections.singletonList(i)).get(0) == mag
                        .getGraph2Indices(Collections.singletonList(j)).get(0);
                if (i == j || equalG1Index || equalG2Index) {
                    assertEquals(0, val);
                } else {
                    assertEquals(1, val);
                }
            }
        }
    }

    /**
     * Creates two completely disconnected Graphs -> association graph entries
     * must be -1 (except for self mappings and dual mappings);
     */
    @org.junit.Test
    public void disconnectedIsomorphicGraphTest() {
        Graph<String, String> g1 = new DefaultGraph<>();
        Graph<String, String> g2 = new DefaultGraph<>();

        int size = 10;

        for (int i = 0; i < size; i++) {
            g1.addNode("A");
            g2.addNode("A");
        }

        MatrixAssociationGraph<String, String> mag = new MatrixAssociationGraph<>(g1, g2);
        int[][] adjacencies = mag.getAdjacencies();

        assertEquals(size * size, adjacencies.length);
        assertEquals(size * size, adjacencies[0].length);

        for (int i = 0; i < adjacencies.length; i++) {
            int[] row = adjacencies[i];
            for (int j = 0; j < row.length; j++) {
                int val = row[j];
                boolean equalG1Index = mag.getGraph1Indices(Collections.singletonList(i)).get(0) == mag
                        .getGraph1Indices(Collections.singletonList(j)).get(0);
                boolean equalG2Index = mag.getGraph2Indices(Collections.singletonList(i)).get(0) == mag
                        .getGraph2Indices(Collections.singletonList(j)).get(0);
                if (i == j || equalG1Index || equalG2Index) {
                    assertEquals(0, val);
                } else {
                    assertEquals(-1, val);
                }
            }
        }
    }

    /**
     * Test by small example
     */
    @org.junit.Test
    public void smallExampleTest() {
        Graph<String, String> g1 = new DefaultGraph<>();
        Graph<String, String> g2 = new DefaultGraph<>();

        Node<String, String> a11 = g1.addNode("A");
        Node<String, String> a12 = g1.addNode("A");
        int b11 = g1.addNode("B").getIndex();
        g1.addEdge(a11, a12, "X");

        Node<String, String> a21 = g2.addNode("A");
        Node<String, String> a22 = g2.addNode("A");
        g2.addNode("B");
        g2.addEdge(a21, a22, "X");

        MatrixAssociationGraph<String, String> mag = new MatrixAssociationGraph<>(g1, g2);
        int[][] adjacencies = mag.getAdjacencies();

        assertEquals(5, adjacencies.length);
        assertEquals(5, adjacencies[0].length);

        for (int i = 0; i < adjacencies.length; i++) {
            int[] row = adjacencies[i];
            for (int j = 0; j < row.length; j++) {
                int val = row[j];

                boolean equalG1Index = mag.getGraph1Indices(Collections.singletonList(i)).get(0) == mag
                        .getGraph1Indices(Collections.singletonList(j)).get(0);
                boolean equalG2Index = mag.getGraph2Indices(Collections.singletonList(i)).get(0) == mag
                        .getGraph2Indices(Collections.singletonList(j)).get(0);
                boolean containsBNode = mag.getGraph1Indices(Collections.singletonList(i)).get(0) == b11
                        || mag.getGraph1Indices(Collections.singletonList(j)).get(0) == b11;

                if (i == j || equalG1Index || equalG2Index) {
                    assertEquals(0, val);
                } else if (containsBNode) {
                    assertEquals(-1, val);
                } else {
                    assertEquals(1, val);
                }
            }
        }
    }
}
