package struclust.mcs;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import graph.DefaultGraph;
import graph.DefaultGraphFactory;

import java.util.List;

import struclust.graph.generators.KuramochiKarypisRGG;

/**
 * JUnit Class to compare different implementations. Each implementation should
 * give the same result.
 * 
 * @author Till Schäfer
 *
 */
public class McsComparisonTests {

    /**
     * Test that randomly generates some graphs and computes the pairwise MCCIS
     * between them. The resulting MCCIS size should not differ for different
     * implementations.
     */
    @org.junit.Test
    public void randomMccisSizeTest() {
        KuramochiKarypisRGG generator = new KuramochiKarypisRGG(30, 10, 3, 3, 2);
        List<DefaultGraph<Integer, Integer>> graphs = generator.createGraphs(10);

        McsEnumAlgorithm<Integer, Integer, DefaultGraph<Integer, Integer>> mccisKriegeMEA = new MccisKriegeMEA<>(
                new DefaultGraphFactory<>());
        McsSizeAlgorithm<Integer, Integer, DefaultGraph<Integer, Integer>> mccisKriegeMSA = new MccisKriegeMSA<>();
        McsSizeAlgorithm<Integer, Integer, DefaultGraph<Integer, Integer>> mccisKriegeColorMSA = new MccisKriegeColorMSA<>();
        McsSizeAlgorithm<Integer, Integer, DefaultGraph<Integer, Integer>> mccisKriegeColor2MSA = new MccisKriegeColor2MSA<>();
        McsSizeDecisionAlgorithm<Integer, Integer, DefaultGraph<Integer, Integer>> mccisKriegeColor2MSDA = new MccisKriegeColor2MSDA<>();

        for (DefaultGraph<Integer, Integer> g1 : graphs) {
            for (DefaultGraph<Integer, Integer> g2 : graphs) {
                if (g1 == g2) {
                    break;
                }
                List<DefaultGraph<Integer, Integer>> mccisKriegeMEAGraphs = mccisKriegeMEA.calc(g1, g2);
                int mccisKriegeMEASize = mccisKriegeMEAGraphs.isEmpty() ? 0 : mccisKriegeMEAGraphs.get(0)
                        .getNodeCount();
                int mccisKriegeMSASize = mccisKriegeMSA.calc(g1, g2);
                int mccisKriegeColorMSASize = mccisKriegeColorMSA.calc(g1, g2);
                int mccisKriegeColor2MSASize = mccisKriegeColor2MSA.calc(g1, g2);
                boolean mccisKriegeColor2MSDAlower = mccisKriegeColor2MSDA.calc(g1, g2, mccisKriegeColor2MSASize);
                boolean mccisKriegeColor2MSDAupper = mccisKriegeColor2MSDA.calc(g1, g2, mccisKriegeColor2MSASize + 1);

                assertEquals("mccisKriegeMEASize differs from mccisKriegeMSASize", mccisKriegeMEASize,
                        mccisKriegeMSASize);
                assertEquals("mccisKriegeMSASize differs from mccisKriegeColorMSASize", mccisKriegeMSASize,
                        mccisKriegeColorMSASize);
                assertEquals("mccisKriegeColorMSASize differs from mccisKriegeColor2MSASize", mccisKriegeColorMSASize,
                        mccisKriegeColor2MSASize);
                assertTrue("mccisKriegeColor2MSDAlower does not align with mccisKriegeColor2MSASize",
                        mccisKriegeColor2MSDAlower);
                assertFalse("mccisKriegeColor2MSDAupper does not align with mccisKriegeColor2MSASize",
                        mccisKriegeColor2MSDAupper);
            }
        }
    }

    /**
     * Test that randomly generates some graphs and computes the pairwise MCIS
     * between them. The resulting MCIS size should not differ for different
     * implementations.
     */
    @org.junit.Test
    public void randomMcisSizeTest() {
        KuramochiKarypisRGG generator = new KuramochiKarypisRGG(20, 10, 3, 3, 2);
        List<DefaultGraph<Integer, Integer>> graphs = generator.createGraphs(6);

        McsSizeAlgorithm<Integer, Integer, DefaultGraph<Integer, Integer>> mcisProsserMcsaMSA = new McisProsserMcsaMSA<>();
        McsSizeAlgorithm<Integer, Integer, DefaultGraph<Integer, Integer>> mcisHsahnMSA = new McisHsahnMSA<>();

        for (DefaultGraph<Integer, Integer> g1 : graphs) {
            for (DefaultGraph<Integer, Integer> g2 : graphs) {
                if (g1 == g2) {
                    break;
                }
                int mcisProsserMcsaMSASize = mcisProsserMcsaMSA.calc(g1, g2);
                int mcisHsahnMSASize = mcisHsahnMSA.calc(g1, g2);

                assertEquals("mcisProsserMcsaMSASize differs from mcisHsahnMSASize", mcisProsserMcsaMSASize,
                        mcisHsahnMSASize);
            }
        }
    }
}
