package struclust.util;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;

import org.apache.commons.math3.util.Pair;

/**
 * JUnit test class for {@link Streams}
 * 
 * @author Till Schäfer
 */
public class StreamsTest {
    /**
     * Test the cross product stream with an example.
     */
    @org.junit.Test
    public void crossProductStreamExampleTest() {
        Collection<Integer> a = new LinkedList<>();
        Collection<Integer> b = new LinkedList<>();

        a.add(1);
        a.add(2);
        a.add(3);

        b.add(5);
        b.add(6);

        List<Pair<Integer, Integer>> pairs = Streams.crossProductStream(a, b).collect(Collectors.toList());

        assertEquals("Number of Pairs is wrong", a.size() * b.size(), pairs.size());

        for (Integer iA : a) {
            for (Integer iB : b) {
                assertTrue(pairs.contains(Pair.create(iA, iB)));
            }
        }
    }

    /**
     * Test the cross product stream with an example in parallel.
     */
    @org.junit.Test
    public void crossProductStreamParallelExampleTest() {
        Collection<Integer> a = new LinkedList<>();
        Collection<Integer> b = new LinkedList<>();

        for (int i = 0; i < 100; i++) {
            a.add(i);
        }

        for (int i = 100; i < 200; i++) {
            b.add(i);
        }

        List<Pair<Integer, Integer>> pairs = Streams.crossProductStream(a, b).parallel().collect(Collectors.toList());

        assertEquals("Number of Pairs is wrong", a.size() * b.size(), pairs.size());

        for (Integer iA : a) {
            for (Integer iB : b) {
                assertTrue(pairs.contains(Pair.create(iA, iB)));
            }
        }
    }

    /**
     * Test the cross product stream with the second collection is empty
     */
    @org.junit.Test
    public void crossProductStreamEmptyTest() {
        Collection<Integer> a = new LinkedList<>();
        Collection<Integer> b = new LinkedList<>();

        b.add(5);
        b.add(6);

        List<Pair<Integer, Integer>> pairs = Streams.crossProductStream(a, b).collect(Collectors.toList());

        assertEquals("Number of Pairs must be zero if one Collection is empty", 0, pairs.size());
    }

    /**
     * Test the cross product stream with the first collection is empty
     */
    @org.junit.Test
    public void crossProductStreamEmpty2Test() {
        Collection<Integer> a = new LinkedList<>();
        Collection<Integer> b = new LinkedList<>();

        a.add(5);
        a.add(6);

        List<Pair<Integer, Integer>> pairs = Streams.crossProductStream(a, b).collect(Collectors.toList());

        assertEquals("Number of Pairs must be zero if one Collection is empty", 0, pairs.size());
    }

    /**
     * Test the ordered pairs stream with an example.
     */
    @org.junit.Test
    public void orderedPairsSpliteratorExampleTest() {
        Collection<Integer> a = new LinkedList<>();

        a.add(1);
        a.add(2);
        a.add(3);
        a.add(4);
        a.add(5);

        List<Pair<Integer, Integer>> pairs = Streams.orderedPairsStream(a).collect(Collectors.toList());

        assertEquals("Number of Pairs is wrong", (int) (Math.pow(a.size(), 2) - a.size()) / 2, pairs.size());

        for (Integer i1 : a) {
            for (Integer i2 : a) {
                if (i1 != i2) {
                    int count = 0;
                    count += java.util.Collections.frequency(pairs, Pair.create(i1, i2));
                    count += java.util.Collections.frequency(pairs, Pair.create(i2, i1));
                    assertEquals(1, count);
                }
            }
        }
    }

    /**
     * Test the ordered pairs stream with an example in parallel.
     */
    @org.junit.Test
    public void orderedPairsSpliteratorParallelExampleTest() {
        Collection<Integer> a = new LinkedList<>();

        for (int i = 0; i < 100; i++) {
            a.add(i);
        }

        List<Pair<Integer, Integer>> pairs = Streams.orderedPairsStream(a).parallel().collect(Collectors.toList());

        assertEquals("Number of Pairs is wrong", (int) (Math.pow(a.size(), 2) - a.size()) / 2, pairs.size());

        for (Integer i1 : a) {
            for (Integer i2 : a) {
                if (i1 != i2) {
                    int count = 0;
                    count += java.util.Collections.frequency(pairs, Pair.create(i1, i2));
                    count += java.util.Collections.frequency(pairs, Pair.create(i2, i1));
                    assertEquals(1, count);
                }
            }
        }
    }
    
    /**
     * Test the ordered pairs stream with an empty input.
     */
    @org.junit.Test
    public void orderedPairsSpliteratorEmptyTest() {
        Collection<Integer> a = new LinkedList<>();
        
        List<Pair<Integer, Integer>> pairs = Streams.orderedPairsStream(a).collect(Collectors.toList());

        assertEquals("Number of Pairs must be zero if the Collection is empty", 0, pairs.size());
    }
}
