package struclust.comparison;

import graph.DefaultGraph;
import graph.DefaultGraphFactory;
import graph.Graph;
import graph.LabelEdge;

import java.util.Date;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import struclust.SharedMemorySCC;
import struclust.graph.GraphContainer;
import struclust.graph.Graphs;
import struclust.mcs.MccisKriegeMEA;

import com.google.common.base.Preconditions;

/**
 * The value of this distance is <it>-1 * |mcs(pattern, target)|^pow</it> if
 * there exists an fixed percentage partial embedding of the pattern graph in
 * the target/host graph and 0 otherwise. A fixed percentage partial embedding
 * means that the mcs covers at least a fixed percentage of the pattern graph.
 * 
 * <br>
 * The distance is non-symmetric.
 * 
 * @author Till Schäfer
 *
 * @param <NL>
 *            the node label type
 * @param <EL>
 *            the edge label type
 * @param <G>
 *            the graph type
 */
public class FuzzySubIsoGGD<NL, EL, G extends Graph<NL, EL>> implements GraphGraphDistance<NL, EL, G> {
    private static final Logger logger = LoggerFactory.getLogger(FuzzySubIsoGGD.class);
    private boolean nodeInduced;
    private double patternPercentage;
    private SharedMemorySCC<NL, EL, G> conf;
    private double pow;

    /**
     * Constructor
     * 
     * @param nodeInduced
     *            iff true the node induced MCS is computed, otherwise the edge
     *            induced.
     * @param patternPercentage
     *            the percentage value between 0 and 1 for the fixed percentage
     *            partial embedding
     * @param pow
     *            the distance is powered by this value
     * @param conf
     *            the configuration
     */
    public FuzzySubIsoGGD(boolean nodeInduced, double patternPercentage, double pow, SharedMemorySCC<NL, EL, G> conf) {
        Preconditions.checkArgument(patternPercentage <= 1 && patternPercentage >= 0);

        this.nodeInduced = nodeInduced;
        this.patternPercentage = patternPercentage;
        this.conf = conf;
        this.pow = pow;
    }

    @Override
    public double calc(GraphContainer<NL, EL, G> pattern, GraphContainer<NL, EL, G> target) {
        Date before = new Date();

        int patternCount = nodeInduced ? pattern.getNodeCount() : pattern.getEdgeCount();
        int targetCount = nodeInduced ? target.getNodeCount() : target.getEdgeCount();

        List<? extends Graph<?, ?>> mcciss;
        if (nodeInduced) {
            MccisKriegeMEA<NL, EL, G> algo = new MccisKriegeMEA<>(conf.gFactory);
            mcciss = algo.calc(pattern.getGraph(), target.getGraph(), (int) (patternCount * patternPercentage + 0.5));
        } else {
            DefaultGraph<LabelEdge<NL, EL>, NL> lineGraphP = Graphs.lineGraph(pattern.getGraph());
            DefaultGraph<LabelEdge<NL, EL>, NL> lineGraphT = Graphs.lineGraph(target.getGraph());

            MccisKriegeMEA<LabelEdge<NL, EL>, NL, DefaultGraph<LabelEdge<NL, EL>, NL>> algo = new MccisKriegeMEA<>(
                    new DefaultGraphFactory<>());
            mcciss = algo.calc(lineGraphP, lineGraphT, (int) (patternCount * patternPercentage + 0.5));
        }

        int mcs = mcciss.isEmpty() ? 0 : mcciss.get(0).getNodeCount();

        boolean fixePercPartEmbedding = mcs >= patternCount * patternPercentage;
        double retVal = fixePercPartEmbedding ? -1.0 * Math.pow(pattern.size(conf.gSize), pow) : 0;

        Date after = new Date();
        logger.debug(
                "Value: {}, MCS Size: {} ({}), percentage: {}, pow: {}, count1: {}, count2: {}, time to compute: {}",
                retVal, mcs, nodeInduced ? "nodeInduced" : "edgeInduced", patternPercentage, pow, patternCount,
                targetCount, after.getTime() - before.getTime());
        return retVal;
    }

    @Override
    public double maxDist() {
        return 0;
    }

    @Override
    public String getDescription() {
        return "FuzzySubIsoGGD (" + (nodeInduced ? "node induced" : "edge induced") + ", patternPercentage: "
                + patternPercentage + ", pow: " + pow + ")";
    }

}
